SystemOrganization addCategory: #'SMBase-UI'!
SystemOrganization addCategory: #'SMBase-domain'!
SystemOrganization addCategory: #'SMBase-installer'!
SystemOrganization addCategory: #'SMBase-utilities'!

!UUID class methodsFor: '*smbase-macsafe' stamp: 'dvf 9/10/2004 23:11'!
fromString36: aString
	"Decode the UUID from a base 36 string using 0-9 and lowercase a-z.
	This is the shortest representation still being able to work as
	filenames etc since it does not depend on case nor characters
	that might cause problems."

	| object num |
	object := self nilUUID.
	num := Integer readFrom: aString asUppercase readStream base: 36.
	16 to: 1 by: -1 do: [:i |
		num size < i
			ifTrue: [object at: i put: 0]
			ifFalse: [object at: i put: (num digitAt: i)]].
	^object! !

!UUID methodsFor: '*smbase-macsafe' stamp: 'gk 4/18/2006 22:15'!
asString36
	"Encode the UUID as a base 36 string using 0-9 and lowercase a-z.
	This is the shortest representation still being able to work as
	filenames etc since it does not depend on case nor characters
	that might cause problems, and it fits into short filenames like on
	the old MacOS HFS filesystem. The check for 36r is to make this code
	work in versions before Squeak 3.8."

	| num candidate |
	num := 0.
	1 to: self size do: [:i | num := num + ((256 raisedTo: i - 1) * (self at: i))].
	candidate := num printStringBase: 36.
	^((candidate beginsWith: '36r')
			ifTrue: [candidate copyFrom: 4 to: candidate size]
			ifFalse: [candidate]) asLowercase! !

!ImageSegment methodsFor: '*SMBase-export' stamp: 'stephaneducasse 2/4/2006 20:38'!
writeForExportOn: fileStream
	"Write the segment on the disk with all info needed to reconstruct it in a new image.  For export.  Out pointers are encoded as normal objects on the disk."

	| temp |
	state = #activeCopy ifFalse: [self error: 'wrong state'].
	temp := endMarker.
	endMarker := nil.
	fileStream fileOutClass: nil andObject: self.
		"remember extra structures.  Note class names."
	endMarker := temp.
! !

!SystemVersion methodsFor: '*smbase-extension' stamp: 'jcg 11/2/2004 10:03'!
majorMinorVersion
	"Return the major/minor version number of the form X.Y, without any 'alpha' or 'beta' or other suffix."
	"(SystemVersion new version: 'Squeak3.7alpha') majorMinorVersion" "  -->  'Squeak3.7' "
	"SystemVersion current majorMinorVersion"
	
	| char stream |
	stream := ReadStream on: version, 'x'.
	stream upTo: $..
	char := stream next.
	char ifNil: [^ version].	"eg: 'Jasmine-rc1' has no $. in it."
	[char isDigit]
		whileTrue: [char := stream next].
	^ version copyFrom: 1 to: stream position - 1
! !

!VersionNumber methodsFor: '*smbase-printing' stamp: 'gk 1/23/2004 10:13'!
versionString

	^String streamContents: [ :strm | self versionStringOn: strm ]! !

!VersionNumber methodsFor: '*smbase-override' stamp: 'svp 6/18/2002 17:23'!
versionStringOn: strm

	| first |
	first := true.
	numbers do: [ :ea |
		first ifFalse: [strm nextPut: $.].
		first := false.
		ea printOn: strm]
	! !

TestCase subclass: #SMDependencyTest
	instanceVariableNames: 'engine map goranAccount'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMDependencyTest commentStamp: '<historical>' prior: 0!
Tests for the dependency engine.!

!SMDependencyTest methodsFor: 'as yet unclassified' stamp: 'gk 5/5/2006 02:07'!
setUp
	| package trivial1rel installed1rel installed2rel tricky2rel conf1 conf2 tricky3rel1 tricky3rel2 tricky1rel seaside httpview kom1 kom2 |
	map := SMSqueakMap new reload.
	goranAccount := map newAccount: 'Goran' username: 'Goran' email: 'g@g.com'.
	
	"Add a few packages to test with:
	
	Tricky1 1
		Installed1 1
		Tricky2 1
	Tricky2 1
		Installed1 1
		TrivialToInstall1 1
		Tricky3 1
	
		Installed1 1
		TrivialToInstall1 1
		Tricky3 2
	Tricky3 2
		TrivialToInstall1 1

		Installed2 1
	
	Seaside
		KomHttpServer 1
	
	HttpView
		KomHttpServer 2
	"
	{
		{'A'. {'Squeak3.6'. 'Squeak3.7'. 'Stable'}. 3}.
		{'B'. {'Squeak3.6'. 'Stable'}. 2}.
		{'TrivialToInstall1'. {'Squeak3.6'. 'Squeak3.7'. 'Stable'}. 1}.
		{'Installed1'. {'Squeak3.6'. 'Squeak3.7'. 'Stable'}. 1}.
		{'Installed2'. {'Squeak3.6'. 'Squeak3.7'. 'Stable'}. 1}.
		{'AlreadyInstallable1'. {'Squeak3.6'. 'Squeak3.7'. 'Stable'}. 1}.
		{'Tricky1'. {'Squeak3.6'. 'Squeak3.7'. 'Stable'}. 1}.
		{'Tricky2'. {'Squeak3.6'. 'Squeak3.7'. 'Stable'}. 1}.
		{'Tricky3'. {'Squeak3.6'. 'Squeak3.7'. 'Stable'}. 2}.
		{'Circular1'. {'Squeak3.6'. 'Squeak3.7'. 'Stable'}. 1}.
		{'Circular2'. {'Squeak3.6'. 'Squeak3.7'. 'Stable'}. 1}.
		{'Circular3'. {'Squeak3.6'. 'Squeak3.7'. 'Stable'}. 1}.
		{'Seaside'. {'Squeak3.6'. 'Squeak3.7'. 'Stable'}. 1}.
		{'KomHttpServer'. {'Squeak3.6'. 'Squeak3.7'. 'Stable'}. 2}.
		{'HttpView'. {'Squeak3.6'. 'Squeak3.7'. 'Stable'}. 1}.
	} do: [:arr |
			package := SMPackage newIn: map.
			package name: arr first.
			arr second do: [:cn | package addCategory: (map categoryWithNameBeginning: cn)].
			arr third timesRepeat: [package newRelease ].
			goranAccount addObject: package].
	
	trivial1rel := (map packageWithName: 'TrivialToInstall1') lastRelease.
	trivial1rel publisher: goranAccount.
	
	installed1rel := (map packageWithName: 'Installed1') lastRelease.
	installed1rel publisher: goranAccount; noteInstalled.
	installed2rel := (map packageWithName: 'Installed2') lastRelease.
	installed2rel publisher: goranAccount; noteInstalled.

	((map packageWithName: 'AlreadyInstallable1') lastRelease
		publisher: goranAccount;
		addConfiguration)
				addRequiredRelease: installed1rel.
	"Tricky1 has just a single configuration with one installed and one not installed."
	tricky1rel := (map packageWithName: 'Tricky1') lastRelease.
	tricky2rel := (map packageWithName: 'Tricky2') lastRelease.			
	(tricky1rel publisher: goranAccount; addConfiguration)
				addRequiredRelease: installed1rel; "already installed"
				addRequiredRelease: tricky2rel. "not installed"

	"Tricky2 has two configurations:
		1: an installed, a trivial one and Tricky3 r1.
		2: an installed, a trivial one and Tricky3 r2."
	conf1 := tricky2rel publisher: goranAccount; addConfiguration.
	conf2 := tricky2rel addConfiguration.
	
	tricky3rel1 := (map packageWithName: 'Tricky3') releases first.
	tricky3rel2 := (map packageWithName: 'Tricky3') lastRelease.
	tricky3rel1 publisher: goranAccount.
	tricky3rel2 publisher: goranAccount.

	conf1 addRequiredRelease: installed1rel; addRequiredRelease: trivial1rel; addRequiredRelease: tricky3rel1.
	conf2 addRequiredRelease: installed1rel; addRequiredRelease: trivial1rel; addRequiredRelease: tricky3rel2.
	
	"Tricky3rel2 has two configurations:
		1: trivial1
		2: installed2rel"
	conf1 := tricky3rel2 publisher: goranAccount; addConfiguration.
	conf2 := tricky3rel2 addConfiguration.
	conf1 addRequiredRelease: trivial1rel.
	conf2 addRequiredRelease: installed2rel.
	

	seaside := (map packageWithName: 'Seaside') lastRelease.
	seaside publisher: goranAccount.
	httpview := (map packageWithName: 'HttpView') lastRelease.
	httpview publisher: goranAccount.
	kom1 := (map packageWithName: 'KomHttpServer') firstRelease.
	kom1 publisher: goranAccount.
	kom2 := (map packageWithName: 'KomHttpServer') lastRelease.
	kom2 publisher: goranAccount.
	
	conf1 := seaside addConfiguration.
	conf2 := httpview addConfiguration.
	conf1 addRequiredRelease: kom1.
	conf2 addRequiredRelease: kom2! !

!SMDependencyTest methodsFor: 'as yet unclassified' stamp: 'gk 5/5/2006 02:07'!
test1
	"Test three different scenarios, a trivial, a simple
	and one with conflicting solutions."

	| task bestPath |
	engine := SMDependencyEngine map: map.
	task := engine installPackages: map installedPackages.
	self assert: (task calculate).
	self assert: (task analysis alreadyInstalled allSatisfy: [:r | r isInstalled]).
	self assert: (task analysis trivialToInstall isEmpty).
	self assert: (task analysis alreadyInstallable isEmpty).
	self assert: (task analysis trickyReleases isEmpty).
	self assert: (task allInstallPaths first isEmpty).
	self assert: (task allInstallPaths size = 1).

	"Test to install a trivial one, and an already installable one."
	engine := SMDependencyEngine map: map.
	task := engine installPackages: {map packageWithName: 'TrivialToInstall1'. map packageWithName: 'AlreadyInstallable1'}.
	self assert: (task calculate).
	self assert: (task analysis alreadyInstalled isEmpty).
	self assert: (task analysis trivialToInstall size = 1).
	self assert: (task analysis alreadyInstallable size = 1).
	self assert: (task analysis trickyReleases isEmpty).
	self assert: (task allInstallPaths size = 1).
	self assert: (task allInstallPaths first size = 2).
	
	"Test to install two packages, there are two solutions without conflicts."
	engine := SMDependencyEngine map: map.
	task := engine installPackages: {map packageWithName: 'Tricky1'. map packageWithName: 'Tricky2'}.
	self assert: (task calculate ).
	self assert: (task analysis alreadyInstalled isEmpty).
	self assert: (task analysis trivialToInstall isEmpty).
	self assert: (task analysis alreadyInstallable isEmpty).
	self assert: (task analysis trickyReleases size = 2).
	self assert: (task allInstallPaths size = 4).
	self assert: (task analysis allNormalizedInstallPaths size = 2).
	"Make sure the 4 different install paths are computed correctly. Since the algorithm uses Sets
	the actual installs can vary, that is why we sort etc to check it."
	self assert: ((task allInstallPaths collect: [:oc |
		(oc collect: [:r | r packageNameWithVersion ]) asSortedCollection asString]) asSortedCollection asString =  'a SortedCollection(''a SortedCollection(''''Tricky1 1'''' ''''Tricky2 1'''' ''''Tricky3 1'''' ''''Tricky3 2'''' ''''TrivialToInstall1 1'''')'' ''a SortedCollection(''''Tricky1 1'''' ''''Tricky2 1'''' ''''Tricky3 1'''' ''''Tricky3 2'''' ''''TrivialToInstall1 1'''')'' ''a SortedCollection(''''Tricky1 1'''' ''''Tricky2 1'''' ''''Tricky3 1'''' ''''TrivialToInstall1 1'''')'' ''a SortedCollection(''''Tricky1 1'''' ''''Tricky2 1'''' ''''Tricky3 2'''' ''''TrivialToInstall1 1'''')'')').
	self assert: (task analysis installPathsWithoutConflicts size = 2).
	bestPath := task analysis bestInstallPath.
	self assert: (bestPath size = 4).
	self assert: (bestPath anySatisfy: [:r | r packageNameWithVersion = 'Tricky3 2']).
	! !

!SMDependencyTest methodsFor: 'as yet unclassified' stamp: 'gk 5/5/2006 02:07'!
test2
	"Checking construction of model:
	- verify test for circular dependencies (not allowed)."

	"Circular1rel has a conf to Circular2rel, which in turn has one back."
	| circular1rel circular2rel circular3rel |
	circular1rel := (map packageWithName: 'Circular1') lastRelease.
	circular2rel := (map packageWithName: 'Circular2') lastRelease.
	circular3rel := (map packageWithName: 'Circular3') lastRelease.

	(circular1rel publisher: goranAccount; addConfiguration)
		addRequiredRelease: circular2rel.

	(circular2rel publisher: goranAccount; addConfiguration)
		addRequiredRelease: circular3rel.
	
	self should: [
		(circular3rel publisher: goranAccount; addConfiguration)
			addRequiredRelease: circular1rel]
		raise: Error! !

!SMDependencyTest methodsFor: 'as yet unclassified' stamp: 'gk 5/5/2006 02:07'!
test3
	"Test simplest conflict."

	| task |
	"Test to install two packages, there is one solution with one conflict."
	engine := SMDependencyEngine map: map.
	task := engine installPackages: {map packageWithName: 'Seaside'. map packageWithName: 'HttpView'}.
	self assert: (task calculate).
	self assert: (task allInstallPaths size = 1).
	self assert: (task analysis installPathsWithoutConflicts size = 0).
	self assert: (task analysis allNormalizedInstallPaths size = 1).
	self assert: (task analysis bestInstallPath = nil).
	self assert: (task analysis untestedInstallPaths size = 1).
	self assert: (task proposals size = 1).
	self assert: (task proposals first hasDeviations)! !

Object subclass: #RcsDiff
	instanceVariableNames: 'commandLines'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-utilities'!

!RcsDiff classSide methodsFor: 'instance creation' stamp: 'gh 11/22/2001 23:44'!
lines: aString
	"Create a new RcsDiff file."
	^(self new) commandLines: aString; yourself! !

!RcsDiff methodsFor: 'applying' stamp: 'stephaneducasse 2/4/2006 20:38'!
applyTo: aString
	"Apply me to given String and return the patched String."

	| space commandStream originalStream nextCommand nextLine lineCount currentLine |
	space := Character space.
	commandStream := ReadStream on: commandLines.
	originalStream := ReadStream on: aString.
	currentLine := 1.
	^String streamContents: [:stream |
		[nextCommand := commandStream next.
		nextCommand isNil] whileFalse: [ 
			nextLine := (commandStream upTo: space) asNumber.
			lineCount := commandStream nextLine asNumber.
			[currentLine = nextLine]
				whileFalse: [stream nextPutAll: originalStream nextLine; cr. currentLine := currentLine + 1].
			nextCommand = $d
				ifTrue:[ lineCount timesRepeat: [originalStream nextLine. currentLine := currentLine + 1]]
				ifFalse:[ nextCommand = $a
							ifTrue:[ stream nextPutAll: originalStream nextLine; cr.
									currentLine := currentLine + 1.
									lineCount timesRepeat: [
										stream nextPutAll: commandStream nextLine; cr]]]].
		stream nextPutAll: originalStream upToEnd]! !

!RcsDiff methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
commandLines: aString
	commandLines := aString! !

Object subclass: #SMDependencyAnalysis
	instanceVariableNames: 'task map wantedReleases alreadyInstalled trivialToInstall alreadyInstallable combinations suggestedInstallSetsSet workingConfigurations workingConfigurationsSize conflictingInstallSetsSet trickyReleases subAnalysises success'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMDependencyAnalysis commentStamp: '<historical>' prior: 0!
A dependency analysis is instantiated by an SMInstallationTask as a step in calculating how the task can be performed.

The analysis is done using a map and some input - for example a list of package releases that the task wants to get installed. It can then be queried for the results. The analysis is performed in a series of steps and middle results are collected in instvars. It also uses sub instances so the analysis actually forms a tree of several instances of SMDependencyAnalysis where each node describes one level of dependencies.

Instvars:

task - the task that instantiated this analysis.
map - the SMSqueakMap to use, we get it by asking the task.
wantedReleases - the releases that we want to install.
alreadyInstalled - the subset of wantedReleases that are already installed.
trivialToInstall - the subset of wantedReleases that do not have dependencies and can be installed directly.
alreadyInstallable - 	the subset of wantedReleases that do have dependencies but for which at least one configuration is fulfilled and thus the release can already be installed.
trickyReleases - the subset of wantedReleases that do have configurations but none of them are fulfilled, so some dependencies first needs to be installed before these releases can be installed.

workingConfigurations - an OrderedCollection of OrderedCollections holding all working configurations for the trickyReleases.
workingConfigurationsSize - size of workingConfigurations.
combinations - all possible (unordered) combinations of picking one working configuration for each tricky release.
suggestedInstallSetsSet - computed from combinations above. A Set of Sets of required releases. Each Set is a unique combination of the required releases to install in order to fulfill one configuration for each of the tricky releases.
conflictingInstallSetsSet - the subset of suggestedInstallSetsSet that are invalid since it includes multiple releases from the same package.
 !

!SMDependencyAnalysis classSide methodsFor: 'instance creation' stamp: 'gk 7/29/2004 14:31'!
task: ownerTask

	^self new task: ownerTask! !

!SMDependencyAnalysis methodsFor: 'queries' stamp: 'gk 5/5/2006 02:07'!
allInstallPaths
	"For all paths, collect in reverse all releases to install.
	At each level, first we add trivially installable releases
	(those that have no dependencies), then installable releases
	(those that have one configuration fulfilled) and finally
	the tricky releases (those left).
	Note that we also return paths with conflicting releases
	of the same package and paths with releases that conflict with
	already installed releases - those paths can be tweaked - and
	paths that are supersets of other paths."

	| installPaths releases |
	installPaths := OrderedCollection new.
	self allPathsDo: [:path |
		releases := OrderedCollection new.
		path reverseDo: [:ana |
			releases addAll: (ana trivialToInstall difference: releases).
			releases addAll: (ana alreadyInstallable difference: releases).
			releases addAll: (ana trickyReleases difference: releases)
			"Below for debugging
			r := OrderedCollection new.
			r add: ana trivialToInstall; add: ana alreadyInstallable; add: ana trickyReleases.
			releases add: r"].
		installPaths add: releases].
	^ installPaths! !

!SMDependencyAnalysis methodsFor: 'queries' stamp: 'gk 5/5/2006 02:07'!
allNormalizedInstallPaths
	"Same as allInstallPaths, but with paths removed that
	are clear supersets of others."

	| installPaths |
	installPaths := self allInstallPaths.
	installPaths := installPaths reject: [:p1 |
					installPaths anySatisfy: [:p2 |
						(p1 ~~ p2) and: [p1 includesAllOf: p2]]].
	^installPaths! !

!SMDependencyAnalysis methodsFor: 'private' stamp: 'gk 9/22/2004 20:22'!
allPathsDo: aBlock
	"For all paths down the tree, evaluate aBlock."

	^ self allPathsDo: aBlock trail: OrderedCollection new! !

!SMDependencyAnalysis methodsFor: 'private' stamp: 'gk 9/22/2004 20:59'!
allPathsDo: aBlock trail: trail
	"For all paths down the tree, evaluate aBlock."

	trail add: self.
	subAnalysises
		ifNil: [
			aBlock value: trail.]
		ifNotNil: [
			subAnalysises do: [:sub |
				sub allPathsDo: aBlock trail: trail]].
	trail removeLast! !

!SMDependencyAnalysis methodsFor: 'private' stamp: 'gk 5/5/2006 02:07'!
allRoutesDo: aBlock currentRoute: currentRoute level: level
	"Recursively iterate over all routes down the tree."

	| newLevel |
	workingConfigurationsSize = level ifTrue: ["we reached the leaves"
		workingConfigurations last do: [:conf | 
			currentRoute addLast: conf.
			aBlock value: currentRoute.
			currentRoute removeLast].
		^self].
	newLevel := level + 1.
	(workingConfigurations at: level) do: [:conf |
		currentRoute addLast: conf.
		self allRoutesDo: aBlock currentRoute: currentRoute level: newLevel.
		currentRoute removeLast]! !

!SMDependencyAnalysis methodsFor: 'accessing' stamp: 'gk 7/27/2004 15:36'!
alreadyInstallable
	^alreadyInstallable
	! !

!SMDependencyAnalysis methodsFor: 'accessing' stamp: 'gk 7/27/2004 15:36'!
alreadyInstalled
	^alreadyInstalled
	! !

!SMDependencyAnalysis methodsFor: 'queries' stamp: 'gk 5/5/2006 02:07'!
bestInstallPath
	"Using some heuristics we suggest the best path:
		- No conflicts
		- Fewest releases
		- If same packages, the newest releases"

	| paths min points point package sc |
	paths := self installPathsWithoutConflicts.
	paths size = 1 ifTrue: [^paths first].
	min := paths inject: 999 into: [:mi :p | p size < mi ifTrue: [p size] ifFalse: [mi]].
	paths := paths select: [:p | p size = min].
	paths size = 1 ifTrue: [^paths first].
	"Try to pick the one with newest releases"
	points := Dictionary new.
	paths do: [:p |
		point := 0.
		p do: [:r |
			package := r package.
			paths do: [:p2 |
				p2 == p ifFalse: [
					(p2 anySatisfy: [:r2 |
						(r2 package == package) and: [r newerThan: r2]])
							ifTrue:[point := point + 1]]]].
		points at: p put: point].
	points isEmpty ifTrue: [^nil].
	sc := points associations asSortedCollection: [:a :b | a value >= b value].
	^ sc first key! !

!SMDependencyAnalysis methodsFor: 'private' stamp: 'gk 5/5/2006 02:07'!
collectCombinationsOfConfigurations
	"Given the wanted releases, find and return all possible combinations
	of working configurations for all those. Perhaps not possible to do
	given lots of releases and configurations, then we need smarter algorithms."
	
	"Pick out all working configurations first."
	workingConfigurations := (trickyReleases collect: [:r | r workingConfigurations]) asOrderedCollection.
	workingConfigurationsSize := workingConfigurations size.
	
	"We iterate over all possible combinations of configurations
	and collect the unique set of unordered configurations."
	combinations := Set new.
	self allRoutesDo: [:route |
		combinations add: route asSet copy] currentRoute: OrderedCollection new level: 1! !

!SMDependencyAnalysis methodsFor: 'private' stamp: 'gk 5/5/2006 02:07'!
computeInstallSets
	"Given all combinations of configurations, compute all valid combinations
	of depdendency releases - all possible different Sets of required releases
	to install before the trickyReleases can be installed."
	
	"For each unique combination of configurations, collect all required releases
	and produce a Set of unique required release combinations." 
	suggestedInstallSetsSet := (combinations collect: [:comb |
								comb inject: Set new into: [:set :conf |
									set addAll: conf requiredReleases.
									set ]]) asSet.

	"Filter out those Sets that have multiple releases of the same package, they are conflicting
	and thus not valid - we can't have two different releases of the same package
	installed at the same time."
	
	"conflictingInstallSetsSet := suggestedInstallSetsSet select:
								[:set | self detectConflictingReleasesIn: set].
	suggestedInstallSetsSet removeAll: conflictingInstallSetsSet"! !

!SMDependencyAnalysis methodsFor: 'private' stamp: 'gk 5/5/2006 02:07'!
detectConflictingReleasesIn: collectionOfReleases
	"Detect if the Set has multiple releases of the same package."

	| detectedPackages |
	detectedPackages := Set new.
	collectionOfReleases do: [:r |
		(detectedPackages includes: r package)
			ifTrue: [^ true]
			ifFalse: [detectedPackages add: r package]].
	^false! !

!SMDependencyAnalysis methodsFor: 'printing' stamp: 'gk 9/21/2004 23:59'!
indent: level
	^'                                                  '
		last: level * 6! !

!SMDependencyAnalysis methodsFor: 'calculation' stamp: 'gk 5/5/2006 02:07'!
installPackageReleases: packageReleases
	"Given a Set of wanted SMPackageReleases, calculate all possible
	installation scenarios. If the analysis succeeds, return true, otherwise false."
	
	| result subAnalysis |
	wantedReleases := packageReleases copy.
	"First classify the releases in different groups."
	self partitionReleases.
	
	"If there are no tricky releases, we are already done.
	No extra required releases needs to be installed or upgraded."
	trickyReleases isEmpty ifTrue: [^success := true].

	"Ok, that was the easy part. The releases left now needs to be processed
	so that we can find out the different scenarios of required releases that we need
	to install first. First we calculate all combinations of available working configurations
	for the tricky releases."
	self collectCombinationsOfConfigurations.
	
	"Based on all configuration combinations,
	compute possible combinations of dependency releases."
	self computeInstallSets.
	
	"Check if we have failed - meaning that there are no valid scenarios without conflicts."
	suggestedInstallSetsSet isEmpty ifTrue: [^success := false].
	
	"Ok, this means we have at least one solution *on this level*!! But we need to do the
	analysis recursively for all these sets of required releases..."
	subAnalysises := OrderedCollection new.
	success := false.
	suggestedInstallSetsSet do: [:set |
		subAnalysis := SMDependencyAnalysis task: task.
		result := subAnalysis installPackageReleases: set.
		result ifTrue: [success := true].
		subAnalysises add: subAnalysis].
	
	"Did at least one succeed? If so, then we have at least one possible scenario!!
	If not, then we need to do tweaking."
	^success! !

!SMDependencyAnalysis methodsFor: 'queries' stamp: 'gk 9/26/2004 23:44'!
installPathsWithConflicts
	"Same as allInstallPaths, but we only return paths
	with multiple releases of the same package."

	^ self allInstallPaths select: [:path | self detectConflictingReleasesIn: path] ! !

!SMDependencyAnalysis methodsFor: 'queries' stamp: 'gk 9/22/2004 23:32'!
installPathsWithoutConflicts
	"Same as allInstallPaths, but we filter out paths
	with multiple releases of the same package."

	^ self allInstallPaths reject: [:path | self detectConflictingReleasesIn: path] ! !

!SMDependencyAnalysis methodsFor: 'private' stamp: 'gk 5/5/2006 02:07'!
partitionReleases
	"Move releases from wantedReleases to suitable other collections
	if they are either installed, trivial to install, or installable as is."
	
	trickyReleases := wantedReleases copy.
	alreadyInstalled := wantedReleases select: [:r | r isInstalled ].
	trickyReleases removeAll: alreadyInstalled. 
	trivialToInstall := trickyReleases select: [:r | r hasNoConfigurations ].
	trickyReleases removeAll: trivialToInstall.		
	alreadyInstallable := trickyReleases select: [:r | r hasFulfilledConfiguration ].
	trickyReleases removeAll: alreadyInstallable! !

!SMDependencyAnalysis methodsFor: 'printing' stamp: 'gk 9/22/2004 22:44'!
printAllInstallPaths
	"Follow all install paths in the tree."

	^String streamContents: [:s |
		self allInstallPaths do: [:path |
			path do: [:rel |
				s nextPutAll: rel packageNameWithVersion, ', '].
			s cr]] ! !

!SMDependencyAnalysis methodsFor: 'private' stamp: 'gk 5/5/2006 02:07'!
removeOlderReleasesIn: collectionOfReleases
	"Remove older multiple releases of the same package.
	2 scans to retain order."

	| newestReleases rel |
	newestReleases := Dictionary new.
	collectionOfReleases do: [:r |
		rel := newestReleases at: r package ifAbsent: [newestReleases at: r package put: r].
		(r newerThan: rel) ifTrue: [newestReleases at: r package put: r]].
	^collectionOfReleases select: [:r |
		(newestReleases at: r package) == r]! !

!SMDependencyAnalysis methodsFor: 'accessing' stamp: 'gk 9/20/2004 22:59'!
success
	^success
	! !

!SMDependencyAnalysis methodsFor: 'accessing' stamp: 'gk 5/5/2006 02:07'!
task: ownerTask
	task := ownerTask.
	map := task map! !

!SMDependencyAnalysis methodsFor: 'printing' stamp: 'gk 9/21/2004 23:49'!
treeString
	"Return a indented String showing the tree
	structure of all possible scenarios."

	^String streamContents: [:s |
		self treeStringOn: s indent: 0]! !

!SMDependencyAnalysis methodsFor: 'printing' stamp: 'gk 5/5/2006 02:07'!
treeStringOn: stream indent: level
	"Print the tree
	structure of all possible scenarios."

	| i |
	i := self indent: level.
	stream nextPutAll: i, 'Wanted:'; cr.
	wantedReleases do: [:r |
		stream nextPutAll: i ,'  ' , r packageNameWithVersion;cr].
	stream nextPutAll: i, 'Tricky:'; cr.
	trickyReleases do: [:r |
		stream nextPutAll: i ,'  ' , r packageNameWithVersion;cr].
	stream cr.
	subAnalysises ifNotNil: [
		subAnalysises do: [:sub | sub treeStringOn: stream indent: level + 1]]! !

!SMDependencyAnalysis methodsFor: 'accessing' stamp: 'gk 7/28/2004 15:32'!
trickyReleases
	^trickyReleases
	! !

!SMDependencyAnalysis methodsFor: 'accessing' stamp: 'gk 7/27/2004 15:36'!
trivialToInstall
	^trivialToInstall
	! !

!SMDependencyAnalysis methodsFor: 'queries' stamp: 'gk 9/26/2004 23:47'!
untestedInstallPaths
	"We take the paths with conflicts and remove the older releases."

	^self installPathsWithConflicts collect: [:p |
		 self removeOlderReleasesIn: p] ! !

!SMDependencyAnalysis methodsFor: 'accessing' stamp: 'gk 7/27/2004 15:35'!
wantedReleases
	^wantedReleases
	! !

Object subclass: #SMDependencyEngine
	instanceVariableNames: 'map'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMDependencyEngine commentStamp: '<historical>' prior: 0!
A dependency engine is used to perform one or more installation, upgrade or uninstallation tasks.
After creation it is typically configured according to different strategies, policies etc based on the preferences of the user.
Then it is used to calculate what installations, uninstallations or upgrades are needed and in which order to reach certain stated goals, like installing a set of wanted packages or upgrading the installed packages.

The engine instantiates different SMInstallationTask subclasses depending on what it should calculate.

Todo: both the installation information for the image and the engine should probably be kept outside of the SMSqueakMap instance.
!

!SMDependencyEngine classSide methodsFor: 'instance creation' stamp: 'gk 7/29/2004 14:29'!
map: aMap

	^ self new map: aMap! !

!SMDependencyEngine methodsFor: 'tasks' stamp: 'gk 7/29/2004 15:59'!
installPackages: wantedPackages
	"Given a Set of wanted SMPackages, create an installation task to compute
	possible installation scenarios.
	Returns an SMInstallationTask which can be further configured
	and then be sent #calculate after which it can be queried for results."
	
	^SMPackageInstallationTask engine: self wantedPackages: wantedPackages! !

!SMDependencyEngine methodsFor: 'accessing' stamp: 'gk 7/29/2004 14:47'!
map
	^map! !

!SMDependencyEngine methodsFor: 'accessing' stamp: 'gk 5/5/2006 02:07'!
map: aMap
	map := aMap! !

Object subclass: #SMFileCache
	instanceVariableNames: 'map'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMFileCache commentStamp: 'gk 3/8/2004 20:10' prior: 0!
A repository for SMSqueakMap downloads. This behaves like a Set, responding to add: and include:, but also package contents may be forcibly refreshed with download:.

The SqueakMap determines what path the cache resides at. Within the cache, there is a 'packages' directory containing UUID-based directories for each package containing further directories for each release. A 'resources' directory stores UUID-based directories for each Resource, with the file stored within that by its original name. Because the cache follows a Set protocol, it can be automatically traversed within Smalltalk's collection protocol, avoiding manual hassles.!

!SMFileCache classSide methodsFor: 'instance creation' stamp: 'gk 1/23/2004 10:21'!
newFor: aMap
	"This is the default creation method, responsible for ensuring the
	paths and such exist, and filling in defaults."

	^self new forMap: aMap
! !

!SMFileCache methodsFor: 'services' stamp: 'gk 7/16/2004 11:04'!
add: aDownloadable 
	"Conditionally download the downloadable object into the cache.
	Return true on success, otherwise false."

	^(self includes: aDownloadable)
		ifTrue: [true]
		ifFalse: [self download: aDownloadable]! !

!SMFileCache methodsFor: 'private' stamp: 'gk 5/5/2006 02:07'!
cacheUrlFor: aDownloadable
	"Find a cache URL for this downloadable.
	Returns nil if no server is available.
	Could use #relativeUrl also."

	| server |
	server := aDownloadable map class findServer.
	server ifNil: [^ nil].
	^'http://', server, '/object/', aDownloadable id asString, '/cache'! !

!SMFileCache methodsFor: 'services' stamp: 'gk 4/19/2006 01:31'!
contents: anSMObject
	"Return contents of the file for the object
	or nil if not in cache."

	anSMObject isCached
		ifTrue: [^(anSMObject cacheDirectory readOnlyFileNamed: anSMObject downloadFileName) binary; contentsOfEntireFile]
		ifFalse: [^nil]
		! !

!SMFileCache methodsFor: 'accessing' stamp: 'gk 1/23/2004 10:26'!
directory
	^map packageCacheDirectory! !

!SMFileCache methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
directoryForPackage: aPackage
	"Returns the local path for storing the package cache's package file area.
	This also ensures that the path exists."

	| slash path dir |
	slash := FileDirectory slash.
	path := 'packages' , slash , aPackage id asString36 , slash.
	dir := FileDirectory default on: self directory fullName, slash, path.
	dir assureExistence.
	^dir! !

!SMFileCache methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
directoryForPackageRelease: aPackageRelease
	"Returns the local path for storing the package cache's version of a  
	package file. This also ensures that the path exists."

	| slash path dir |
	slash := FileDirectory slash.
	path := 'packages' , slash , aPackageRelease package id asString36 , slash , aPackageRelease automaticVersionString.
	dir := FileDirectory default on: self directory fullName, slash, path.
	dir assureExistence.
	^dir! !

!SMFileCache methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
directoryForResource: aResource
	"Returns the local path for storing the package cache's version of a  
	resource file. This also ensures that the path exists."

	| slash path dir |
	slash := FileDirectory slash.
	path := 'resources' , slash , aResource id asString36.
	dir := FileDirectory default on: self directory fullName, slash, path.
	dir assureExistence.
	^dir! !

!SMFileCache methodsFor: 'services' stamp: 'gk 5/5/2006 02:07'!
download: aDownloadable 
	"Download the file for this SMObject into the local file cache.
	If the file already exists, delete it.
	No unpacking or installation into the running image."

	| stream file fileName dir |
	[fileName := aDownloadable downloadFileName.
	fileName
		ifNil: [self inform: 'No download url, can not download.'.
			^ false].
	fileName isEmpty
		ifTrue: [self inform: 'Download url lacks filename, can not download.'.
			^ false].
	dir := aDownloadable cacheDirectory.
	[stream := self getStream: aDownloadable.
	stream ifNil: [^ false].
	(dir fileExists: fileName)
		ifTrue: [dir deleteFileNamed: fileName].
	file := dir newFileNamed: fileName.
	file binary; nextPutAll: stream contents]
		ensure: [file ifNotNil: [file close]]]
		on: Error
		do: [^ false].
	^ true! !

!SMFileCache methodsFor: 'initialize' stamp: 'stephaneducasse 2/4/2006 20:38'!
forMap: aMap
	"Initialize the ache, make sure the cache dir exists."

	map := aMap! !

!SMFileCache methodsFor: 'private' stamp: 'gk 5/5/2006 02:07'!
getStream: aDownloadable 
	"Get the stream, either from the original url
	or if that fails, from the server cache - unless
	this is the actual server of course. :)
	We also verify that the sha1sum is correct."

	| stream |
	[stream := aDownloadable downloadUrl asUrl retrieveContents contentStream binary.
	(aDownloadable correctSha1sum: stream contents)
		ifFalse: [self error: 'Incorrect SHA checksum of file from orginal URL']]
		on: Exception do: [:ex |
			Transcript show: 'Download from original url (', aDownloadable downloadUrl, ') failed with this exception: ', ex messageText;cr.
			SMUtilities isServer
				ifTrue: [^nil]
				ifFalse: [
					Transcript show: 'Trying server cache instead.'; cr.
					[stream := (self cacheUrlFor: aDownloadable) asUrl retrieveContents contentStream binary.
					(stream contents size = 21 and: [stream contents asString = 'SMFILEMISSINGONSERVER'])
						ifTrue: [self error: 'File missing in server cache'].
					(stream contents size = 24 and: [stream contents asString = 'SMRELEASENOTDOWNLOADABLE'])
						ifTrue: [self error: 'Release not downloadable'].
					(aDownloadable correctSha1sum: stream contents)
						ifFalse: [self error: 'Incorrect SHA checksum of file from server']]
							on: Exception do: [:ex2 | | msg |
								msg := 'Download from server cache of ', aDownloadable printName, ' failed with this exception: ', ex2 messageText.
								Transcript show: msg; cr.
								self error: msg]]].
	^ stream! !

!SMFileCache methodsFor: 'services' stamp: 'gk 7/16/2004 11:10'!
includes: anSMObject
	"Check if the cache holds the file for the object."
	
	^(anSMObject cacheDirectory)
		fileExists: anSMObject downloadFileName! !

!SMFileCache methodsFor: 'accessing' stamp: 'btr 5/27/2003 16:24'!
map
	^ map! !

Object subclass: #SMInstallationDeviation
	instanceVariableNames: 'brokenConfigurations selectedRelease'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMInstallationDeviation commentStamp: '<historical>' prior: 0!
An installation deviation is when the user decides to install or upgrade to a release that is newer than one or more used configurations specify.
This means that the other installed releases which configurations will be broken may not work correctly.
Instvar selectedRelease refers to the release selected to install, brokenConfigurations is a collection of all configurations that need another release of this package.!

!SMInstallationDeviation classSide methodsFor: 'instance creation' stamp: 'gk 10/13/2004 00:51'!
selectedRelease: release releases: releases
	^self new selectedRelease: release releases: releases! !

!SMInstallationDeviation methodsFor: 'accessing' stamp: 'gk 5/5/2006 02:07'!
otherReleases
	| package |
	package := selectedRelease package.
	^ brokenConfigurations collect: [:conf |
		conf releases detect: [:r | r package == package]]! !

!SMInstallationDeviation methodsFor: 'accessing' stamp: 'gk 10/4/2004 11:14'!
selectedRelease
	^ selectedRelease! !

!SMInstallationDeviation methodsFor: 'initialize-release' stamp: 'gk 5/5/2006 02:07'!
selectedRelease: aRelease releases: releases

	| p others otherRequired |
	selectedRelease := aRelease.
	p := selectedRelease package.
	brokenConfigurations := OrderedCollection new.
	others := releases copyWithout: aRelease.
	others := others select: [:r | r package ~= p].
	others do: [:rel |
		rel workingConfigurations do: [:conf |
			otherRequired := conf requiredReleases select: [:r | r package ~= p].
			((others includesAllOf: otherRequired) and:
				[(conf requiredReleases includes: selectedRelease) not])
					ifTrue: [brokenConfigurations add: conf]]]! !

Object subclass: #SMInstallationProposal
	instanceVariableNames: 'installList comment deviations task'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMInstallationProposal commentStamp: '<historical>' prior: 0!
This is primarily an ordered list of package release installations or upgrades to achieve a given installation task.
!

!SMInstallationProposal classSide methodsFor: 'instance creation' stamp: 'gk 10/1/2004 10:01'!
installList: anOrderedCollection
	^ self new installList: anOrderedCollection! !

!SMInstallationProposal methodsFor: 'initialize-release' stamp: 'gk 5/5/2006 02:07'!
calculateComment

	comment := ''! !

!SMInstallationProposal methodsFor: 'initialize-release' stamp: 'gk 5/5/2006 02:07'!
calculateDeviations
	"Calculate deviations. Currently we just pick the newest release."

	| conflicts newest |
	deviations := OrderedCollection new.
	conflicts := self collectConflictsIn: installList.
	conflicts keysAndValuesDo: [:package :releases |
		newest := releases first.
		releases do: [:r | (r newerThan: newest) ifTrue: [newest := r]].
		deviations add: (SMInstallationDeviation selectedRelease: newest releases: installList)]! !

!SMInstallationProposal methodsFor: 'initialize-release' stamp: 'gk 5/5/2006 02:07'!
collectConflictsIn: collectionOfReleases
	"Collect all conflicts where there are either
		- multiple releases of the same package and/or
		- another release of the same package already installed
	Return the conflicts as an IdentityDictionary with
	the package as key and the value being a Set of releases."

	| conflicts set |
	conflicts := IdentityDictionary new.
	collectionOfReleases do: [:r |
		set := conflicts at: r package ifAbsent: [
				conflicts at: r package put: OrderedCollection new].
		set add: r].
	"Add the installed releases too"
	conflicts keysAndValuesDo: [:key :value |
		key isInstalled ifTrue: [value add: key installedRelease]].
	"Prune release sets with only one member"
	^conflicts select: [:releaseSet | releaseSet size > 1]! !

!SMInstallationProposal methodsFor: 'initialize-release' stamp: 'gk 10/13/2004 01:26'!
hasDeviations
	^ deviations notEmpty! !

!SMInstallationProposal methodsFor: 'initialize-release' stamp: 'gk 5/5/2006 02:07'!
installList: anOrderedCollection

	installList := anOrderedCollection.
	self calculateDeviations.
	self calculateComment! !

Object subclass: #SMInstallationRegistry
	instanceVariableNames: 'installedPackages installCounter map'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMInstallationRegistry commentStamp: '<historical>' prior: 0!
A registry instance keeps track of installations into an image. Typically used by the SMSqueakMap instance when installing package releases.!

!SMInstallationRegistry classSide methodsFor: 'instance creation' stamp: 'gk 8/1/2004 17:42'!
map: aMap
	"Create a new registry and make it use the given map."

	^self new map: aMap! !

!SMInstallationRegistry methodsFor: 'queries' stamp: 'gk 8/1/2004 17:51'!
clearInstalledPackageWithId: aPackageId
	"Clear the fact that any release of this package is installed.
	Can be used even when the map isn't loaded."

	^installedPackages ifNotNil: [
		installedPackages removeKey: (UUID fromString: aPackageId) ifAbsent: [nil]]! !

!SMInstallationRegistry methodsFor: 'services' stamp: 'gk 5/5/2006 02:07'!
clearInstalledPackages
	"Simply clear the dictionary with information on installed packages.
	Might be good if things get corrupted etc. Also see
	SMSqueakMap class>>recreateInstalledPackagesFromChangeLog"

	installedPackages := nil.
	installCounter := 0! !

!SMInstallationRegistry methodsFor: 'private' stamp: 'gk 5/5/2006 02:07'!
countInstall
	"Increase the install counter."

	installCounter ifNil: [installCounter := 0].
	^installCounter := installCounter + 1
! !

!SMInstallationRegistry methodsFor: 'accessing' stamp: 'gk 5/5/2006 02:07'!
installCounter: anInteger
	"Set counter directly."

	installCounter := anInteger! !

!SMInstallationRegistry methodsFor: 'queries' stamp: 'gk 5/5/2006 02:07'!
installedPackages
	"Answer all packages that we know are installed.
	Lazily initialize. The Dictionary contains the installed packages
	using their UUIDs as keys and the version string as the value."

	| result p |
	result := OrderedCollection new.
	installedPackages ifNil: [^#()]
		ifNotNil: [installedPackages keys
					do: [:k |
						p := map object: k.
						p ifNotNil: [result add: p]]].
	^result! !

!SMInstallationRegistry methodsFor: 'accessing' stamp: 'gk 8/1/2004 17:44'!
installedPackagesDictionary
	"Access the dictionary directly. The UUID of the installed package is the key.
	The value is an OrderedCollection of Arrays.
	The arrays have the smartVersion of the package, the time of the
	installation in seconds and the sequence number (installCounter)."

	^installedPackages ifNil: [Dictionary new]! !

!SMInstallationRegistry methodsFor: 'accessing' stamp: 'gk 5/5/2006 02:07'!
installedPackagesDictionary: aDict
	"Set dictionary directly."

	installedPackages := aDict! !

!SMInstallationRegistry methodsFor: 'queries' stamp: 'gk 5/5/2006 02:07'!
installedReleaseOf: aPackage
	"If the package is installed, return the release.
	Otherwise return nil. SM2 stores the version as
	an Association to be able to distinguish it."

	| autoVersionOrOld |
	installedPackages ifNil: [^nil].
	autoVersionOrOld := (installedPackages at: aPackage id ifAbsent: [^nil]) last first.
	(autoVersionOrOld isKindOf: Association)
		ifTrue: [
			^aPackage releaseWithAutomaticVersion: autoVersionOrOld value]
		ifFalse: [
			^aPackage releaseWithVersion: autoVersionOrOld]! !

!SMInstallationRegistry methodsFor: 'queries' stamp: 'gk 5/5/2006 02:07'!
installedVersionOf: aPackage
	"If the package is installed, return the version as a String.
	If it is a package installed during SM1 it will return the manual version String,
	for SM2 it returns the automatic version as a String.
	If package is not installed - return nil. If you want it to work without the map loaded you
	should instead use #installedVersionOfPackageWithId:."

	| versionOrString |
	versionOrString := self installedVersionOfPackageWithId: aPackage id.
	versionOrString ifNil: [^nil].
	^versionOrString isString
		ifTrue: [versionOrString]
		ifFalse: [versionOrString versionString]! !

!SMInstallationRegistry methodsFor: 'queries' stamp: 'gk 5/5/2006 02:07'!
installedVersionOfPackageWithId: anId
	"If the package is installed, return the automatic version or version String.
	Otherwise return nil. This can be used without the map loaded."

	| autoVersionOrOld |
	installedPackages ifNil: [^nil].
	autoVersionOrOld := (installedPackages at: anId ifAbsent: [^nil]) last first.
	(autoVersionOrOld isKindOf: Association)
		ifTrue: [
			^autoVersionOrOld value]
		ifFalse: [
			^autoVersionOrOld]! !

!SMInstallationRegistry methodsFor: 'accessing' stamp: 'gk 5/5/2006 02:07'!
map: aMap

	map := aMap! !

!SMInstallationRegistry methodsFor: 'private' stamp: 'gk 5/5/2006 02:07'!
markInstalled: uuid version: version time: time counter: num
	"Private. Mark the installation. SM2 uses an Association
	to distinguish the automatic version from old versions."


	| installs |
	installedPackages ifNil: [installedPackages := Dictionary new].
	installs := installedPackages at: uuid
				ifAbsent: [installedPackages at: uuid put: OrderedCollection new].
	installs add:
		(Array with: 2->version
				with: time
				with: num)! !

!SMInstallationRegistry methodsFor: 'installation-changelog' stamp: 'gk 5/5/2006 02:07'!
noteInstalledPackage: uuidString version: version
	"Mark a specific version of a package as installed.
	This method is called when replaying a logged installation
	from before SqueakMap 1.07. Such logged installations lacked
	a timestamp and a count. We take the current time and a
	count starting from -10000 and upwards. This should keep
	the sorting order correct."

	"Find the lowest installed count."
	| lowest |
	lowest := 0.
	installedPackages ifNotNil: [
		installedPackages valuesDo: [:oc |
			oc do: [:array |
				array last < lowest ifTrue: [lowest := array last]]]]
		ifNil: [lowest := -10000].
	lowest negative ifFalse: [lowest := -10000].
	^self noteInstalledPackage: uuidString version: version
		atSeconds: Time totalSeconds number: lowest + 1! !

!SMInstallationRegistry methodsFor: 'installation-changelog' stamp: 'gk 5/5/2006 02:07'!
noteInstalledPackage: uuidString version: version atSeconds: time number: num
	"Mark a package as installed in the Dictionary.
	This method is called when replaying a logged installation.
	<time> is the point in time as totalSeconds of the installation.
	<num> is the installCount of the installation.
	This method is typically called from a doIt in the changelog
	in order to try to keep track of packages installed."

	num negative ifFalse: ["Not an emulated count from prior SM1.07"
		installCounter := num max: installCounter].
	self markInstalled: (UUID fromString: uuidString) version: version time: time counter: num! !

!SMInstallationRegistry methodsFor: 'installation-changelog' stamp: 'gk 5/5/2006 02:07'!
noteInstalledPackageWithId: uuidString autoVersion: version atSeconds: time number: num
	"Mark a package as installed in the Dictionary.
	This method is called when replaying a logged installation.
	<time> is the point in time as totalSeconds of the installation.
	<num> is the installCount of the installation.
	This method is typically called from a doIt in the changelog
	in order to try to keep track of packages installed."

	num negative ifFalse: ["Not an emulated count from prior SM1.07"
		installCounter := num max: installCounter].
	self markInstalled: (UUID fromString: uuidString) version: version time: time counter: num! !

!SMInstallationRegistry methodsFor: 'installation' stamp: 'gk 5/5/2006 02:07'!
noteInstalledPackageWithId: aPackageId autoVersion: aVersion name: aName
	"The package release was just successfully installed.
	Can be used to inform SM of an installation not been
	done using SM, even when the map isn't loaded.

	We record the fact in our Dictionary of installed packages
	and log a 'do it' to mark this in the changelog.
	The doit helps keeping track of the packages when
	recovering changes etc - not a perfect solution but should help.
	The map used is the default map.
	The id of the package is the key and the value is an OrderedCollection
	of Arrays with the release auto version, the point in time and the current installCounter."

	| time name id v |
	v := aVersion isString ifTrue: [aVersion asVersion] ifFalse: [aVersion].
	aName ifNil: [name := '<unknown package name>'] ifNotNil: [name := aName].
	id := UUID fromString: aPackageId.
	time := Time totalSeconds.
	self countInstall.
	self markInstalled: id version: v time: time counter: installCounter.
	(((Smalltalk classNamed: 'SmalltalkImage') ifNotNilDo: [:si | si current]) ifNil: [Smalltalk])
		logChange: '"Installed ', name, ' auto version ', v versionString, '".
(Smalltalk at: #SMSqueakMap ifAbsent: []) ifNotNil:[
	SMSqueakMap noteInstalledPackageWithId: ', id asString storeString, ' autoVersion: ', v storeString, ' atSeconds: ', time asString, ' number: ', installCounter asString, ']'! !

!SMInstallationRegistry methodsFor: 'installation' stamp: 'gk 5/5/2006 02:07'!
noteUninstalledPackageWithId: aPackageId autoVersion: aVersion name: aName
	"The package release was just successfully uninstalled.
	Can be used to inform SM of an uninstallation not been
	done using SM, even when the map isn't loaded.

	We record the fact in our Dictionary of installed packages
	and log a 'do it' to mark this in the changelog.
	The doit helps keeping track of the packages when
	recovering changes etc - not a perfect solution but should help.
	The map used is the default map.
	The id of the package is the key and the value is an OrderedCollection
	of Arrays with the release auto version, the point in time and the current installCounter."

	| time name id v |
	v := aVersion isString ifTrue: [aVersion asVersion] ifFalse: [aVersion].
	aName ifNil: [name := '<unknown package name>'] ifNotNil: [name := aName].
	id := UUID fromString: aPackageId.
	time := Time totalSeconds.
	self countInstall. "Used for both installs and uninstalls"
	self clearInstalled: id version: v time: time counter: installCounter.
	(((Smalltalk classNamed: 'SmalltalkImage') ifNotNilDo: [:si | si current]) ifNil: [Smalltalk])
		logChange: '"Uninstalled ', name, ' auto version ', v versionString, '".
(Smalltalk at: #SMSqueakMap ifAbsent: []) ifNotNil:[
	SMSqueakMap noteUninstalledPackageWithId: ', id asString storeString, ' autoVersion: ', v storeString, ' atSeconds: ', time asString, ' number: ', installCounter asString, ']'! !

Object subclass: #SMInstallationTask
	instanceVariableNames: 'map engine'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMInstallationTask commentStamp: '<historical>' prior: 0!
An SMInstallationProposal is effectively a list of releases to install or upgrade to in a specific order to achieve an SMInstallationTask.
The task can be either an upgrade or a new installation - the proposal still involved an ordered list of installations or upgrades.
!

!SMInstallationTask methodsFor: 'calculation' stamp: 'gk 9/20/2004 20:49'!
calculate
	"Calculate how the task should be performed.
	After calculation the task can be inspected and presented
	to the user for confirmation. Actually performing the task
	is done by #execute."
	
	self subclassResponsibility ! !

!SMInstallationTask methodsFor: 'accessing' stamp: 'gk 5/5/2006 02:07'!
engine: anEngine
	engine := anEngine.
	map := engine map! !

!SMInstallationTask methodsFor: 'calculation' stamp: 'gk 9/20/2004 20:49'!
execute
	"Actually perform the task."

	self subclassResponsibility ! !

!SMInstallationTask methodsFor: 'accessing' stamp: 'gk 7/29/2004 14:48'!
map
	^map! !

SMInstallationTask subclass: #SMPackageInstallationTask
	instanceVariableNames: 'wantedPackages wantedReleases analysis'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMPackageInstallationTask commentStamp: '<historical>' prior: 0!
A package installation task is to install one or more given SMPackages (not specified releases) into the image.

First it tries to calculate the ideal releases of the given packages that it will try to install given the policy and preferences set by the user. Then it runs an analysis to find how to install those wanted releases. This typically results in zero, one or more possible scenarios.
!

!SMPackageInstallationTask classSide methodsFor: 'instance creation' stamp: 'gk 7/29/2004 12:37'!
engine: engine wantedPackages: wantedPackages
	^self new engine: engine; wantedPackages: wantedPackages! !

!SMPackageInstallationTask methodsFor: 'queries' stamp: 'gk 9/22/2004 22:47'!
allInstallPaths
	"Return all different ways to install - the ones requested plus all dependencies.
	This includes ways where different releases of the same package are combined."

	^analysis allInstallPaths! !

!SMPackageInstallationTask methodsFor: 'accessing' stamp: 'gk 9/20/2004 20:50'!
analysis
	"Return the analysis of the task."

	^analysis! !

!SMPackageInstallationTask methodsFor: 'calculation' stamp: 'gk 5/5/2006 02:07'!
calculate
	"First calculate the wanted releases. Then perform a dependency analysis.
	We return the most basic result of the analysis - does there exist at least one
	working installation scenario without tweaks?"

	self calculateWantedReleases.
	analysis := SMDependencyAnalysis task: self.
	analysis installPackageReleases: wantedReleases.
	^analysis success! !

!SMPackageInstallationTask methodsFor: 'private' stamp: 'gk 5/5/2006 02:07'!
calculateWantedReleases
	"The user gave us wanted packages.
	We need to figure out which actual releases of those
	we should try to install."

	| rel |
	wantedReleases := Set new.
	wantedPackages do: [:p | rel := self idealReleaseFor: p.
		rel ifNotNil: [wantedReleases add: rel]]! !

!SMPackageInstallationTask methodsFor: 'private' stamp: 'gk 9/20/2004 20:51'!
idealReleaseFor: aPackage
	"Return the most suitable release to install for <aPackage>."

	^ aPackage lastPublishedReleaseForCurrentSystemVersion! !

!SMPackageInstallationTask methodsFor: 'queries' stamp: 'gk 10/1/2004 10:06'!
proposals
	"Return all different possible proposals to install
	sorted with the best proposal first."

	^analysis allNormalizedInstallPaths collect: [:path | SMInstallationProposal installList: path]! !

!SMPackageInstallationTask methodsFor: 'accessing' stamp: 'gk 5/5/2006 02:07'!
wantedPackages: packages

	wantedPackages := packages! !

SMPackageInstallationTask subclass: #SMPackageUpgradeTask
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMPackageUpgradeTask commentStamp: '<historical>' prior: 0!
A package upgrade task is to upgrade one or more given SMPackages (not specified releases) in the image to newer releases.

First it tries to calculate the newest available releases of the given packages that it will try to upgrade given the policy and preferences set by the user. Then it runs an analysis to find how to upgrade to those wanted releases. This typically results in zero, one or more possible scenarios.
!

Object subclass: #SMInstaller
	instanceVariableNames: 'packageRelease'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-installer'!

!SMInstaller commentStamp: '<historical>' prior: 0!
An installer takes care of installing SqueakMap packages represented by SMCards.
Deciding if a package is installable and instantiating the correct installer class is done on the class side, see implementors of #canInstall:. Two methods need to be implemented by subclasses - download and install. Upgrade can also be specialized by implementing #upgrade, otherwise it will default to #install.!

!SMInstaller classSide methodsFor: 'changeset utilities' stamp: 'nk 11/30/2002 17:18'!
basicNewChangeSet: newName 
	"This method copied here to ensure SqueakMap is independent of 
	ChangeSorter. "
	Smalltalk
		at: #ChangeSorter
		ifPresentAndInMemory: [:cs | ^ cs basicNewChangeSet: newName].
	(self changeSetNamed: newName)
		ifNotNil: [self error: 'The name ' , newName , ' is already used'].
	^ ChangeSet basicNewNamed: newName! !

!SMInstaller classSide methodsFor: 'testing' stamp: 'gk 11/16/2003 23:49'!
canInstall: aPackageRelease
	"Nope, I am an abstract class and can not install anything.
	But my subclasses should reimplement this."

	^ false! !

!SMInstaller classSide methodsFor: 'changeset utilities' stamp: 'gh 10/31/2002 10:11'!
changeSetNamed: newName
	"This method copied here to ensure SqueakMap is independent of ChangeSorter."

	Smalltalk at: #ChangeSorter ifPresentAndInMemory: [ :cs | ^cs changeSetNamed: newName ].
	^ChangeSet allInstances detect: [ :cs | cs name = newName ] ifNone: [ nil ].! !

!SMInstaller classSide methodsFor: 'instance creation' stamp: 'gk 11/16/2003 23:33'!
classForPackageRelease: aPackageRelease
	"Decide which subclass to instantiate. 
	We detect and return the first subclass
	that wants to handle the release going
	recursively leaf first so that subclasses gets
	first chance if several classes compete over
	the same packages, like for example SMDVSInstaller
	that also uses the .st file extension."

	self subclasses do: [:ea |
		(ea classForPackageRelease: aPackageRelease)
			ifNotNilDo: [:class | ^ class]].
	^(self canInstall: aPackageRelease)
		ifTrue: [self]! !

!SMInstaller classSide methodsFor: 'deprecated' stamp: 'gk 5/5/2006 02:07'!
forPackageRelease: aPackageRelease
	"Instantiate the first class suitable to install the package release.
	If no installer class is found we raise an Error."

	| class |
	aPackageRelease ifNil: [self error: 'No package release specified to find installer for.'].
	class := self classForPackageRelease: aPackageRelease.
	^class
		ifNil: [self error: 'No installer found for package ', aPackageRelease name, '.']
		ifNotNil: [class new packageRelease: aPackageRelease]! !

!SMInstaller classSide methodsFor: 'testing' stamp: 'gk 11/16/2003 23:50'!
isInstallable: aPackageRelease
	"Detect if any subclass can handle the package release."

	aPackageRelease ifNil: [^false].
	^(self classForPackageRelease: aPackageRelease) notNil! !

!SMInstaller classSide methodsFor: 'testing' stamp: 'gk 11/16/2003 23:49'!
isUpgradeable: aPackageRelease
	"Detect if any subclass can handle the release.
	Currently we assume that upgrade is the same as install."

	^self isInstallable: aPackageRelease! !

!SMInstaller methodsFor: 'services' stamp: 'gk 11/16/2003 21:56'!
download
	"This service should bring the package release to
	the client and also unpack it on disk if needed.
	It will not install it into the running image though.
	Raises errors if operation does not succeed."

	self subclassResponsibility ! !

!SMInstaller methodsFor: 'services' stamp: 'gk 11/16/2003 21:57'!
install
	"This service should bring the package release to the client,
	unpack it if necessary and install it into the image.
	The package release should be notified of the installation using
	'packageRelease noteInstalled'."

	self subclassResponsibility ! !

!SMInstaller methodsFor: 'testing' stamp: 'gk 11/23/2005 00:06'!
isCached
	"Check if it is in the cache."

	^packageRelease isCached! !

!SMInstaller methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
packageRelease: aPackageRelease
	packageRelease := aPackageRelease! !

!SMInstaller methodsFor: 'private' stamp: 'gk 7/13/2004 02:43'!
silent
	"Can we ask questions?"
	
	^packageRelease ifNotNil: [packageRelease map silent] ifNil: [false]! !

!SMInstaller methodsFor: 'services' stamp: 'gk 7/14/2004 15:38'!
upgrade
	"This service performs an upgrade to the selected release.
	Currently it just defaults to the same operation as an install -
	which is handled fine by Monticello, but not necessarily for
	other formats."

	^self install! !

SMInstaller subclass: #SMSimpleInstaller
	instanceVariableNames: 'fileName dir unpackedFileName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-installer'!

!SMSimpleInstaller commentStamp: '<historical>' prior: 0!
This is a base class that you can subclass if your package format can be downloaded using
a single file url and possibly also be decompressed using gzip.!

SMSimpleInstaller subclass: #SMDefaultInstaller
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-installer'!

!SMDefaultInstaller commentStamp: '<historical>' prior: 0!
An installer takes care of installing SqueakMap packages represented by SMCards.
This installer handles packages that consist of classical fileins (single changesets and .st-files) and optional gzip-decompression of those. Deciding if a package is installable and instantiating the correct installer class is done on the class side in SMInstaller, to see how this installer gets chosen - see SMDefaultInstaller class>>canInstall:.

!

SMDefaultInstaller subclass: #SMDVSInstaller
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-installer'!

!SMDVSInstaller commentStamp: '<historical>' prior: 0!
This is an installer class for DVS packages. It handles packages categorized with package
format as DVS and with a download filename with extensions .st or .st.gz.

This class can function without DVS installed, needed classes are looked up dynamically.!

!SMDVSInstaller classSide methodsFor: 'testing' stamp: 'stephaneducasse 2/4/2006 20:38'!
canInstall: aPackage
	"Can I install this? First we check if class StreamPackageLoader
	is available, otherwise DVS isn't installed.
	Then we check if the package is categorized with package
	format DVS - currently we have hardcoded the id of that category."

	| fileName |
	Smalltalk at: #StreamPackageLoader ifPresentAndInMemory: [ :loader |
		fileName := aPackage downloadFileName.
		fileName ifNil: [^false].
		fileName := fileName asLowercase.
		^((fileName endsWith: '.st') or: [fileName endsWith: '.st.gz'])
			and: [aPackage categories includes: "The DVS format category"
					(SMSqueakMap default
						categoryWithId: 'b02f51f4-25b4-4117-9b65-f346215a8e41')]].
	^false! !

!SMDVSInstaller classSide methodsFor: 'loading' stamp: 'gk 10/1/2003 13:51'!
loadDVS
	"Load the DVS package from SqueakMap."

	SMSqueakMap default installPackageWithId: '100d59d0-bf81-4e74-a4fe-5a2fd0c6b4ec'! !

!SMDVSInstaller methodsFor: 'services' stamp: 'stephaneducasse 2/4/2006 20:38'!
install
	"Install using DVS."

	| imagePackageLoader streamPackageLoader packageInfo packageManager baseName current new manager |
	self cache; unpack.
	imagePackageLoader := Smalltalk at: #ImagePackageLoader ifAbsent: [].
	streamPackageLoader := Smalltalk at: #StreamPackageLoader ifAbsent: [].
	packageInfo := Smalltalk at: #PackageInfo ifAbsent: [].
	packageManager := Smalltalk at: #FilePackageManager ifAbsent: [].

	({ imagePackageLoader. streamPackageLoader. packageInfo. packageManager } includes: nil)
		ifTrue: [ (self confirm: ('DVS support is not loaded, but would be helpful in loading ', unpackedFileName, '.
It isn''t necessary, but if you intend to use DVS later it would be a good idea to load it now.
Load it from SqueakMap?'))
			ifTrue: [ self class loadDVS. ^self install ]
			ifFalse: [ ^self fileIn ]].

	baseName := packageRelease name.
	dir rename: unpackedFileName toBe: (baseName, '.st').
	unpackedFileName := baseName, '.st'.

	(manager := packageManager allManagers detect: [ :pm | pm packageName = baseName ] ifNone: [])
		ifNotNil: [
			current := imagePackageLoader new package: (packageInfo named: baseName).
			new := streamPackageLoader new stream: (dir readOnlyFileNamed: unpackedFileName).
			(new changesFromBase: current) fileIn ]
		ifNil: [
			self fileIn.
			manager := packageManager named: baseName. ].

	manager directory: dir.
	packageManager changed: #allManagers.
	packageRelease noteInstalled! !

!SMDefaultInstaller classSide methodsFor: 'testing' stamp: 'gk 3/31/2006 09:59'!
canInstall: aPackage
	"Answer if this class can install/upgrade the package.
	This installer handles .st, .cs, .mst, .mcs (Squeak 3.9+)
	with or without .gz suffix."

	| fileName |
	fileName := aPackage downloadFileName.
	fileName ifNil: [^false].
	fileName := fileName asLowercase.
	^self sourceFileSuffixes anySatisfy: [:each | 
			(fileName endsWith: (FileDirectory dot, each)) or: [
				fileName endsWith: (FileDirectory dot, each, '.gz')]]! !

!SMDefaultInstaller classSide methodsFor: 'private' stamp: 'gk 3/31/2006 10:05'!
multiSuffixes
	"Unfortunately we can not tell which suffixes use multibyte encoding.
	So we guess that they begin with $m."

	^self sourceFileSuffixes select: [:suff | suff first = $m]! !

!SMDefaultInstaller classSide methodsFor: 'private' stamp: 'gk 3/31/2006 10:06'!
nonMultiSuffixes
	"Unfortunately we can not tell which suffixes use multibyte encoding.
	So we guess that they begin with $m."

	^self sourceFileSuffixes reject: [:suff | suff first = $m]! !

!SMDefaultInstaller classSide methodsFor: 'private' stamp: 'gk 3/31/2006 09:59'!
sourceFileSuffixes
	"Trying to play nice with all Squeak versions."

	^(FileStream respondsTo: #sourceFileSuffixes)
			ifTrue: [FileStream sourceFileSuffixes]
			ifFalse: [#(cs st)].! !

!SMDefaultInstaller methodsFor: 'private' stamp: 'gk 4/18/2006 23:31'!
fileIn
	"Installing in the standard installer is simply filing in.
	Both .st and .cs files will file into a ChangeSet of their own.
	We let the user confirm filing into an existing ChangeSet
	or specify another ChangeSet name if
	the name derived from the filename already exists."
	
	| fileStream |
	(self class nonMultiSuffixes anySatisfy: [:each | unpackedFileName endsWith: (FileDirectory dot, each)])
		ifTrue:[
			fileStream := dir readOnlyFileNamed: unpackedFileName.
			(fileStream respondsTo: #setConverterCode) ifTrue: [fileStream setConverterForCode].
			self fileIntoChangeSetNamed: (fileStream localName sansPeriodSuffix) fromStream: fileStream.
			^self].
	(self class multiSuffixes anySatisfy: [:each | unpackedFileName endsWith: (FileDirectory dot, each)])
		ifTrue:[
			fileStream := dir readOnlyFileNamed: unpackedFileName.
			"Only images with converters should have multi suffixes"
			fileStream converter: (Smalltalk at: #UTF8TextConverter) new.
			self fileIntoChangeSetNamed: (fileStream localName sansPeriodSuffix) fromStream: fileStream.
			^self].
	self error: 'Filename should end with a proper extension'.
! !

!SMDefaultInstaller methodsFor: 'services' stamp: 'gk 11/16/2003 20:52'!
install
	"This service should bring the package to the client,
	unpack it if necessary and install it into the image.
	The package is notified of the installation."

	self cache; unpack; fileIn.
	packageRelease noteInstalled
! !

SMSimpleInstaller subclass: #SMLanguageInstaller
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-installer'!

!SMLanguageInstaller classSide methodsFor: 'testing' stamp: 'gk 3/31/2006 00:22'!
canInstall: aPackage
	"Answer if this class can install the package.
	We handle .translation files optionally compressed."

	| fileName |
	((Smalltalk includesKey: #Language)
		or: [Smalltalk includesKey: #NaturalLanguageTranslator]) ifFalse: [^false].
	fileName := aPackage downloadFileName.
	fileName ifNil: [^false].
	fileName := fileName asLowercase.
	^(fileName endsWith: '.translation') or: [
		(fileName endsWith: '.tra') or: [
			(fileName endsWith: '.tra.gz') or: [
				fileName endsWith: '.translation.gz']]]! !

!SMLanguageInstaller methodsFor: 'services' stamp: 'gk 3/31/2006 00:23'!
install
	"This service should bring the package to the client, 
	unpack it if necessary and install it into the image. 
	The package is notified of the installation."

	| translator |
	self cache; unpack.
	translator := Smalltalk at: #Language ifAbsent: [Smalltalk at: #NaturalLanguageTranslator].
	[translator mergeTranslationFileNamed: unpackedFileName]
			ensure: [packageRelease noteInstalled]! !

SMSimpleInstaller subclass: #SMMcInstaller
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-installer'!

!SMMcInstaller commentStamp: 'gk 10/8/2003 14:28' prior: 0!
I am a SMInstaller that knows how to install .mcz (Monticello) files. If Monticello is installed I use that (MCMczReader), otherwise I file in the code more simply using the package MCInstaller (MczInstaller).!

!SMMcInstaller classSide methodsFor: 'testing' stamp: 'stephaneducasse 2/4/2006 20:38'!
canInstall: aPackage
	"Is this a Monticello package and do I have MCInstaller
	or Monticello available?"

	| fileName |
	((Smalltalk includesKey: #MCMczReader) or: [
		 Smalltalk includesKey: #MczInstaller])
			ifTrue: [
				fileName := aPackage downloadFileName.
				fileName ifNil: [^false].
				^ 'mcz' = (FileDirectory extensionFor: fileName) asLowercase].
	^false! !

!SMMcInstaller methodsFor: 'private' stamp: 'stephaneducasse 2/4/2006 20:38'!
fileIn
	| extension |
	extension := (FileDirectory extensionFor: fileName) asLowercase.
	extension = 'mcz'
		ifTrue: [self installMcz]
		ifFalse: [self error: 'Cannot install file of type .', extension]! !

!SMMcInstaller methodsFor: 'private' stamp: 'ab 8/8/2003 18:33'!
fullFileName 
	^ dir fullNameFor: fileName! !

!SMMcInstaller methodsFor: 'services' stamp: 'gk 11/16/2003 21:55'!
install
	"This service should bring the package to the client,
	unpack it if necessary and install it into the image.
	The package is notified of the installation."

	self cache; fileIn.
	packageRelease noteInstalled! !

!SMMcInstaller methodsFor: 'private' stamp: 'gk 7/13/2004 02:44'!
installMcz
	"Install the package, we already know that either MCInstaller or Monticello is available."

	| installer monticello |
	installer := MczInstaller.
	(Smalltalk hasClassNamed: #MCMczReader) ifFalse: [
		packageRelease package isInstalled ifTrue: [
			(self silent ifFalse: [
				(self confirm:
'A release of package ''', packageRelease package name, ''' is already installed.
You only have MCInstaller and not Monticello
installed and MCInstaller can not properly upgrade packages.
Do you wish to install Monticello first and then proceed?
If you answer no MCInstaller will be used - but at your own risk.
Cancel cancels the installation.' orCancel: [self error: 'Installation cancelled.'])]
			ifTrue: [false])
				ifTrue: [
					monticello := packageRelease map packageWithName: 'Monticello'.
					monticello lastPublishedRelease
						ifNotNil: [monticello lastPublishedRelease install]
						ifNil: [monticello lastRelease install].
					installer := (Smalltalk at: #MCMczReader)]]
	] ifTrue: [installer := (Smalltalk at: #MCMczReader)].
	installer loadVersionFile: self fullFileName! !

SMSimpleInstaller subclass: #SMProjectInstaller
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-installer'!

!SMProjectInstaller commentStamp: '<historical>' prior: 0!
I am a SMInstaller that knows how to install .pr (Project) files.!

!SMProjectInstaller classSide methodsFor: 'testing' stamp: 'stephaneducasse 2/4/2006 20:38'!
canInstall: aPackage
	"Answer if this class can install the package.
	We handle .pr files (upper and lowercase)"

	| fileName |
	fileName := aPackage downloadFileName.
	fileName ifNil: [^false].
	^'pr' = (FileDirectory extensionFor: fileName) asLowercase! !

!SMProjectInstaller methodsFor: 'services' stamp: 'gk 11/16/2003 21:55'!
install
	"This service should bring the package to the client, 
	unpack it if necessary and install it into the image. 
	The package is notified of the installation."

	Project canWeLoadAProjectNow ifFalse: [self error: 'Can not load Project now, probably because not in Morphic.'].
	self cache.
	[[ ProjectLoading openFromDirectory: dir andFileName: fileName ]
		on: ProgressTargetRequestNotification do: [ :ex | ex resume ]]
			ensure: [packageRelease noteInstalled]! !

SMSimpleInstaller subclass: #SMSARInstaller
	instanceVariableNames: 'zip'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-installer'!

!SMSARInstaller commentStamp: '<historical>' prior: 0!
I am a SqueakMap installer that knows how to deal with Zip format change-set archives.
I recognize them by the file extension ".sar" (Squeak Archive).

These have a couple of members with special names:

install/preamble
install/postscript

These are loaded in order. Either or both can further load other members using fileInMemberNamed:.

Inside a postscript or preamble, the pseudo-variable "self" is set to an instance of SARInstaller; you can then get to its ZipArchive using the method "zip". Or you can call its methods for filing in change sets, extracting files, etc.

You can test this loading with:
(SMSARInstaller new) directory: FileDirectory default; fileName: 'test.sar'; fileIn.

See ChangeSet>>fileOutAsZipNamed: for one way to make these files. Here is another way of creating a multi change set archive installable by SqueakMap:

"The following doit will create a .sar file with HVs preamble and postscript as
separate entries and the included changesets included as normal.
Given a preamble as described below this will autoinstall in SqueakMap."
(ChangeSorter changeSetNamed: 'HV')
	fileOutAsZipNamed: 'httpview-021023.sar'
	including: {
		ChangeSorter changeSetNamed: 'HVFixes'.
		ChangeSorter changeSetNamed: 'kom412'}

Preamble in changeset HV that will install the changesets:

"Change Set:		HV
Date:			23 October 2002
Author:			Gran Hultgren

This is my latest developer code drop of HttpView packaged as a Squeak selfextracting archive (courtesy Ned Konz)."

"Standard SqueakMap installing code follows:"
(self isKindOf: SARInstaller) ifTrue:[
	self fileInMemberNamed: 'HVFixes'.
	self fileInMemberNamed: 'kom412'.
	self fileInMemberNamed: 'HV'
]

!

!SMSARInstaller classSide methodsFor: 'testing' stamp: 'stephaneducasse 2/4/2006 20:38'!
canInstall: aPackage
	"Answer if this class can install the package.
	We handle it if the filename has the extension
	.sar (upper and lowercase) and SARInstaller is
	present in the image to handle the install."

	| fileName |
	fileName := aPackage downloadFileName.
	fileName ifNil: [^false].
	Smalltalk at: #SARInstaller ifPresentAndInMemory: [ :installer |
			^'sar' = (FileDirectory extensionFor: fileName) asLowercase].
	^false! !

!SMSARInstaller methodsFor: 'private' stamp: 'gh 10/31/2002 11:19'!
fileIn

	Smalltalk at: #SARInstaller ifPresentAndInMemory: [:installer |
		(installer directory: dir fileName: fileName) fileIn. ^self].
	self error: 'SAR support not installed in image, can not install.'! !

!SMSARInstaller methodsFor: 'services' stamp: 'gk 11/16/2003 21:55'!
install
	"This service should bring the package to the client,
	unpack it if necessary and install it into the image.
	The package is notified of the installation."

	self cache; fileIn.
	packageRelease noteInstalled! !

!SMSimpleInstaller classSide methodsFor: 'testing' stamp: 'gh 10/22/2002 11:55'!
canInstall: aPackage
	"Answer if this class can install the package.
	This class is abstract so we return false."

	^false! !

!SMSimpleInstaller methodsFor: 'services' stamp: 'stephaneducasse 2/4/2006 20:38'!
cache
	"Download object into cache if needed.
	Set the directory and fileName for subsequent unpacking and install."

	packageRelease ensureInCache ifTrue: [
		fileName := packageRelease downloadFileName.
		dir := packageRelease cacheDirectory]! !

!SMSimpleInstaller methodsFor: 'accessing' stamp: 'gh 10/23/2002 10:54'!
directory
	^dir! !

!SMSimpleInstaller methodsFor: 'services' stamp: 'stephaneducasse 2/4/2006 20:38'!
download
	"This service downloads the last release of the package
	even if it is in the cache already."

	packageRelease download ifTrue: [
		fileName := packageRelease downloadFileName.
		dir := packageRelease cacheDirectory]! !

!SMSimpleInstaller methodsFor: 'services' stamp: 'rbb 3/1/2005 11:12'!
fileIntoChangeSetNamed: aString fromStream: stream
	"We let the user confirm filing into an existing ChangeSet
	or specify another ChangeSet name if
	the name derived from the filename already exists."
	
	| changeSet newName oldChanges global |
	newName := aString.
	changeSet := SMInstaller changeSetNamed: newName.
	changeSet ifNotNil: [
		newName := self silent ifNil: [UIManager default
									request: 'ChangeSet already present, just confirm to overwrite or enter a new name:' 
									initialAnswer: newName]
						ifNotNil: [newName].
		newName isEmpty ifTrue:[self error: 'Cancelled by user'].
		changeSet := SMInstaller changeSetNamed: newName].
		changeSet ifNil:[changeSet := SMInstaller basicNewChangeSet: newName].
		changeSet ifNil:[self error: 'User did not specify a valid ChangeSet name'].
		oldChanges := (SystemVersion current highestUpdate < 5302)
						ifFalse: [global := ChangeSet. ChangeSet current]
						ifTrue: [global := Smalltalk. Smalltalk changes].
 		[global newChanges: changeSet.
		stream fileInAnnouncing: 'Loading ', newName, ' into change set ''', newName, ''''.
		stream close] ensure: [global newChanges: oldChanges]! !

!SMSimpleInstaller methodsFor: 'accessing' stamp: 'gh 10/21/2002 14:39'!
fileName
	^fileName! !

!SMSimpleInstaller methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
fileName: aFileName
	fileName := aFileName! !

!SMSimpleInstaller methodsFor: 'accessing' stamp: 'nk 2/22/2004 13:12'!
fullFileName 
	^ self directory fullNameFor: self fileName! !

!SMSimpleInstaller methodsFor: 'services' stamp: 'gk 4/18/2006 22:59'!
unpack
	"This basic installer simply checks the file extension of
	the downloaded file to choose suitable method for unpacking.
	Currently it only supports .gz decompression.
	If a file exists with the same name it is first deleted.
	The unpacked filename is set on succesfull decompression or
	if the file was not recognized as a compressed file."

	| unzipped zipped buffer |
	(fileName endsWith: '.gz')
		ifTrue:[
			unpackedFileName := fileName copyUpToLast: FileDirectory extensionDelimiter.
			(dir fileExists: unpackedFileName) ifTrue:[ dir deleteFileNamed: unpackedFileName ].
			unzipped := dir newFileNamed: unpackedFileName.
			unzipped binary.
			zipped := GZipReadStream on: ((dir readOnlyFileNamed: fileName) binary; yourself).
			buffer := ByteArray new: 50000.
			'Extracting ' , fileName
				displayProgressAt: Sensor cursorPoint
				from: 0
				to: zipped sourceStream size
				during: [:bar | 
					[zipped atEnd]
						whileFalse: 
							[bar value: zipped sourceStream position.
							unzipped nextPutAll: (zipped nextInto: buffer)].
					zipped close.
					unzipped close]]
		ifFalse:[unpackedFileName := fileName]! !

!SMSimpleInstaller methodsFor: 'accessing' stamp: 'gh 10/21/2002 14:39'!
unpackedFileName
	^unpackedFileName! !

Object subclass: #SMObject
	instanceVariableNames: 'id map created updated name summary url'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMObject commentStamp: 'gk 9/23/2003 20:26' prior: 0!
SMObject is the abstract superclass for all objects living in an SMSqueakMap.

It has a unique UUID and a reference to the owning SMSqueakMap.
It has timestamps to record the birthtime and the last modification.
It has basic attributes like name, oneline summary and url.





!

SMObject subclass: #SMCategorizableObject
	instanceVariableNames: 'categories resources'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMCategorizableObject commentStamp: '<historical>' prior: 0!
A categorizable object can be associated with one or more SMCategories.
The relation between the categories and the SMCategorizableObject is bidirectional.
A categorizable object can also have attached resources, see SMResource.

The categories are used to classify the categorizable object for different purposes.
Package and package releases are classified in different ways, but so can resources and accounts be.
!

!SMCategorizableObject methodsFor: 'private' stamp: 'stephaneducasse 2/4/2006 20:38'!
addCategory: aCategory
	"Add <aCategory> to me. If I already have it do nothing."

	categories ifNil: [categories := OrderedCollection new].
	(categories includes: aCategory) ifFalse:[
		aCategory addObject: self.
		categories add: aCategory].
	^aCategory! !

!SMCategorizableObject methodsFor: 'resources' stamp: 'gk 5/5/2006 02:07'!
addResource: aResource
	"Lazily initialize the resources collection."
	
	resources ifNil: [resources := OrderedCollection new].
	aResource object: self.
	^resources add: aResource! !

!SMCategorizableObject methodsFor: 'accessing' stamp: 'gk 9/23/2003 20:44'!
categories
	"Lazily initialized."

	^categories ifNil: [OrderedCollection new]! !

!SMCategorizableObject methodsFor: 'categories' stamp: 'gk 9/23/2003 20:45'!
categoriesDo: aBlock
	"Evaluate aBlock for each of the categories."

	categories ifNil: [^self].
	categories do: aBlock! !

!SMCategorizableObject methodsFor: 'categories' stamp: 'gk 11/24/2005 11:18'!
categoryForParent: aCategory
	"Answer one of my categories with parent <aCategory>, if I have it."

	categories ifNil: [^nil].
	^categories detect: [:cat | cat parent = aCategory ] ifNone: [nil]! !

!SMCategorizableObject methodsFor: 'private' stamp: 'gk 8/8/2003 02:35'!
delete
	"Delete me. Disconnect me from my categories."

	super delete.
	self removeFromCategories! !

!SMCategorizableObject methodsFor: 'printing' stamp: 'dew 3/17/2004 16:28'!
describeCategoriesOn: aStream indent: tabs 
	"Show a full listing of categories and their dscription on aStream, indented by the given number of tabs."

	categories isEmptyOrNil
		ifFalse: [aStream cr;
				withAttribute: TextEmphasis bold
				do: [aStream nextPutAll: 'Categories: ']; cr.
			(self categories asSortedCollection: [:a :b | a path < b path])
				do: [:c | 
					aStream tab: tabs.
					c
						parentsDo: [:p | aStream nextPutAll: p name;
								 nextPut: $/].
					aStream nextPutAll: c name;
						 nextPutAll: ' - ';
						
						withAttributes: {TextEmphasis italic. TextIndent tabs: tabs + 1 }
						do: [aStream nextPutAll: c summary];
						 cr]]! !

!SMCategorizableObject methodsFor: 'resources' stamp: 'gk 7/27/2004 18:12'!
embeddedResources
	"Return all embedded resources."
	
	^resources ifNil: [#()]
		ifNotNil: [resources select: [:r | r isEmbedded ]]
	! !

!SMCategorizableObject methodsFor: 'categories' stamp: 'gh 11/27/2002 12:35'!
hasCategory: aCategory
	"Answer true if I am in it."

	^categories notNil and: [categories includes: aCategory]! !

!SMCategorizableObject methodsFor: 'categories' stamp: 'gk 7/9/2004 02:57'!
hasCategoryOrSubCategoryOf: aCategory
	"Answer true if I am in aCategory or if I am in any
	of its sub categories recursively."

	aCategory allCategoriesDo: [:cat |
		(self hasCategory: cat) ifTrue: [^ true]].
	^false! !

!SMCategorizableObject methodsFor: 'private' stamp: 'gh 11/27/2002 12:35'!
removeCategory: aCategory
	"Remove category from me if I am in it."

	(categories notNil and: [categories includes: aCategory]) ifTrue:[
		aCategory removeObject: self.
		categories remove: aCategory].
	^aCategory! !

!SMCategorizableObject methodsFor: 'private' stamp: 'gh 11/27/2002 12:35'!
removeFromCategories
	"Remove me from all my categories."

	categories ifNotNil:[
		categories copy do: [:cat | self removeCategory: cat ]]! !

!SMCategorizableObject methodsFor: 'resources' stamp: 'gk 11/24/2005 11:23'!
removeResource: aResource
	"Disconnect and remove the resource."
	
	aResource object: nil.
	^resources remove: aResource! !

!SMCategorizableObject methodsFor: 'accessing' stamp: 'gk 7/27/2004 13:09'!
resources
	"Lazily initialized."

	^resources ifNil: [OrderedCollection new]! !

SMCategorizableObject subclass: #SMPackageRelease
	instanceVariableNames: 'publisher automaticVersion version note downloadUrl package repository sha1sum'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMPackageRelease commentStamp: '<historical>' prior: 0!
A package release refers to a specific version of the package.
Releases are auto numbered (with a VersionNumber) and also has a designated version name which can be whatever the maintainer wants.
There is also a release note and the URL for download. The inherited url is for any homepage for the release.
The instvar publisher refers to the SMAccount that owned the package at the time of the release and the instvar package refers to the owning package.
The instvar repository holds a String that is used to connect to the live repository for the package release, for example a Monticello repository.!

!SMPackageRelease classSide methodsFor: 'instance creation' stamp: 'gk 12/7/2005 14:16'!
newFromRelease: aPackageRelease package: aPackage
	"Create a new release from a given release."

	^super new initializeFromRelease: aPackageRelease package: aPackage! !

!SMPackageRelease methodsFor: 'configurations' stamp: 'gk 7/27/2004 17:58'!
addConfiguration
	"Create and add a new SMPackageReleaseConfiguration and return it."

	^ self addResource: (SMPackageReleaseConfiguration newIn: map)! !

!SMPackageRelease methodsFor: 'accessing' stamp: 'gh 11/27/2002 11:22'!
automaticVersion
	"Return the VersionNumber for me."

	^automaticVersion! !

!SMPackageRelease methodsFor: 'accessing' stamp: 'gk 8/12/2003 17:21'!
automaticVersionString
	"Return my VersionNumber as a String."

	^automaticVersion versionString! !

!SMPackageRelease methodsFor: 'accessing' stamp: 'gk 8/13/2003 15:25'!
cacheDirectory
	^ map cache directoryForPackageRelease: self! !

!SMPackageRelease methodsFor: 'accessing' stamp: 'gk 11/23/2005 01:13'!
calculateSha1sum
	"Return the checksum of the currently cached file contents."

	^SecureHashAlgorithm new hashMessage: self contents
	
	
		! !

!SMPackageRelease methodsFor: 'configurations' stamp: 'gk 7/27/2004 14:29'!
configurations
	"Return all SMPackageReleaseConfigurations attached to this release."


	^ self embeddedResources select: [:er | er isConfiguration]! !

!SMPackageRelease methodsFor: 'accessing' stamp: 'gk 11/22/2005 22:41'!
contents
	"Return the contents of the cached file.
	If it is not downloadable, or if the file
	is not cached, return nil."

	^map cache contents: self! !

!SMPackageRelease methodsFor: 'accessing' stamp: 'gk 11/23/2005 01:50'!
correctSha1sum: content
	"Return if the checksum of the content is correct.
	If we have none, then we consider that to be correct."
	
	^sha1sum isNil or: [sha1sum = (SecureHashAlgorithm new hashMessage: content)]
	
	
		! !

!SMPackageRelease methodsFor: 'deletion' stamp: 'gh 11/28/2002 21:32'!
delete
	super delete.
	package removeRelease: self! !

!SMPackageRelease methodsFor: 'services' stamp: 'gk 8/13/2003 15:43'!
download
	"Force a download into the cache regardless if it is already there."

	^map cache download: self! !

!SMPackageRelease methodsFor: 'accessing' stamp: 'gh 11/27/2002 14:03'!
downloadFileName
	"Cut out the filename from the url."

	downloadUrl isEmpty ifTrue: [^nil].
	^downloadUrl asUrl path last! !

!SMPackageRelease methodsFor: 'accessing' stamp: 'gh 11/27/2002 15:31'!
downloadUrl
	^downloadUrl! !

!SMPackageRelease methodsFor: 'accessing' stamp: 'gh 11/27/2002 15:31'!
downloadUrl: anObject
	^downloadUrl := anObject! !

!SMPackageRelease methodsFor: 'services' stamp: 'gk 7/13/2004 00:48'!
eitherVersion
	"Return either version:
		1. If the maintainer entered a version then we use that.
		2. Otherwise we use the automatic version with an 'r' prepended."

	^version notEmpty
			ifTrue:[version]
			ifFalse:['r', automaticVersion versionString]! !

!SMPackageRelease methodsFor: 'services' stamp: 'gk 7/10/2004 03:53'!
ensureInCache
	"Makes sure the file is in the cache.
	Return true on success, otherwise false."

	^map cache add: self! !

!SMPackageRelease methodsFor: 'printing' stamp: 'yo 7/26/2004 22:06'!
fullDescription
	"Return a full textual description of the package release."

	| s |
	s := TextStream on: (Text new: 400).
	self describe: self package name withBoldLabel: 'Package name: ' on: s.

	self 
		describe: self version
		withBoldLabel: 'version: '
		on: s.

	categories isEmptyOrNil 
		ifFalse: 
			[s
				cr;
				withAttribute: TextEmphasis bold do: [s nextPutAll: 'Categories: '];
				cr.
			self categoriesDo: 
					[:c | 
					s
						tab;
						withAttribute: TextEmphasis italic
							do: 
								[c parentsDo: 
										[:p | 
										s
											nextPutAll: p name;
											nextPutAll: '/'].
								s nextPutAll: c name];
						nextPutAll: ' - ' , c summary;
						cr].
			s cr].

	self note isEmptyOrNil 
		ifFalse: 
			[s
				cr;
				withAttribute: TextEmphasis bold do: [s nextPutAll: 'Version Comment:'].
			s cr.
			s withAttribute: (TextIndent tabs: 1) do: [s nextPutAll: self note].
			s
				cr;
				cr].
	url isEmptyOrNil 
		ifFalse: 
			[s
				withAttribute: TextEmphasis bold do: [s nextPutAll: 'Homepage:'];
				tab;
				withAttribute: (TextURL new url: url) do: [s nextPutAll: url];
				cr].
	self downloadUrl isEmptyOrNil 
		ifFalse: 
			[s
				withAttribute: TextEmphasis bold do: [s nextPutAll: 'Download:'];
				tab;
				withAttribute: (TextURL new url: self downloadUrl)
					do: [s nextPutAll: self downloadUrl];
				cr].
	^s contents.

! !

!SMPackageRelease methodsFor: 'services' stamp: 'gk 7/12/2004 16:04'!
fullVersion
	"Return version followed by the automatic version
	with r prepended in parenthesis."

	^version, ' (r', automaticVersion versionString, ')'! !

!SMPackageRelease methodsFor: 'view' stamp: 'gk 10/12/2005 12:25'!
getLink: aBuilder
	"Return a link for using on the web."

	^self getLink: aBuilder view: aBuilder view! !

!SMPackageRelease methodsFor: 'view' stamp: 'gk 10/12/2005 12:27'!
getLink: aBuilder view: aView
	"Return a link for using on the web."

	^aBuilder getLinkTop: self relativeUrl text: self packageNameWithVersion! !

!SMPackageRelease methodsFor: 'view' stamp: 'gk 10/12/2005 12:27'!
getShortLink: aBuilder

	^aBuilder getLinkTop: self relativeUrl text: self listName! !

!SMPackageRelease methodsFor: 'configurations' stamp: 'gk 9/20/2004 21:17'!
hasFulfilledConfiguration
	"Is any of the configurations already fulfilled?
	A fulfilled configuration has all required releases
	already installed, this means the release can be
	trivially installed."
	
	^self workingConfigurations anySatisfy: [:c | c isFulfilled]! !

!SMPackageRelease methodsFor: 'configurations' stamp: 'gk 9/20/2004 21:16'!
hasNoConfigurations
	"Does this release lack configurations,
	both working or failed ones? This is interpreted
	as if the release has no dependencies."

	^self configurations isEmpty! !

!SMPackageRelease methodsFor: 'initialize-release' stamp: 'gk 5/5/2006 02:07'!
initializeFromRelease: parentRelease package: aPackage
	"Initialize package release from a given parent.
	Branch if needed."

	self map: aPackage map id: UUID new.
	package := aPackage.
	automaticVersion :=
		parentRelease
			ifNil: [VersionNumber first]
			ifNotNil: [parentRelease nextOrBranch].
	version := note := downloadUrl := ''! !

!SMPackageRelease methodsFor: 'services' stamp: 'gk 2/16/2004 20:04'!
install
	"Install this package release."

	^map installPackageRelease: self! !

!SMPackageRelease methodsFor: 'testing' stamp: 'gk 8/13/2003 15:32'!
isCached
	"Delegate to last release."

	^map cache includes: self! !

!SMPackageRelease methodsFor: 'testing' stamp: 'gk 5/5/2006 01:54'!
isCompatibleWithCurrentSystemVersion
	"Return true if this release is listed as being compatible with the SystemVersion of the current image.  Only checks major/minor version number; does not differentiate between alpha/beta/gamma releases.  Checks version categories of both the SMPackageRelease and the parent SMPackage."

	| current |
	current := self majorMinorVersionFrom: SystemVersion current version.
	self categories, self package categories do: [:c |
		((c parent name = 'Squeak versions') and: [
			(self majorMinorVersionFrom: c name) = current])
				ifTrue: [^true]].
	^ false
		
"	^ (self categories, self package categories
		detect:
			[:cat | (cat parent name = 'Squeak versions')
					and: [(SystemVersion new version: cat name) majorMinorVersion = SystemVersion current majorMinorVersion]]
		ifNone: []) notNil
"! !

!SMPackageRelease methodsFor: 'testing' stamp: 'gk 10/15/2003 12:37'!
isDownloadable
	"Answer if I can be downloaded.
	We simply verify that the download url
	ends with a filename."

	^self downloadFileName isEmptyOrNil not! !

!SMPackageRelease methodsFor: 'testing' stamp: 'gk 11/16/2003 23:48'!
isInstallable
	"Answer if there is any installer for me.
	This depends typically on the filename of
	the download url, but can in the future
	depend on other things too.
	It does *not* say if the release is installed or not."

	^SMInstaller isInstallable: self! !

!SMPackageRelease methodsFor: 'testing' stamp: 'gk 7/13/2004 13:59'!
isInstalled
	"Answer if this release is installed."

	^(map installedReleaseOf: package) == self! !

!SMPackageRelease methodsFor: 'testing' stamp: 'gh 12/1/2002 19:52'!
isPackageRelease
	^true! !

!SMPackageRelease methodsFor: 'testing' stamp: 'gk 9/23/2003 21:09'!
isPublished
	"It is published when the publisher is set."

	^publisher notNil! !

!SMPackageRelease methodsFor: 'testing' stamp: 'gk 11/17/2003 12:01'!
isUpgradeable
	"Answer if there is any installer that can upgrade me.
	This depends typically on the filename of
	the download url, but can in the future
	depend on other things too.
	It does *not* say if the package is installed or not
	or if there is a newer version available or not."

	^SMInstaller isUpgradeable: self! !

!SMPackageRelease methodsFor: 'printing' stamp: 'gk 7/10/2004 03:35'!
listName
	"Return something suitable for showing in lists.
	We list the manual version after a dash if it is available.
	We don't list the release name."

	^version isEmpty
		ifFalse: [self automaticVersion versionString , '-', version]
		ifTrue: [self automaticVersion versionString] ! !

!SMPackageRelease methodsFor: 'private' stamp: 'gk 5/5/2006 01:54'!
majorMinorVersionFrom: aVersionName

	| start |
	start := aVersionName indexOf: $..
	start = 0 ifTrue: [^ aVersionName].
	aVersionName size = start ifTrue: [^ aVersionName].
	start + 1 to: aVersionName size do: [:i |
		(aVersionName at: i) isDigit ifFalse: [^aVersionName copyFrom: 1 to: i - 1]].
	^aVersionName! !

!SMPackageRelease methodsFor: 'testing' stamp: 'gk 7/13/2004 13:24'!
newerThan: aRelease
	"Answer if this release was made after <aRelease>."
	
	^aRelease automaticVersion < automaticVersion! !

!SMPackageRelease methodsFor: 'services' stamp: 'gk 5/5/2006 02:07'!
nextOrBranch
	"Return a new automaticVersion that is either
	the next following my version, or if that is taken
	a branch, or if that is taken too - a branch from it and so on.
	Yes, it sucks, but I don't have time hacking VersionNumber right now."

	| nextVersion nextBranch |
	nextVersion := automaticVersion next.
	(package releaseWithAutomaticVersion: nextVersion) ifNil: [^nextVersion].
	nextBranch := automaticVersion branchNext.
	[(package releaseWithAutomaticVersion: nextBranch) notNil]
		whileTrue: [nextBranch := nextBranch branchNext].
	^nextBranch 
! !

!SMPackageRelease methodsFor: 'accessing' stamp: 'gh 11/27/2002 12:43'!
note
	^note! !

!SMPackageRelease methodsFor: 'accessing' stamp: 'gh 11/27/2002 12:43'!
note: anObject
	^note := anObject! !

!SMPackageRelease methodsFor: 'services' stamp: 'gk 11/17/2003 01:25'!
noteInstalled
	"This package release was just successfully installed.
	We tell the map so that it can keep track of what
	package releases are installed."

	map noteInstalled: self! !

!SMPackageRelease methodsFor: 'services' stamp: 'gk 1/23/2004 10:01'!
noteUninstalled
	"This package release was just successfully uninstalled.
	We tell the map so that it can keep track of what
	package releases are installed."

	self error: 'Uninstall is not working yet!!'.
	map noteUninstalled: self! !

!SMPackageRelease methodsFor: 'testing' stamp: 'gk 7/13/2004 13:24'!
olderThan: aRelease
	"Answer if this release was made before <aRelease>."
	
	^automaticVersion < aRelease automaticVersion! !

!SMPackageRelease methodsFor: 'accessing' stamp: 'gh 3/15/2003 20:55'!
package
	"Get the package that I belong to."

	^package! !

!SMPackageRelease methodsFor: 'private' stamp: 'stephaneducasse 2/4/2006 20:38'!
package: aPackage
	"Set when I am created."

	package := aPackage! !

!SMPackageRelease methodsFor: 'printing' stamp: 'gk 7/10/2004 03:38'!
packageNameWithVersion
	"Return '<packageName> <autoVersion>-<version>' like:
		'SqueakMap 5-0.92'	"

	^package name, ' ', self listName! !

!SMPackageRelease methodsFor: 'services' stamp: 'gk 10/12/2005 13:25'!
parentRelease
	"Return my parent release based on the automatic
	version number."

	^package parentReleaseFor: self! !

!SMPackageRelease methodsFor: 'services' stamp: 'gk 10/12/2005 13:23'!
previousRelease
	"Return the release before me.
	Returns nil if there is none.
	This is chronological order and not how they relate."

	^package previousReleaseFor: self! !

!SMPackageRelease methodsFor: 'printing' stamp: 'gk 7/10/2004 03:38'!
printName
	"Return a String identifying object without context."

	^self packageNameWithVersion! !

!SMPackageRelease methodsFor: 'printing' stamp: 'gk 12/9/2003 00:13'!
printOn: aStream

	aStream nextPutAll: self class name, '[', self packageNameWithVersion, ']'! !

!SMPackageRelease methodsFor: 'accessing' stamp: 'gk 6/26/2003 14:59'!
publisher
	^publisher! !

!SMPackageRelease methodsFor: 'accessing' stamp: 'gk 6/26/2003 15:00'!
publisher: anObject
	publisher := anObject! !

!SMPackageRelease methodsFor: 'services' stamp: 'gk 10/12/2005 00:24'!
relativeUrl
	"Return the relative url for this release on an SM server."
	
	^'package/', package id asString, '/autoversion/', automaticVersion versionString! !

!SMPackageRelease methodsFor: 'accessing' stamp: 'gk 7/15/2004 16:26'!
repository
	^repository! !

!SMPackageRelease methodsFor: 'accessing' stamp: 'gk 5/5/2006 02:07'!
repository: aString
	repository := aString! !

!SMPackageRelease methodsFor: 'accessing' stamp: 'gk 11/22/2005 22:29'!
sha1sum
	^sha1sum! !

!SMPackageRelease methodsFor: 'accessing' stamp: 'gk 5/5/2006 02:07'!
sha1sum: aString
	sha1sum := aString! !

!SMPackageRelease methodsFor: 'services' stamp: 'gk 7/13/2004 00:49'!
smartVersion
	"This method is used to ensure that we always have a
	version name for the package release even if the maintainer didn't
	bother to enter one. Is is calculated like this:
		1. If the maintainer entered a version then we use that.
		2. Otherwise we use the automatic version with an 'r' prepended.
		3. If the release is not published we enclose it in parenthesis."

	^ self isPublished ifTrue: [self eitherVersion] ifFalse: ['(', self eitherVersion, ')']! !

!SMPackageRelease methodsFor: 'printing' stamp: 'gk 11/14/2003 00:11'!
type

	^'Package release'! !

!SMPackageRelease methodsFor: 'services' stamp: 'stephaneducasse 2/4/2006 20:38'!
upgrade
	"Upgrade this package release if there is a new release available."

	| newRelease |
	newRelease := package lastPublishedReleaseForCurrentSystemVersionNewerThan: self.
	newRelease ifNotNil: [(SMInstaller forPackageRelease: newRelease) upgrade]! !

!SMPackageRelease methodsFor: 'accessing' stamp: 'gh 11/27/2002 15:30'!
version
	^version! !

!SMPackageRelease methodsFor: 'accessing' stamp: 'gh 11/27/2002 15:30'!
version: anObject
	^version := anObject! !

!SMPackageRelease methodsFor: 'view' stamp: 'gh 3/15/2003 19:37'!
viewFor: uiObject
	"This is a double dispatch mechanism for multiple views
	for multiple uis."

	^uiObject packageReleaseViewOn: self! !

!SMPackageRelease methodsFor: 'configurations' stamp: 'gk 7/27/2004 14:55'!
workingConfigurations
	"Return all working configurations."
	
	^ self configurations select: [:c | c isWorking ]! !

SMCategorizableObject subclass: #SMRootedObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMRootedObject commentStamp: '<historical>' prior: 0!
An SMRootedObject is an SMObject that can be "rooted" in a given homeMap. This concept is for the upcoming new architecture with a tree of SM servers. Not used yet.!

SMRootedObject subclass: #SMAccount
	instanceVariableNames: 'initials email signature password newPassword advogatoId objects coObjects isAdmin'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMAccount commentStamp: '<historical>' prior: 0!
SMAccount is the object representing a user account in SqueakMap - a Squeaker that owns information in SqueakMap.

It keeps track of the email address, developer initials and two passwords used to access the account. There is also an advogatoId (people.squeakfoundation.org username) and a signature field (not used yet). The flag isAdmin is a crude way of marking a user as a superuser, this will possibly be changed in the future and instead expressed using a category.

Passwords are stored as secure hashes. The extra password (newPassword) is used when the regular password is forgotten - it is then randomly set and an email is sent out containing it to the registered email. This enables the original password to still work. When logging in, the user gets a chance to enter a new regular password overwriting the old one and clearing the random new password in the process.

The instvar objects holds all SMPersonalObjects (instances of subclasses) that this account "owns" - these are typically instances of SMPackages and SMResources, but are not limited to be.

The instvar coObjects holds all SMPersonalObjects that this account is co-maintaining - these are typically instances of SMPackages and SMResources.

Finally the account also maintains a directory with uploaded files on the server. This directory has the UUID of the account as its name and it is located under sm/accounts!

!SMAccount methodsFor: 'objects' stamp: 'gk 11/11/2003 16:16'!
addCoObject: anObject
	"Add <anObject> to this account.
	Only called from #addMaintainer:."
	
	(coObjects includes: anObject)
		ifFalse:[coObjects add: anObject]! !

!SMAccount methodsFor: 'objects' stamp: 'gk 11/11/2003 20:49'!
addObject: anObject
	"Add <anObject> to this account. Also makes sure the
	reverse reference is correct."
	
	(objects includes: anObject) ifFalse:[
		objects add: anObject.
		anObject owner: self.
		map addObject: anObject]! !

!SMAccount methodsFor: 'accessing' stamp: 'gk 11/13/2003 23:06'!
advogatoId
	^advogatoId! !

!SMAccount methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
advogatoId: aString
	advogatoId := aString! !

!SMAccount methodsFor: 'objects' stamp: 'gk 11/24/2005 11:13'!
coPackageWithId: anIdString
	"Return the correct package or nil."

	^self withId: anIdString in: self coPackages! !

!SMAccount methodsFor: 'objects' stamp: 'gk 11/24/2005 11:10'!
coPackages
	"Return all co-maintained packages."

	^coObjects select: [:o | o isPackage]! !

!SMAccount methodsFor: 'passwords' stamp: 'stephaneducasse 2/4/2006 20:38'!
correctPassword: aPassword
	"We store the password as a SHA hash so that we can let the slave maps
	have them too. Also check the optional new random password."

	| try |
	aPassword isEmptyOrNil ifTrue:[^false].
	try := SecureHashAlgorithm new hashMessage: aPassword.
	^password = try or: [newPassword = try]! !

!SMAccount methodsFor: 'passwords' stamp: 'gk 5/5/2006 02:07'!
createRandomPassword
	"Create a random password and set it
	in parallell to the regular one."

	| randomPass |
	randomPass := String streamContents: [:stream | 10 timesRepeat: [ stream nextPut: 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789' atRandom]].
	self setNewPassword: randomPass.
	^randomPass! !

!SMAccount methodsFor: 'files' stamp: 'gk 3/8/2004 19:26'!
delete
	"Delete this account. First delete all SM objects we own
	and disconnect this account from those we co-maintain."

	objects do: [:o | o delete].
	coObjects do: [:co | co removeMaintainer: self].
	super delete
! !

!SMAccount methodsFor: 'files' stamp: 'stephaneducasse 2/4/2006 20:38'!
deleteFiles: fileNames
	"Delete all fileNames from the uploads directory."

	| dir |
	dir := self uploadsDirectory.
	fileNames do: [:fn | dir deleteFileNamed: fn]
! !

!SMAccount methodsFor: 'files' stamp: 'stephaneducasse 2/4/2006 20:38'!
directory
	"Get the directory for the account."

	| dir |
	dir := (map directory directoryNamed: 'accounts') assureExistence; yourself.
	^(dir directoryNamed: id asString) assureExistence; yourself
! !

!SMAccount methodsFor: 'accessing' stamp: 'gk 6/26/2003 14:26'!
email
	^email! !

!SMAccount methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
email: address
	email := address! !

!SMAccount methodsFor: 'files' stamp: 'gk 3/8/2004 19:26'!
entries
	"Return all file entries in the upload directory."

	^self uploadsDirectory entries! !

!SMAccount methodsFor: 'files' stamp: 'gk 8/15/2003 12:08'!
files
	"Return filenames for uploaded files."

	^self uploadsDirectory fileNames
! !

!SMAccount methodsFor: 'view' stamp: 'gk 10/12/2005 12:24'!
getLink: aBuilder
	"Return a link for using on the web.
	Always from the top."

	^aBuilder getLinkTop: 'accountbyid/', id asString text: self nameWithInitials! !

!SMAccount methodsFor: 'initialize-release' stamp: 'stephaneducasse 2/4/2006 20:38'!
initialize
	"Initialize account."

	super initialize.
	initials := signature := advogatoId := ''.
	isAdmin := false.
	objects := OrderedCollection new.
	coObjects := OrderedCollection new! !

!SMAccount methodsFor: 'accessing' stamp: 'gk 7/30/2003 14:10'!
initials
	^initials! !

!SMAccount methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
initials: aString
	"If these are changed we need to update the dictionary in the map."

	initials ~= aString ifTrue: [
		initials := aString.
		map clearUsernames]! !

!SMAccount methodsFor: 'testing' stamp: 'gk 6/26/2003 16:47'!
isAccount
	^true! !

!SMAccount methodsFor: 'accessing' stamp: 'gk 11/17/2003 11:49'!
isAdmin
	^isAdmin ifNil: [false] ifNotNil: [isAdmin]! !

!SMAccount methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
isAdmin: aBoolean
	isAdmin := aBoolean! !

!SMAccount methodsFor: 'view' stamp: 'gk 8/4/2003 14:10'!
logout
	"Automatically called upon logout. Do nothing."! !

!SMAccount methodsFor: 'objects' stamp: 'gk 8/7/2003 21:00'!
moveObject: aPersonalObject toAccount: anAccount
	"Transfer the ownership of the given personal object to <anAccount>."

	self removeObject: aPersonalObject.
	anAccount addObject: aPersonalObject! !

!SMAccount methodsFor: 'accessing' stamp: 'gk 11/24/2005 11:09'!
nameAndEmail
	"This is not really correct, the name needs to be
	mime encoded."

	^name , ' <', email, '>'! !

!SMAccount methodsFor: 'view' stamp: 'gk 8/5/2003 13:16'!
nameWithInitials
	"Return name and developer initials within parentheses."

	^name, ' (', (initials isEmptyOrNil ifTrue: ['not entered'] ifFalse: [initials]) , ')'! !

!SMAccount methodsFor: 'files' stamp: 'gk 4/18/2006 22:53'!
newFile: fileName block: aBlock
	"Create a new file. Let <aBlock> fill the file with content by calling it with a stream."

	| dir stream |
	dir := self uploadsDirectory.
	[(dir fileExists: fileName) ifTrue:[dir deleteFileNamed: fileName].
	stream := dir newFileNamed: fileName.
	stream binary.
	aBlock value: stream] ensure: [stream close]! !

!SMAccount methodsFor: 'accessing' stamp: 'gk 8/4/2003 16:34'!
newPassword
	"Get the parallell password hash."

	^newPassword! !

!SMAccount methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
newPassword: aHashNumber
	"Set the parallell password hash."

	newPassword := aHashNumber! !

!SMAccount methodsFor: 'testing' stamp: 'gk 11/11/2003 17:24'!
owns: anObject
	^objects includes: anObject! !

!SMAccount methodsFor: 'objects' stamp: 'gk 11/24/2005 11:14'!
packageWithId: anIdString
	"Return the correct package or nil."

	^self withId: anIdString in: self packages! !

!SMAccount methodsFor: 'objects' stamp: 'gk 11/24/2005 11:14'!
packages
	"Return all owned packages."

	^objects select: [:o | o isPackage]! !

!SMAccount methodsFor: 'accessing' stamp: 'gk 8/4/2003 15:42'!
password
	"Get the password hash."

	^password! !

!SMAccount methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
password: aHashNumber
	"Set the password hash."

	password := aHashNumber! !

!SMAccount methodsFor: 'view' stamp: 'gk 8/8/2003 00:09'!
publicViewFor: uiObject
	"This is a double dispatch mechanism for multiple views
	for multiple uis."

	^uiObject publicAccountViewOn: self! !

!SMAccount methodsFor: 'objects' stamp: 'gk 11/11/2003 16:16'!
removeCoObject: anObject
	"Remove <anObject> from this account.
	Only called from #removeMaintainer:."

	(coObjects includes: anObject) ifTrue: [
		coObjects remove: anObject]! !

!SMAccount methodsFor: 'objects' stamp: 'gk 8/7/2003 20:56'!
removeObject: anObject
	"Remove <anObject> from this account. Also makes sure the
	reverse reference is cleared."

	(objects includes: anObject) ifTrue: [
		anObject owner: nil.
		objects remove: anObject]! !

!SMAccount methodsFor: 'passwords' stamp: 'stephaneducasse 2/4/2006 20:38'!
setNewPassword: aString
	"Set a new parallell password the user can use to get in
	if the old password is forgotten. We don't delete the old
	password since the request for this new password is made
	anonymously. Note that the password is stored as a secured
	hash large integer."

	newPassword := SecureHashAlgorithm new hashMessage: aString! !

!SMAccount methodsFor: 'passwords' stamp: 'stephaneducasse 2/4/2006 20:38'!
setPassword: aString
	"We also clear the random extra password."

	password := SecureHashAlgorithm new hashMessage: aString.
	newPassword := nil! !

!SMAccount methodsFor: 'accessing' stamp: 'gk 8/4/2003 15:56'!
signature
	"Get the signature."

	^signature! !

!SMAccount methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
signature: aSignature
	"Set the signature."

	signature := aSignature! !

!SMAccount methodsFor: 'files' stamp: 'gk 4/3/2006 23:58'!
streamForFile: fileName
	"Return a readonly stream for file <fileName>.
	If the file does not exist return nil."

	| stream |
	[stream := StandardFileStream oldFileNamed: (self uploadsDirectory fullNameFor: fileName)]
		on: FileDoesNotExistException do: [^nil].
	^stream! !

!SMAccount methodsFor: 'printing' stamp: 'gk 11/24/2005 11:09'!
type
	"Used in various views."

	^'Account'! !

!SMAccount methodsFor: 'files' stamp: 'gk 8/14/2003 14:23'!
uploadsDirectory
	"Get the directory for uploaded files, create it if missing."

	^(self directory directoryNamed: 'uploads') assureExistence; yourself
! !

!SMAccount methodsFor: 'view' stamp: 'gk 6/26/2003 16:01'!
viewFor: uiObject
	"This is a double dispatch mechanism for multiple views
	for multiple uis."

	^uiObject accountViewOn: self! !

SMRootedObject subclass: #SMPersonalObject
	instanceVariableNames: 'owner'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMPersonalObject commentStamp: 'gk 7/27/2004 13:28' prior: 0!
SMPersonalObject is the abstract base class for things that belong/are owned by a user account in SqueakMap. Most things are personal objects - but the SMCategories aren't for example.

A personal object has a reference to the SMAccount owning it.!

SMPersonalObject subclass: #SMMaintainableObject
	instanceVariableNames: 'maintainers rss feedbackEmail'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMMaintainableObject commentStamp: '<historical>' prior: 0!
A mainainable object is a personal object that is of such a complexity that it seems beneficial to optionally co-maintain with other people.
It thus has a potential list of maintainers - other accounts that also can modify the object. It also has a field to be able to refer to an RSS feed regarding the object.!

SMMaintainableObject subclass: #SMDocument
	instanceVariableNames: 'description author'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMDocument commentStamp: '<historical>' prior: 0!
An SMDocument refers typically to a piece of digital information accessible through a URL. :-)
This means it can be downloaded and cached.

The instvar description describes the document and instvar author references the name and/or email of the original author.

SMDocument has one subclasses - SMPackage. Since SqueakMap is primarily meant for
keeping track of installable source packages of Squeak software, a specific subclass handles those.!

!SMDocument methodsFor: 'accessing' stamp: 'gk 3/8/2004 19:51'!
author
	^author! !

!SMDocument methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
author: aString
	author := aString! !

!SMDocument methodsFor: 'accessing' stamp: 'gk 3/8/2004 19:52'!
description
	^description! !

!SMDocument methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
description: aString
	description := aString! !

!SMDocument methodsFor: 'cache' stamp: 'gk 3/8/2004 19:51'!
ensureInCache
	"Makes sure the file is in the cache."

	self subclassResponsibility ! !

!SMDocument methodsFor: 'initialize-release' stamp: 'gk 12/7/2005 13:51'!
initialize

	super initialize.
	description := author := ''! !

!SMDocument methodsFor: 'cache' stamp: 'gk 8/12/2003 17:09'!
isCached
	"Is the file corresponding to me in the local file cache?"

	self subclassResponsibility ! !

!SMDocument methodsFor: 'testing' stamp: 'gk 7/16/2004 11:08'!
isDownloadable
	"Answer if I can be downloaded.
	Default is false."
	
	^ false
	
	
! !

SMDocument subclass: #SMPackage
	instanceVariableNames: 'releases packageInfoName repository'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMPackage commentStamp: '<historical>' prior: 0!
An SMPackage represents a named piece of an installable "thing" in the image. Typically it is a code package, but it can be other things too.
It owns a collection of SMPackageReleases. Each release represents a version of the package, and each release has a URL that refers to the actual content that can be installed.

An SMPackage also knows its packageInfoName which can tie it into the image.!

!SMPackage methodsFor: 'private' stamp: 'gk 10/12/2005 15:44'!
addRelease: aRelease
	"Add the release. Make sure package is set."

	releases add: aRelease.
	aRelease package: self.
	^aRelease! !

!SMPackage methodsFor: 'cache' stamp: 'gk 8/12/2003 17:20'!
cacheDirectory
	^ self lastRelease cacheDirectory! !

!SMPackage methodsFor: 'accessing' stamp: 'gk 7/12/2004 16:05'!
currentVersion
	^self isPublished ifTrue: [self lastPublishedRelease version]! !

!SMPackage methodsFor: 'private' stamp: 'gh 11/28/2002 21:25'!
delete
	"Delete me. Delete my releases."

	super delete.
	self deleteReleases! !

!SMPackage methodsFor: 'private' stamp: 'gk 10/20/2005 00:17'!
deleteReleases
	"Delete my releases."

	releases copy do: [:release | release delete]! !

!SMPackage methodsFor: 'cache' stamp: 'gk 3/8/2004 19:56'!
download
	"Force download into cache."

	self isReleased ifFalse: [self error: 'There is no release for this package to download.'].
	^self lastRelease download! !

!SMPackage methodsFor: 'cache' stamp: 'gk 7/16/2004 11:01'!
ensureInCache
	"Makes sure all release files are in the cache."

	self releases do: [:rel | rel ensureInCache ]! !

!SMPackage methodsFor: 'services' stamp: 'gk 9/26/2004 23:34'!
firstRelease
	"Return the first release."

	^releases isEmpty ifTrue: [nil] ifFalse: [releases first]! !

!SMPackage methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
fullDescription
	"Return a full textual description of the package. 
	Most of the description is taken from the last release."
	| s publishedRelease |
	s := TextStream on: (Text new: 400).

	self
		describe: name
		withBoldLabel: 'Name:		'
		on: s.

	summary isEmptyOrNil
		ifFalse: [self
				describe: summary
				withBoldLabel: 'Summary:	'
				on: s ].

	author isEmptyOrNil
		ifFalse: [s
				withAttribute: TextEmphasis bold
				do: [s nextPutAll: 'Author:'];
				 tab;
				 tab.
			s
				withAttribute: (PluggableTextAttribute
						evalBlock: [self userInterface
										sendMailTo: (SMUtilities stripEmailFrom: author)
										regardingPackageRelease: self lastRelease])
				do: [s nextPutAll: author];
				 cr].
	self owner
		ifNotNil: [s
				withAttribute: TextEmphasis bold
				do: [s nextPutAll: 'Owner:'];
				 tab; tab.
			s
				withAttribute: (PluggableTextAttribute
						evalBlock: [self userInterface
										sendMailTo: self owner email
										regardingPackageRelease: self lastRelease])
				do: [s nextPutAll: self owner nameAndEmail];	
				 cr].

	self maintainers isEmpty ifFalse: [
		s withAttribute: TextEmphasis bold do: [s nextPutAll: 'Co-maintainers:']; tab.
		self maintainers do: [:com |
			com = self maintainers first ifFalse: [s nextPutAll: ', '].
			s
				withAttribute:
					(PluggableTextAttribute
						evalBlock: [self userInterface
									sendMailTo: com email
									regardingPackageRelease: self lastRelease])
				do: [s nextPutAll: com nameAndEmail]].
				s cr].

	description isEmptyOrNil
		ifFalse: [s cr.
			s
				withAttribute: TextEmphasis bold
				do: [s nextPutAll: 'Description:'].
			s cr.
			s
				withAttribute: (TextIndent tabs: 1)
				do: [s next: (description findLast: [ :c | c isSeparator not ]) putAll: description].
			s cr ].

	self describeCategoriesOn: s indent: 1.

	s cr.
	publishedRelease := self lastPublishedRelease.
	self
		describe: (self publishedVersion ifNil: ['<not published>'])
		withBoldLabel: 'Published version: '
		on: s.
	self isPublished ifTrue: [
		s
			withAttribute: TextEmphasis bold do: [ s nextPutAll: 'Created: ' ];
			print: publishedRelease created;
			cr.
			self note isEmptyOrNil
				ifFalse: [s
					withAttribute: TextEmphasis bold
					do: [s nextPutAll: 'Release note:'].
			s cr.
			s
				withAttribute: (TextIndent tabs: 1)
				do: [s nextPutAll: publishedRelease note].
			s cr ]].

	url isEmptyOrNil
		ifFalse: [s cr;
				withAttribute: TextEmphasis bold
				do: [s nextPutAll: 'Homepage:'];
				 tab;
				withAttribute: (TextURL new url: url)
				do: [s nextPutAll: url];
				 cr].

	^ s contents! !

!SMPackage methodsFor: 'view' stamp: 'gk 10/12/2005 13:10'!
getCoEditLink: aBuilder
	"Return a link for using on the web.
	Relative to the current view."

	^self getCoEditLink: aBuilder view: aBuilder view! !

!SMPackage methodsFor: 'view' stamp: 'gk 10/12/2005 13:09'!
getCoEditLink: aBuilder view: aView
	"Return a link for using on the web."

	^aBuilder getLink: 'copackage/', id asString, '/edit' text: 'edit' view: aView! !

!SMPackage methodsFor: 'view' stamp: 'gk 10/12/2005 13:11'!
getCoEditReleasesLink: aBuilder
	"Return a link for using on the web.
	Relative to the current view."

	^self getCoEditReleasesLink: aBuilder view: aBuilder view! !

!SMPackage methodsFor: 'view' stamp: 'gk 10/12/2005 13:11'!
getCoEditReleasesLink: aBuilder view: aView
	"Return a link for using on the web."

	^aBuilder getLink: 'copackage/', id asString, '/editreleases' text: 'edit releases' view: aView! !

!SMPackage methodsFor: 'view' stamp: 'gk 10/12/2005 13:06'!
getCoLink: aBuilder
	"Return a link for using on the web.
	Relative to the current view."

	^self getCoLink: aBuilder view: aBuilder view! !

!SMPackage methodsFor: 'view' stamp: 'gk 10/12/2005 13:06'!
getCoLink: aBuilder view: aView
	"Return a link for using on the web."

	^aBuilder getLink: 'copackage/', id asString text: name view: aView! !

!SMPackage methodsFor: 'view' stamp: 'gk 10/12/2005 13:13'!
getEditLink: aBuilder
	"Return a link for using on the web.
	Relative to the current view."

	^self getEditLink: aBuilder view: aBuilder view! !

!SMPackage methodsFor: 'view' stamp: 'gk 10/12/2005 13:13'!
getEditLink: aBuilder view: aView
	"Return a link for using on the web."

	^aBuilder getLink: 'package/', id asString, '/edit' text: 'edit' view: aView! !

!SMPackage methodsFor: 'view' stamp: 'gk 10/12/2005 13:13'!
getEditReleasesLink: aBuilder
	"Return a link for using on the web.
	Relative to the current view."

	^self getEditReleasesLink: aBuilder view: aBuilder view! !

!SMPackage methodsFor: 'view' stamp: 'gk 10/12/2005 13:13'!
getEditReleasesLink: aBuilder view: aView
	"Return a link for using on the web."

	^aBuilder getLink: 'package/', id asString, '/editreleases' text: 'edit releases' view: aView! !

!SMPackage methodsFor: 'view' stamp: 'gk 10/12/2005 12:24'!
getLink: aBuilder
	"Return a link for using on the web.
	Relative to the current view."

	^self getLink: aBuilder view: aBuilder view! !

!SMPackage methodsFor: 'view' stamp: 'gk 10/12/2005 12:19'!
getLink: aBuilder view: aView
	"Return a link for using on the web."

	^aBuilder getLink: 'package/', id asString text: name view: aView! !

!SMPackage methodsFor: 'initialize-release' stamp: 'gk 5/5/2006 02:07'!
initialize
	"Initialize package."

	super initialize.
	packageInfoName := ''.
	releases := OrderedCollection new! !

!SMPackage methodsFor: 'installation' stamp: 'gk 7/14/2004 17:32'!
install
	"Install the latest newer published version for this version of Squeak."

	^map installPackage: self! !

!SMPackage methodsFor: 'installation' stamp: 'gk 11/17/2003 02:27'!
installedRelease
	"Return the installed release.
	We ask the map. Return nil if this package is not installed."

	^map installedReleaseOf: self! !

!SMPackage methodsFor: 'printing' stamp: 'gk 11/17/2003 10:14'!
installedVersion
	"Return the version String for the installed version.
	We ask the map. Return nil if this package is not installed."

	^self installedRelease ifNotNilDo: [:r | r smartVersion]! !

!SMPackage methodsFor: 'testing' stamp: 'gk 7/13/2004 15:11'!
isAvailable
	"Answer if I am old or not installed regardless of
	if there is installer support for me. It also does
	not care if the newer release is not published
	or no for this Squeak version."

	^self isOld or: [self isInstalled not]! !

!SMPackage methodsFor: 'testing' stamp: 'gk 7/13/2004 14:55'!
isCached
	"Is the last release corresponding to me in the local file cache?
	NOTE: This doesn't honour #published nor if the release is
	intended for the current Squeak version."

	^self isReleased and: [self lastRelease isCached]! !

!SMPackage methodsFor: 'testing' stamp: 'gk 7/15/2004 17:21'!
isInstallable
	"Answer if any of my releases can be installed."

	^ releases anySatisfy: [:rel | rel isInstallable]! !

!SMPackage methodsFor: 'testing' stamp: 'gh 10/25/2002 11:08'!
isInstallableAndNotInstalled
	"Answer if there is any installer that
	can install me and I am not yet installed."

	^self isInstallable and: [self isInstalled not]! !

!SMPackage methodsFor: 'testing' stamp: 'gk 11/17/2003 11:02'!
isInstalled
	"Answer if any version of me is installed."

	^(map installedReleaseOf: self) notNil! !

!SMPackage methodsFor: 'testing' stamp: 'gk 7/13/2004 15:10'!
isOld
	"Answer if I am installed and there also is a
	newer version available *regardless* if it is
	not published or not for this Squeak version.
	This is for people who want to experiment!!"

	| installed |
	installed := map installedReleaseOf: self.
	^installed
		ifNil: [false]
		ifNotNil: [
			self releases anySatisfy: [:r |
				r newerThan: installed ]]! !

!SMPackage methodsFor: 'testing' stamp: 'gh 12/1/2002 19:40'!
isPackage
	^true! !

!SMPackage methodsFor: 'testing' stamp: 'gk 9/23/2003 21:35'!
isPublished
	"Answer if I have public releases."

	^releases anySatisfy: [:rel | rel isPublished]! !

!SMPackage methodsFor: 'testing' stamp: 'btr 11/20/2003 00:05'!
isReleased
	^ releases isEmpty not! !

!SMPackage methodsFor: 'testing' stamp: 'gk 7/13/2004 15:37'!
isSafeToInstall
	"Answer if I am NOT installed and there also is a
	published version for this version of Squeak available."

	^self isInstalled not and: [
		self lastPublishedReleaseForCurrentSystemVersion notNil]! !

!SMPackage methodsFor: 'testing' stamp: 'gk 7/14/2004 16:15'!
isSafelyAvailable
	"Answer if I am old or not installed regardless of
	if there is installer support for me. The
	newer release should be published
	and meant for this Squeak version."

	^self isSafeToInstall or: [self isSafelyOld]! !

!SMPackage methodsFor: 'testing' stamp: 'stephaneducasse 2/4/2006 20:38'!
isSafelyOld
	"Answer if I am installed and there also is a
	newer published version for this version of Squeak available."

	| installed |
	installed := self installedRelease.
	^installed ifNil: [false] ifNotNil: [
		^(self lastPublishedReleaseForCurrentSystemVersionNewerThan: installed) notNil]! !

!SMPackage methodsFor: 'testing' stamp: 'stephaneducasse 2/4/2006 20:38'!
isSafelyOldAndUpgradeable
	"Answer if I am installed and there also is a
	newer published version for this version of Squeak available
	that can be upgraded to (installer support)."

	| installed newRelease |
	installed := self installedRelease.
	^installed ifNil: [false] ifNotNil: [
		newRelease := self lastPublishedReleaseForCurrentSystemVersionNewerThan: installed.
		^newRelease ifNil: [false] ifNotNil: [newRelease isUpgradeable]]! !

!SMPackage methodsFor: 'services' stamp: 'gk 2/16/2004 20:14'!
lastPublishedRelease
	"Return the latest published release."

	^releases isEmpty ifTrue: [nil] ifFalse: [
		releases reversed detect: [:r | r isPublished] ifNone:[nil]]! !

!SMPackage methodsFor: 'services' stamp: 'gk 7/13/2004 13:28'!
lastPublishedReleaseForCurrentSystemVersion
	"Return the latest published release marked
	as compatible with the current SystemVersion."

	^releases isEmpty ifTrue: [nil] ifFalse: [
		releases reversed detect: [:r |
			r isPublished and: [r isCompatibleWithCurrentSystemVersion]]
				ifNone:[nil]]! !

!SMPackage methodsFor: 'services' stamp: 'gk 7/14/2004 17:15'!
lastPublishedReleaseForCurrentSystemVersionNewerThan: aRelease
	"Return the latest published release marked
	as compatible with the current SystemVersion
	that is newer than the given release."

	^releases isEmpty ifTrue: [nil] ifFalse: [
		releases reversed detect: [:r |
			(r isPublished and: [r newerThan: aRelease])
				and: [r isCompatibleWithCurrentSystemVersion]]
				 	ifNone:[nil]]! !

!SMPackage methodsFor: 'services' stamp: 'gk 8/4/2003 11:49'!
lastRelease
	"Return the latest release."

	^releases isEmpty ifTrue: [nil] ifFalse: [releases last]! !

!SMPackage methodsFor: 'accessing' stamp: 'gk 1/25/2004 16:23'!
maintainer
	"Deprecated"

	^self owner! !

!SMPackage methodsFor: 'deprecated' stamp: 'btr 11/20/2003 00:38'!
modulePath: p moduleVersion: v moduleTag: t versionComment: vc
	"Deprecated. Only kept for migration from SM 1.0x.
	Method used when recreating from storeOn: format."

	self isReleased ifTrue: [self lastRelease note: vc]! !

!SMPackage methodsFor: 'installation' stamp: 'gk 11/18/2003 02:14'!
nameWithVersionLabel
	^name, ' (', self versionLabel, ')'! !

!SMPackage methodsFor: 'private' stamp: 'gk 12/7/2005 14:15'!
newChildReleaseFrom: aRelease
	"Create a new release."

	^self addRelease: (map newObject: (SMPackageRelease newFromRelease: aRelease package: self))! !

!SMPackage methodsFor: 'private' stamp: 'gk 12/7/2005 14:07'!
newRelease
	"Create a new release. Just use the last
	chronological release as parent, if this is the first release
	that is nil."

	^self newChildReleaseFrom: self lastRelease! !

!SMPackage methodsFor: 'accessing' stamp: 'gk 7/12/2004 17:15'!
note
	^self isPublished ifTrue: [self lastPublishedRelease note]! !

!SMPackage methodsFor: 'accessing' stamp: 'gk 11/6/2003 15:24'!
packageInfoName
	^packageInfoName! !

!SMPackage methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
packageInfoName: aString
	packageInfoName := aString! !

!SMPackage methodsFor: 'services' stamp: 'gk 5/5/2006 02:07'!
parentReleaseFor: aPackageRelease
	"If there is none (the given release is automaticVersion '1'), return nil."

	| previousVersion |
	previousVersion := aPackageRelease automaticVersion previous.
	^releases detect: [:r | r automaticVersion = previousVersion] ifNone: [nil]! !

!SMPackage methodsFor: 'services' stamp: 'gh 11/27/2002 12:33'!
previousReleaseFor: aPackageRelease
	"If there is none, return nil."
	
	^releases before: aPackageRelease ifAbsent: [nil]! !

!SMPackage methodsFor: 'accessing' stamp: 'gk 7/12/2004 17:13'!
publishedVersion
	^self isPublished ifTrue: [self lastPublishedRelease version]! !

!SMPackage methodsFor: 'services' stamp: 'gk 11/18/2003 17:39'!
releaseWithAutomaticVersion: aVersion
	"Look up a specific package release of mine. Return nil if missing.
	They are few so we just do a #select:."

	^releases detect: [:rel | rel automaticVersion = aVersion ] ifNone: [nil]! !

!SMPackage methodsFor: 'services' stamp: 'gk 11/18/2003 17:39'!
releaseWithAutomaticVersionString: aVersionString
	"Look up a specific package release of mine. Return nil if missing.
	They are few so we just do a #select:."

	^self releaseWithAutomaticVersion: aVersionString asVersion! !

!SMPackage methodsFor: 'services' stamp: 'stephaneducasse 2/4/2006 20:38'!
releaseWithId: anIdString 
	"Look up a specific package release of mine. Return nil if missing.
	They are few so we just do a #select:."

	| anId |
	anId := UUID fromString: anIdString.
	releases detect: [:rel | rel id = anId ].
	^nil! !

!SMPackage methodsFor: 'services' stamp: 'gk 11/18/2003 17:41'!
releaseWithVersion: aVersionString
	"Look up a specific package release of mine. Return nil if missing.
	They are few so we just do a #select:."

	^releases detect: [:rel | rel version = aVersionString ] ifNone: [nil]! !

!SMPackage methodsFor: 'accessing' stamp: 'gh 12/1/2002 21:21'!
releases
	^releases! !

!SMPackage methodsFor: 'private' stamp: 'gh 11/28/2002 21:33'!
removeRelease: aRelease
	"Remove the release."

	releases remove: aRelease! !

!SMPackage methodsFor: 'accessing' stamp: 'gk 10/12/2005 23:28'!
repository
	^repository! !

!SMPackage methodsFor: 'accessing' stamp: 'gk 5/5/2006 02:07'!
repository: aString
	repository := aString! !

!SMPackage methodsFor: 'services' stamp: 'gk 7/13/2004 14:37'!
smartVersion
	"Delegate to last published release for this SystemVersion."

	| r |
	r := self lastPublishedReleaseForCurrentSystemVersion.
	^r ifNotNil: [r smartVersion] ifNil: ['']! !

!SMPackage methodsFor: 'printing' stamp: 'gk 11/14/2003 00:12'!
type

	^'Package'! !

!SMPackage methodsFor: 'installation' stamp: 'stephaneducasse 2/4/2006 20:38'!
upgrade
	"Upgrade to the latest newer published version for this version of Squeak."

	| installed |
	installed := self installedRelease.
	installed
		ifNil: [self error: 'No release installed, can not upgrade.']
		ifNotNil: [^installed upgrade]! !

!SMPackage methodsFor: 'installation' stamp: 'stephaneducasse 2/4/2006 20:38'!
upgradeOrInstall
	"Upgrade to or install the latest newer published version for this version of Squeak."

	| installed |
	installed := self installedRelease.
	installed
		ifNil: [^self install]
		ifNotNil: [^installed upgrade]! !

!SMPackage methodsFor: 'installation' stamp: 'stephaneducasse 2/4/2006 20:38'!
versionLabel
	"Return a label indicating installed and available version as:
		'1.0'      = 1.0 is installed and no new published version for this version of Squeak is available
		'1.0->1.1' = 1.0 is installed and 1.1 is published for this version of Squeak
		'->1.1'    = No version is installed and 1.1 is published for this version of Squeak
		'->(1.1)	 = No version is installed and there is only a non published version available for this version of Squeak

	The version showed is the one that #smartVersion returns.
	If a version name is in parenthesis it is not published."

	| installedVersion r r2 |
	r := self installedRelease.
	r ifNotNil: [
		installedVersion := r smartVersion.
		r2 := self lastPublishedReleaseForCurrentSystemVersionNewerThan: r]
	ifNil: [
		installedVersion := ''.
		r2 := self lastPublishedReleaseForCurrentSystemVersion ].
	^r2 ifNil: [installedVersion ] ifNotNil: [installedVersion, '->', r2 smartVersion].! !

!SMPackage methodsFor: 'view' stamp: 'gh 3/15/2003 18:05'!
viewFor: uiObject
	"This is a double dispatch mechanism for multiple views
	for multiple uis."

	^uiObject packageViewOn: self! !

!SMMaintainableObject methodsFor: 'maintainers' stamp: 'gk 5/5/2006 02:07'!
addMaintainer: anAccount
	"Add anAccount as a maintainer."

	maintainers ifNil: [maintainers := OrderedCollection new].
	maintainers add: anAccount.
	anAccount addCoObject: self! !

!SMMaintainableObject methodsFor: 'deletion' stamp: 'gk 7/27/2004 13:27'!
delete
	"Disconnect from maintainers."

	super delete.
	maintainers ifNotNil: [
		maintainers copy do: [:m | self removeMaintainer: m]]! !

!SMMaintainableObject methodsFor: 'maintainers' stamp: 'gk 10/12/2005 23:25'!
feedbackEmail
	^feedbackEmail! !

!SMMaintainableObject methodsFor: 'maintainers' stamp: 'gk 5/5/2006 02:07'!
feedbackEmail: anEmail
	feedbackEmail := anEmail! !

!SMMaintainableObject methodsFor: 'testing' stamp: 'gk 7/27/2004 13:26'!
isOwnerOrMaintainer: anAccount
	^ owner = anAccount or: [self maintainers includes: anAccount]! !

!SMMaintainableObject methodsFor: 'maintainers' stamp: 'gk 7/27/2004 13:26'!
maintainers
	"Return all maintainers."

	^maintainers ifNil: [#()]! !

!SMMaintainableObject methodsFor: 'maintainers' stamp: 'gk 7/27/2004 13:26'!
removeMaintainer: anAccount
	"Remove anAccount as a maintainer."

	maintainers ifNil: [^self].
	maintainers remove: anAccount.
	anAccount removeCoObject: self! !

!SMMaintainableObject methodsFor: 'maintainers' stamp: 'gk 7/27/2004 13:26'!
rss
	^rss! !

!SMMaintainableObject methodsFor: 'maintainers' stamp: 'gk 5/5/2006 02:07'!
rss: anUrl
	rss := anUrl! !

!SMPersonalObject methodsFor: 'deletion' stamp: 'gk 10/20/2005 00:12'!
delete
	"Disconnect from owner."

	super delete.
	owner removeObject: self! !

!SMPersonalObject methodsFor: 'accessing' stamp: 'gk 8/7/2003 20:56'!
owner
	^owner! !

!SMPersonalObject methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
owner: anAccount
	owner := anAccount! !

SMPersonalObject subclass: #SMResource
	instanceVariableNames: 'object version'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMResource commentStamp: '<historical>' prior: 0!
A resource is a document that is NOT a package. Thus, it is used for all the things interesting to register on SM that aren't packages. The are three major differences with resources:

- A resource keeps no track of version history like packages do with package releases. It only has a field for the current version.
- A resource can be embedded inside the map instead of being a document reached by a URL.
- A resource can be associated with another SMObject, the instvar object.

However, resources respond to some of the same actions as PackageReleases.!

SMResource subclass: #SMEmbeddedResource
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMEmbeddedResource commentStamp: '<historical>' prior: 0!
An embedded resource is a resource that is stored inside the map. No download is needed.
This means that embedded resources should be "small" and typically only be used for information that
is needed to be available at all times without downloading. A typical example is meta data for other SMObjects.

!

!SMEmbeddedResource methodsFor: 'services' stamp: 'gk 3/8/2004 19:58'!
download
	"A dummy method to respond as other resources would."

	^true! !

!SMEmbeddedResource methodsFor: 'services' stamp: 'gk 7/16/2004 11:03'!
ensureInCache
	"Makes sure the file is in the cache.
	An embedded resource doesn't have a file."

	^true! !

!SMEmbeddedResource methodsFor: 'testing' stamp: 'gk 8/12/2003 17:10'!
isCached
	"Is the file corresponding to me in the local file cache?
	Well consider it as true since I am embedded in the map."

	^true! !

!SMEmbeddedResource methodsFor: 'testing' stamp: 'gk 7/27/2004 14:31'!
isEmbedded
	^ true! !

SMEmbeddedResource subclass: #SMGenericEmbeddedResource
	instanceVariableNames: 'contents'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMGenericEmbeddedResource commentStamp: 'gk 10/12/2005 23:01' prior: 0!
A generic embedded resource is simply some kind of object, held in instvar #contents, that is stored inside the map. No download is needed.
Embedded resources should be "small" and typically only be used for resources that
are needed to be available at all times without downloading. A typical example is meta data for other SMObjects.

!

!SMGenericEmbeddedResource methodsFor: 'as yet unclassified' stamp: 'gk 10/12/2005 23:00'!
contents
	"Answers the contents object."

	^contents! !

!SMGenericEmbeddedResource methodsFor: 'as yet unclassified' stamp: 'gk 5/5/2006 02:07'!
contents: obj
	"Sets the actual contents of this resource.
	SM does not know what it is."
	
	contents := obj! !

SMGenericEmbeddedResource subclass: #SMKabunguHint
	instanceVariableNames: 'type'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMKabunguHint methodsFor: 'as yet unclassified' stamp: 'gk 10/12/2005 20:47'!
account
	
	^ self owner! !

!SMKabunguHint methodsFor: 'as yet unclassified' stamp: 'mist 8/22/2005 23:52'!
type

	^ type! !

!SMKabunguHint methodsFor: 'as yet unclassified' stamp: 'gk 5/5/2006 02:07'!
type: t

	type := t! !

SMEmbeddedResource subclass: #SMPackageReleaseConfiguration
	instanceVariableNames: 'requiredReleases status'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMPackageReleaseConfiguration commentStamp: '<historical>' prior: 0!
A package release configuration describes the result of testing the specific release with a set of other releases that it depends on.
The status instvar holds a symbol which reflects the result. Currently there are two valid values:
	#working
	#failing
	
The intention is that users and maintainers post these configurations to the map as "known working combinations of required releases".
Each SMPackageRelease can then have multiple of these configurations.!

!SMPackageReleaseConfiguration methodsFor: 'dependencies' stamp: 'gk 5/5/2006 02:07'!
addRequiredRelease: aRelease
	"Add <aRelease> as a required release. The release added
	can not indirectly refer back to this release."
	
	(self isCircular: aRelease) ifTrue: [self error: 'Circular dependencies not allowed.'].
	requiredReleases := requiredReleases copyWith: aRelease.
	^aRelease! !

!SMPackageReleaseConfiguration methodsFor: 'initialize-release' stamp: 'gk 5/5/2006 02:07'!
initialize
	super initialize.
	requiredReleases := #().
	status := #working! !

!SMPackageReleaseConfiguration methodsFor: 'private' stamp: 'gk 9/25/2004 00:22'!
isCircular: aRelease
	"Answer if there is a reference that goes back
	to the release of this configuration."

	"This is the base case"
	aRelease == object ifTrue: [^ true].
	
	aRelease configurations do: [:conf |
		conf requiredReleases do: [:rel |
			(self isCircular: rel) ifTrue: [^ true]]].
	^false! !

!SMPackageReleaseConfiguration methodsFor: 'testing' stamp: 'gk 7/27/2004 14:29'!
isConfiguration
	^true! !

!SMPackageReleaseConfiguration methodsFor: 'testing' stamp: 'gk 7/27/2004 18:23'!
isFailing
	^status == #failing! !

!SMPackageReleaseConfiguration methodsFor: 'testing' stamp: 'gk 7/27/2004 14:56'!
isFulfilled
	"Are all my required releases already installed?"
	
	^requiredReleases allSatisfy: [:r | r isInstalled ]! !

!SMPackageReleaseConfiguration methodsFor: 'testing' stamp: 'gk 7/27/2004 18:23'!
isWorking
	^status == #working! !

!SMPackageReleaseConfiguration methodsFor: 'printing' stamp: 'gk 7/28/2004 11:43'!
printOn: aStream

	aStream nextPutAll: 'Cfg['.
	requiredReleases do: [:r |
		aStream nextPutAll: r printString; space].
	aStream nextPutAll: ']'! !

!SMPackageReleaseConfiguration methodsFor: 'dependencies' stamp: 'gk 5/5/2006 02:07'!
removeRequiredRelease: aRelease
	"Remove <aRelease> as a required release."
	
	requiredReleases := requiredReleases copyWithout: aRelease.
	^ aRelease! !

!SMPackageReleaseConfiguration methodsFor: 'accessing' stamp: 'gk 7/27/2004 14:22'!
requiredReleases
	^ requiredReleases! !

!SMPackageReleaseConfiguration methodsFor: 'accessing' stamp: 'gk 7/27/2004 14:22'!
status
	^ status! !

SMResource subclass: #SMExternalResource
	instanceVariableNames: 'downloadUrl'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMExternalResource commentStamp: 'gk 3/8/2004 20:09' prior: 0!
An external resource is a downloadable resource.
The instance variable downloadUrl holds the URL to the resource and the resource is cacheable in the FileCache for the SqueakMap.
An external resource can be used for any kind of document that is to be attached to another SMObject.!

!SMExternalResource methodsFor: 'accessing' stamp: 'btr 5/28/2003 04:13'!
cacheDirectory
	^ map cache directoryForResource: self! !

!SMExternalResource methodsFor: 'accessing' stamp: 'gk 4/19/2006 01:32'!
contents
	"Return the contents of a stream from the downloaded resource.
	Not yet tested, this resource returns the stream and not its contents."

	map cache add: self.
	^(self cacheDirectory readOnlyFileNamed: self downloadFileName) binary; yourself! !

!SMExternalResource methodsFor: 'services' stamp: 'gk 8/13/2003 15:42'!
download
	"Force a download into the cache regardless if it is already there."

	^map cache download: self! !

!SMExternalResource methodsFor: 'accessing' stamp: 'btr 5/28/2003 04:13'!
downloadFileName
	"Cut out the filename from the url."

	downloadUrl isEmpty ifTrue: [^nil].
	^downloadUrl asUrl path last! !

!SMExternalResource methodsFor: 'accessing' stamp: 'btr 5/28/2003 04:14'!
downloadUrl
	^ downloadUrl! !

!SMExternalResource methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
downloadUrl: anUrl
	downloadUrl := anUrl! !

!SMExternalResource methodsFor: 'services' stamp: 'gk 8/13/2003 15:54'!
ensureInCache
	"Makes sure the file is in the cache."

	^map cache add: self! !

!SMExternalResource methodsFor: 'testing' stamp: 'gk 8/13/2003 15:32'!
isCached
	"Is the file corresponding to me in the local file cache?"

	^map cache includes: self! !

!SMExternalResource methodsFor: 'accessing' stamp: 'gk 7/16/2004 11:09'!
isDownloadable
	"Answer if I can be downloaded.
	We simply verify that the download url
	ends with a filename."

	^self downloadFileName isEmptyOrNil not! !

!SMResource classSide methodsFor: 'instance creation' stamp: 'btr 5/28/2003 04:31'!
forString: aString
	^ SMEmbeddedResource new content: aString! !

!SMResource classSide methodsFor: 'instance creation' stamp: 'btr 5/28/2003 04:30'!
forUrl: anUrl
	^ SMExternalResource new downloadUrl: anUrl; yourself! !

!SMResource methodsFor: 'testing' stamp: 'gk 7/27/2004 14:30'!
isConfiguration
	^ false! !

!SMResource methodsFor: 'testing' stamp: 'gk 7/27/2004 14:31'!
isEmbedded
	^false! !

!SMResource methodsFor: 'testing' stamp: 'btr 5/28/2003 04:25'!
isResource
	^ true! !

!SMResource methodsFor: 'accessing' stamp: 'gk 7/27/2004 13:12'!
object
	^object! !

!SMResource methodsFor: 'accessing' stamp: 'gk 5/5/2006 02:07'!
object: anSMCategorizableObject
	object := anSMCategorizableObject! !

!SMResource methodsFor: 'accessing' stamp: 'gk 9/23/2003 21:55'!
version
	^ version! !

!SMResource methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
version: aVersion
	version := aVersion! !

SMObject subclass: #SMCategory
	instanceVariableNames: 'mandatory subCategories parent objects'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMCategory commentStamp: 'gk 3/8/2004 19:44' prior: 0!
An SMCategory is a "tag" that can be attached to SMCategorizableObjects in order to classify them.

The SMCategories are arranged in a strict hierarchy and each SMCategory both knows it's parent and it's subcategories.
The instvar objects holds all SMObjects belonging to this category.
Instvars name and summary are already inherited from SMObject and describe the category.
The instvar url can be used to refer to a web page that can explain the category in more detail, typically a page at the Squeak Swiki.
SMCategory adds an instance variable called mandatory holding a Set with the classes (SMPackage, SMPackageRelease, SMAccount, SMResource etc) that must belong to at least one subcategory of this SMCategory. Obviously not many categories will be mandatory for each class.

The category tree is maintained by a few trusted people so that chaos will not reign. :-)
!

!SMCategory methodsFor: 'private' stamp: 'stephaneducasse 2/4/2006 20:38'!
addCategory: cat
	"Add a category as a subcategory to self.
	The collection of subcategories is lazily instantiated."

	subCategories ifNil: [subCategories := OrderedCollection new].
	cat parent ifNotNil: [cat parent removeCategory: cat ].
	subCategories add: cat.
	cat parent: self.
	^cat! !

!SMCategory methodsFor: 'private' stamp: 'gk 9/23/2003 21:59'!
addObject: anObject
	"Add <anObject> to this category. This should only be called
	from SMCategorizableObject>>addCategory: to ensure consistency."
	
	(objects includes: anObject) ifFalse:[objects add: anObject]! !

!SMCategory methodsFor: 'services' stamp: 'gk 11/17/2003 10:48'!
allCategoriesDo: aBlock
	"Evaluate <aBlock> for all categories below me including me,
	bottom up breadth-first."

	self allSubCategoriesDo: aBlock.
	aBlock value: self! !

!SMCategory methodsFor: 'services' stamp: 'gk 7/9/2004 02:59'!
allSubCategoriesDo: aBlock
	"Evaluate <aBlock> for all categories below me NOT including me,
	bottom up breadth-first."

	subCategories ifNil: [^self].
	subCategories do: [:sub |
		sub allSubCategoriesDo: aBlock.
		aBlock value: sub]! !

!SMCategory methodsFor: 'services' stamp: 'gh 8/5/2002 17:10'!
categoryBefore
	"Return the category listed before me in my parent.
	If I am first or I have no parent, return nil."

	parent isNil ifTrue:[^nil].
	parent subCategories first = self ifTrue:[^nil].
	^parent subCategories before: self
	! !

!SMCategory methodsFor: 'private' stamp: 'gk 11/24/2005 11:37'!
delete
	"Delete me. Disconnect me from my objects and my parent.
	Then delete my subcategories."

	super delete.
	self removeFromObjects; removeFromParent.
	self subCategories do: [:c | c delete ]! !

!SMCategory methodsFor: 'view' stamp: 'gk 10/12/2005 12:24'!
getLink: aBuilder
	"Return a link for using on the web.
	Always from the top."

	^aBuilder getLinkTop: 'category/', id asString text: name! !

!SMCategory methodsFor: 'testing' stamp: 'gh 12/1/2002 20:24'!
hasSubCategories
	^subCategories isEmptyOrNil not! !

!SMCategory methodsFor: 'testing' stamp: 'gk 8/7/2003 22:31'!
includes: anObject
	"Answer if <anObject> is in this category."

	^objects includes: anObject! !

!SMCategory methodsFor: 'initialize-release' stamp: 'stephaneducasse 2/4/2006 20:38'!
initialize
	super initialize.
	name := summary := url := ''.
	objects := OrderedCollection new! !

!SMCategory methodsFor: 'testing' stamp: 'gh 12/1/2002 19:51'!
isCategory
	^true! !

!SMCategory methodsFor: 'testing' stamp: 'gh 6/27/2002 13:31'!
isTopCategory
	^parent isNil! !

!SMCategory methodsFor: 'accessing' stamp: 'gk 11/17/2003 13:51'!
mandatory
	^mandatory! !

!SMCategory methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
mandatory: aSet
	mandatory := aSet! !

!SMCategory methodsFor: 'accessing' stamp: 'gk 11/17/2003 21:35'!
mandatoryFor: aClass
	"Is this category mandatory for instances of <aClass>?"

	^mandatory ifNil: [false] ifNotNil: [mandatory includes: aClass]! !

!SMCategory methodsFor: 'services' stamp: 'gh 8/5/2002 14:36'!
move: cat toAfter: before
	"Move a category to be after the category <before>."

	subCategories remove: cat.
	before ifNil: [subCategories addFirst: cat] ifNotNil: [subCategories add: cat after: before]! !

!SMCategory methodsFor: 'accessing' stamp: 'gh 8/1/2002 16:54'!
objects
	"Return all objects in this category."

	^objects! !

!SMCategory methodsFor: 'accessing' stamp: 'gk 8/7/2003 23:42'!
packages
	"Return all packages in this category."

	^objects select: [:p | p isPackage]! !

!SMCategory methodsFor: 'accessing' stamp: 'gh 6/27/2002 16:01'!
parent
	^parent! !

!SMCategory methodsFor: 'private' stamp: 'stephaneducasse 2/4/2006 20:38'!
parent: aCategory
	"Change the parent category.
	This method relies on that somebody else
	updates the parent's subCategories collection."
	
	parent := aCategory! !

!SMCategory methodsFor: 'services' stamp: 'gh 8/1/2002 17:30'!
parentsDo: aBlock
	"Run a block for all my parents starting from the top."

	parent ifNotNil: [
		parent parentsDo: aBlock.
		aBlock value: parent]! !

!SMCategory methodsFor: 'printing' stamp: 'gk 1/29/2004 00:14'!
path
	"Return my name with a full path of my
	parent names separated with slashes like:
		'Squeak versions/Squeak3.5' "

	^String streamContents: [:s |
		self parentsDo: [:cat |
			s nextPutAll: cat name; nextPutAll: '/'].
		s nextPutAll: self name]! !

!SMCategory methodsFor: 'printing' stamp: 'gh 8/16/2002 06:04'!
printOn: aStream

	aStream nextPutAll: self class name, ': ', name! !

!SMCategory methodsFor: 'private' stamp: 'gk 11/24/2005 11:34'!
removeDeepFromObjects
	"Remove myself from my objects and then ask
	my subCategories to do the same."

	self removeFromObjects.
	subCategories do: [:cat | cat removeDeepFromObjects]! !

!SMCategory methodsFor: 'private' stamp: 'gh 12/1/2002 20:03'!
removeFromObjects
	"Remove myself from my objects."

	objects copy do: [:obj | obj removeCategory: self]! !

!SMCategory methodsFor: 'private' stamp: 'gh 12/1/2002 20:28'!
removeFromParent
	"Remove me from my parent."

	parent ifNotNil: [parent removeCategory: self]! !

!SMCategory methodsFor: 'private' stamp: 'gh 12/1/2002 20:28'!
removeObject: anObject
	"Remove <anObject> from this category. This should only be called
	from SMCategorizableObject>>removeCategory: to ensure consistency."
	
	^objects remove: anObject! !

!SMCategory methodsFor: 'accessing' stamp: 'gh 6/27/2002 12:25'!
subCategories
	subCategories ifNil: [^#()].
	^subCategories! !

!SMCategory methodsFor: 'printing' stamp: 'gk 11/14/2003 00:13'!
type

	^'Category'! !

!SMCategory methodsFor: 'view' stamp: 'gh 8/1/2002 16:38'!
viewFor: uiObject
	"This is a double dispatch mechanism for multiple views
	for multiple uis."

	^uiObject categoryViewOn: self! !

!SMObject classSide methodsFor: 'instance creation' stamp: 'gk 12/7/2005 13:53'!
newIn: aMap
	"Create a new object in a given map with an UUID to ensure unique identity."

	^(super basicNew) map: aMap id: UUID new! !

!SMObject methodsFor: 'comparing' stamp: 'gk 9/22/2004 23:21'!
<= anSMObject

	^name <= anSMObject name! !

!SMObject methodsFor: 'accessing' stamp: 'gh 3/16/2002 23:12'!
created
	^TimeStamp fromSeconds: created! !

!SMObject methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
created: c
	created := c! !

!SMObject methodsFor: 'accessing' stamp: 'gh 3/16/2002 23:12'!
createdAsSeconds
	^created! !

!SMObject methodsFor: 'deletion' stamp: 'gk 8/8/2003 10:10'!
delete
	"Delete from map."

	map deleteObject: self! !

!SMObject methodsFor: 'printing' stamp: 'gk 11/14/2003 14:22'!
describe: string withBoldLabel: label on: stream
	"Helper method for doing styled text."

	stream withAttribute: (TextEmphasis bold) do: [ stream nextPutAll: label ].
	stream nextPutAll: string; cr! !

!SMObject methodsFor: 'accessing' stamp: 'gh 3/16/2002 23:12'!
id
	^id! !

!SMObject methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
id: anId
	id := anId! !

!SMObject methodsFor: 'initialize-release' stamp: 'stephaneducasse 2/4/2006 20:38'!
initialize
	"Initialize the receiver."

	updated := created := TimeStamp current asSeconds.
	name := summary := url := ''.! !

!SMObject methodsFor: 'testing' stamp: 'gk 6/26/2003 16:47'!
isAccount
	^false! !

!SMObject methodsFor: 'testing' stamp: 'gh 12/1/2002 19:51'!
isCategory
	^false! !

!SMObject methodsFor: 'testing' stamp: 'gh 12/1/2002 19:40'!
isPackage
	^false! !

!SMObject methodsFor: 'testing' stamp: 'gh 12/1/2002 19:51'!
isPackageRelease
	^false! !

!SMObject methodsFor: 'testing' stamp: 'gh 12/1/2002 19:51'!
isResource
	^false! !

!SMObject methodsFor: 'accessing' stamp: 'gh 8/15/2002 08:50'!
map
	^map! !

!SMObject methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
map: aMap
	map := aMap! !

!SMObject methodsFor: 'initialize-release' stamp: 'stephaneducasse 2/4/2006 20:38'!
map: aMap id: anId
	"Initialize the receiver."

	self initialize.
	map := aMap.
	id := anId! !

!SMObject methodsFor: 'accessing' stamp: 'gh 6/26/2002 15:31'!
name
	^name! !

!SMObject methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
name: aName
	name := aName! !

!SMObject methodsFor: 'printing' stamp: 'gk 7/10/2004 03:39'!
printName
	"Return a String identifying receiver without a context.
	Default is name."

	^self name! !

!SMObject methodsFor: 'printing' stamp: 'gk 8/4/2003 11:56'!
printOn: aStream

	aStream nextPutAll: self class name, '[', name, ']'! !

!SMObject methodsFor: 'updating' stamp: 'stephaneducasse 2/4/2006 20:38'!
stampAsUpdated
	"This method should be called whenever the object is modified."

	updated := TimeStamp current asSeconds! !

!SMObject methodsFor: 'accessing' stamp: 'gh 6/26/2002 15:33'!
summary
	^summary! !

!SMObject methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
summary: aString
	summary := aString! !

!SMObject methodsFor: 'printing' stamp: 'gk 11/14/2003 00:11'!
type

	^'Object'! !

!SMObject methodsFor: 'accessing' stamp: 'gh 3/17/2002 01:44'!
updated
	^TimeStamp fromSeconds: updated! !

!SMObject methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
updated: c
	updated := c! !

!SMObject methodsFor: 'accessing' stamp: 'gh 3/17/2002 01:45'!
updatedAsSeconds
	^updated! !

!SMObject methodsFor: 'accessing' stamp: 'gh 6/26/2002 15:36'!
url
	^url! !

!SMObject methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
url: aString
	url := aString! !

!SMObject methodsFor: 'accessing' stamp: 'gh 11/27/2002 12:21'!
userInterface
	"Return the object that we use for interacting with the user."

	^SMUtilities! !

!SMObject methodsFor: 'private' stamp: 'gk 5/5/2006 02:07'!
withId: aUUIDString in: aCollection
	"Return the object with the corresponding id
	and nil if not found."

	| uuid |
	uuid := UUID fromString: aUUIDString.
	^aCollection detect: [:o | o id = uuid ] ifNone: [nil]! !

Object subclass: #SMSqueakMap
	instanceVariableNames: 'packages accounts objects categories dir adminPassword fileCache users mutex isDirty checkpointNumber silent registry'
	classVariableNames: 'DefaultMap ServerList'
	poolDictionaries: ''
	category: 'SMBase-domain'!

!SMSqueakMap commentStamp: '<historical>' prior: 0!
SqueakMap is a Squeak meta catalog, primarily a catalog of all available Squeak packages.
SMSqueakMap is the class for the domain model.

One master instance lives on a server on the Internet at map1.squeakfoundation.org (but there are fallback servers too). Then each Squeak connected to the Internet has one instance that is synchronized against the master. This way every user can have an updated catalog of all available Squeak software on the planet locally on their machine. :-)

Typically you only need one instance of SMSqueakMap per image and it is held by a singleton class variable reached with "SMSqueakMap default". If it is not there it will then be created together with it's own directory on disk by default the "sm" directory created in your default directory.

Synching with the master is the only action that affects the map so you can actually use the same map from multiple images (they will by default use the same snapshot files if the images have the same default directory) and multiple tools (SMLoader, SMBrowser or others).

An instance of SMSqueakMap contains instances of SMPackage which which represent Squeak packages and SMCategories which are centrally registered values that can be included as attributes in an SMPackage. An example of such a category would be different kinds of licenses, topic etc. etc.

It also contains instances of SMAccount which are registered Squeak developers, the maintainers of the packages.

Finally SMSqueakMap also refers to a registry object (SMInstallationRegistry) which records what packages and releases of them have been installed in the image
--------------------
"Simplest use of SMSqueakMap - this will create a map if you don't have one and open a simple UI""
SMLoader open

Use these doits to play "hands on" with a SqueakMap.

"Creating another SqueakMap in directory 'slavemap' instead of default 'sm'"
Smalltalk at: #AnotherSqueakMap put: (SMSqueakMap newIn: 'slavemap')

"Update the default map by synching it with the master - check Transcript for some numbers."
SMSqueakMap default loadUpdates

"If the map is broken in some way, reload it from disk"
SMSqueakMap default reload

"Clear out the contents of the map to save image space, this does not remove the registry."
SMSqueakMap default purge
!

!SMSqueakMap classSide methodsFor: 'changelog replay' stamp: 'stephaneducasse 2/4/2006 20:38'!
askUser
	"Ask user about how to handle a replayed installation note
	when there is no current SqueakMap in the image."

	| choice |
	[choice := UIManager default chooseFrom: #('Yes' 'No' 'More info')
				title:
'There is no SqueakMap in this image,
do you wish to create/recreate it? (typical answer is Yes)' .
			choice = 3] whileTrue: [self inform:
'When packages are installed using SqueakMap a little mark is made
in the change log. When an image is reconstructed from the changelog
these marks are intended to keep your map informed about what packages
are installed. You probably already have a map on disk which will automatically be
reloaded if you choose ''Yes'', otherwise an new empty map will be created.
If you choose ''No'', it will only result in that SqueakMap will not know that this package
is installed in your image.
If you are still unsure - answer ''Yes'' since that is probably the best.'].
	^choice = 1! !

!SMSqueakMap classSide methodsFor: 'bootstrap upgrade' stamp: 'gk 3/31/2006 01:05'!
bootStrap
	"Bootstrap upgrade. Only used when SqueakMap itself is too old to
	communicate with the server. This relies on the existence of a package
	called SqueakMap that is a .st loadscript. The loadscript needs to do its
	own changeset management."

	| server url |
	server := self findServer.
	server ifNotNil: ["Ok, found a SqueakMap server"
		url := (('http://', server, '/packagebyname/squeakmap/downloadurl')
				asUrl retrieveContents content) asUrl.
		(url retrieveContents content unzipped readStream)
				fileInAnnouncing: 'Upgrading SqueakMap...']! !

!SMSqueakMap classSide methodsFor: 'instance creation' stamp: 'stephaneducasse 2/4/2006 20:38'!
clear
	"Clear out the model in the image. This will forget
	about what packages are installed and what versions.
	The map is itself on disk though and will be reloaded.
	
	If you only want to reload the map and not forget about
	installed packages then use 'SMSqueakMap default reload'.

	If you want to throw out the map perhaps when shrinking
	an image, then use 'SMSqueakMap default purge'."

	"SMSqueakMap clear"

	DefaultMap := nil! !

!SMSqueakMap classSide methodsFor: 'instance creation' stamp: 'stephaneducasse 2/4/2006 20:38'!
default
	"Return the default map, create one if missing."

	"SMSqueakMap default"

	^DefaultMap ifNil: [DefaultMap := self new]! !

!SMSqueakMap classSide methodsFor: 'instance creation' stamp: 'gh 10/21/2002 15:05'!
defaultNoCreate
	"Return the default map or nil if there is none."

	"SMSqueakMap defaultNoCreate"

	^DefaultMap! !

!SMSqueakMap classSide methodsFor: 'discarding' stamp: 'gk 7/14/2003 15:04'!
discardSM
	"Discard SqueakMapBase. All the map state is kept in
	the class var DefaultMap in SMSqueakMap and is thus also removed."

	"SMSqueakMap discardSM"

	SystemOrganization removeCategoriesMatching: 'SM-domain'.! !

!SMSqueakMap classSide methodsFor: 'server detection' stamp: 'stephaneducasse 2/4/2006 20:38'!
findServer
	"Go through the list of known master servers, ping 
	each one using simple http get on a known 'ping'-url 
	until one responds return the server name. 
	If some servers are bypassed we write that to Transcript. 
	If all servers are down we inform the user and return nil."

	| notAnswering deafServers |
	Socket initializeNetwork.
	notAnswering := OrderedCollection new.
	Cursor wait
		showWhile: [ServerList
				do: [:server | (self pingServer: server)
						ifTrue: [notAnswering isEmpty
								ifFalse: [deafServers := String
												streamContents: [:str | notAnswering
														do: [:srvr | str nextPutAll: srvr printString;
																 nextPut: Character cr]].
									Transcript show: ('These SqueakMap master servers did not respond:\' , deafServers , 'Falling back on ' , server printString , '.') withCRs].
							^ server]
						ifFalse: [notAnswering add: server]]].
	deafServers := String
				streamContents: [:str | notAnswering
						do: [:srvr | str nextPutAll: srvr printString;
								 nextPut: Character cr]].
	self error: ('All SqueakMap master servers are down:\' , deafServers , '\ \Can not update SqueakMap...') withCRs.
	^ nil! !

!SMSqueakMap classSide methodsFor: 'class initialization' stamp: 'gk 5/5/2006 02:07'!
initialize
	"Initialize the list of master servers.
	The last one is for debugging/development."

	"self initialize"

	ServerList := #('map.squeak.org' 'map1.squeakfoundation.org' 'map2.squeakfoundation.org' '127.0.0.1:8080')! !

!SMSqueakMap classSide methodsFor: 'instance creation' stamp: 'gh 10/22/2002 17:41'!
new
	"Create a new server in a new directory
	under the default directory called 'sm'."

	^super new initializeOn: 'sm'! !

!SMSqueakMap classSide methodsFor: 'instance creation' stamp: 'gh 8/15/2002 07:02'!
newIn: directoryName
	"Create a new server in a new directory
	under the default directory called <directoryName>."

	^super new initializeOn: directoryName! !

!SMSqueakMap classSide methodsFor: 'changelog replay' stamp: 'stephaneducasse 2/4/2006 20:38'!
noteInstalledPackage: uuidString version: version
	"We are replaying a change that indicates that a package
	was just installed. If there is a map we let it record this,
	otherwise we ask the user if we should create/recreate the map."

	| choice |
	DefaultMap
		ifNotNil: [DefaultMap noteInstalledPackage: uuidString version: version]
		ifNil: [
			[choice := UIManager default chooseFrom: #('Yes' 'No' 'More info')
				title:
'There is no SqueakMap in this image,
do you wish to create/recreate it? (typical answer is Yes)' .
			choice = 3] whileTrue: [self inform:
'When packages are installed using SqueakMap a little mark is made
in the change log. When an image is reconstructed from the changelog
these marks are intended to keep your map informed about what packages
are installed. You probably already have a map on disk which will automatically be
reloaded if you choose ''Yes'', otherwise an new empty map will be created.
If you choose ''No'', it will only result in that SqueakMap will not know that this package
is installed in your image.
If you are still unsure - answer ''Yes'' since that is probably the best.'].
			choice = 1
				ifTrue:[self default noteInstalledPackage: uuidString version: version]]! !

!SMSqueakMap classSide methodsFor: 'changelog replay' stamp: 'gk 7/29/2003 01:01'!
noteInstalledPackage: uuidString version: version atSeconds: sec number: num
	"We are replaying a change that indicates that a package
	was just installed. If there is a map we let it record this,
	otherwise we ask the user if we should create/recreate the map."

	DefaultMap
		ifNotNil: [DefaultMap noteInstalledPackage: uuidString version: version
					atSeconds: sec number: num]
		ifNil: [
			self askUser
				ifTrue:[self default noteInstalledPackage: uuidString version: version
							atSeconds: sec number: num]]! !

!SMSqueakMap classSide methodsFor: 'changelog replay' stamp: 'gk 11/17/2003 01:20'!
noteInstalledPackageWithId: uuidString autoVersion: version atSeconds: sec number: num
	"We are replaying a change that indicates that a package release
	was just installed using SM2. If there is a map we let it record this,
	otherwise we ask the user if we should create/recreate the map."

	DefaultMap
		ifNotNil: [DefaultMap noteInstalledPackageWithId: uuidString autoVersion: version
					atSeconds: sec number: num]
		ifNil: [
			self askUser
				ifTrue:[self default noteInstalledPackageWithId: uuidString autoVersion: version
							atSeconds: sec number: num]]! !

!SMSqueakMap classSide methodsFor: 'changelog replay' stamp: 'gk 11/17/2003 01:18'!
noteInstalledPackageWithId: uuidString version: version atSeconds: sec number: num
	"We are replaying a change that indicates that a package release
	was just installed using SM2. If there is a map we let it record this,
	otherwise we ask the user if we should create/recreate the map."

	DefaultMap
		ifNotNil: [DefaultMap noteInstalledPackageWithId: uuidString version: version
					atSeconds: sec number: num]
		ifNil: [
			self askUser
				ifTrue:[self default noteInstalledPackageWithId: uuidString version: version
							atSeconds: sec number: num]]! !

!SMSqueakMap classSide methodsFor: 'server detection' stamp: 'gk 5/5/2006 02:07'!
pingServer: aServerName
	"Check if the SqueakMap server is responding.
	For an old image we first make sure the name resolves -
	the #httpGet: had such a long timeout (and hanging?)
	for resolving the name."

	| url answer |
	"Only test name lookup first if image is before the network rewrite,
	after the rewrite it works."
	[(SystemVersion current highestUpdate < 5252)
		ifTrue: [NetNameResolver addressForName: (aServerName upTo: $:) timeout: 5].
	url := 'http://', aServerName, '/ping'.
	answer := HTTPSocket httpGet: url]
				on: Error do: [:ex | ^false].
	^answer isString not and: [answer contents = 'pong']! !

!SMSqueakMap classSide methodsFor: 'migration' stamp: 'stephaneducasse 2/4/2006 20:38'!
recreateInstalledPackagesFromChangeLog
	"Clear and recreate the Dictionary with information on installed packages.

	NOTE: This takes some time to run and will only find packages installed using SM
	and since the last changelog condense.

	For packages installed prior to SqueakMap 1.07 there is no timestamp nor counter
	logged. These packages will be given the time of the replay and a separate count
	(from -10000 upwards) maintaining correct order of installation."

	"SMSqueakMap recreateInstalledPackagesFromChangeLog"

	| changesFile chunk |
	SMSqueakMap default clearInstalledPackages.
	changesFile := (SourceFiles at: 2) readOnlyCopy.
	[changesFile atEnd]
		whileFalse: [
			chunk := changesFile nextChunk.
			((chunk beginsWith: '"Installed') and: [
				(chunk indexOfSubCollection: 'SMSqueakMap noteInstalledPackage:'
					startingAt: 10) > 0])
				ifTrue: [Compiler evaluate: chunk logged: false]].
	changesFile close! !

!SMSqueakMap classSide methodsFor: 'constants' stamp: 'gk 4/4/2006 01:13'!
version
	"This is the protocol version number used for clients to decide if
	they need to update SMSqueakMap before synching with
	the master. In short - only increase this if changes have made
	the clients incompatible so that they need to be updated.

	2.0: Removed Module stuff and added Package releases.
	2.1: Various changes/additions and class shape changes.
	2.2: Various 3.9 related fixes and bug fix in segment compression etc."

	^'2.2'! !

!SMSqueakMap methodsFor: 'queries' stamp: 'gk 8/4/2003 11:50'!
accountForEmail: email
	"Find account given email."

	^self accounts detect: [:a | a email = email] ifNone: [nil]! !

!SMSqueakMap methodsFor: 'queries' stamp: 'stephaneducasse 2/4/2006 20:38'!
accountForName: name
	"Find account given full name. Disregarding case
	and allows up to 2 different characters.
	Size must match though, someone else can be smarter -
	this is just for migrating accounts properly."

	| lowerName size aName |
	lowerName := name asLowercase.
	size := lowerName size.
	^self accounts
		detect: [:a |
			aName := a name asLowercase.
			(aName size = size) and: [| errors |
				errors := 0.
				aName with: lowerName do: [:c1 :c2 |
					c1 ~= c2 ifTrue: [errors := errors + 1]].
				errors < 3
			]]
		ifNone: [nil]
		! !

!SMSqueakMap methodsFor: 'queries' stamp: 'gk 7/30/2003 14:11'!
accountForUsername: username
	"Find account given username. The username used
	is the developer initials of the account."

	^self users at: username ifAbsent: [nil]! !

!SMSqueakMap methodsFor: 'queries' stamp: 'stephaneducasse 2/4/2006 20:38'!
accountWithId: anIdString 
	"Look up an account. Return nil if missing.
	Raise error if it is not an account."

	| account |
	account := self objectWithId: anIdString.
	account ifNil: [^nil].
	account isAccount ifTrue:[^account].
	self error: 'UUID did not map to a account.'! !

!SMSqueakMap methodsFor: 'queries' stamp: 'gk 2/7/2004 14:55'!
accountWithName: aName
	"Look up an account by name. Return nil if missing."

	^self accounts values detect: [:a | a name = aName ] ifNone: [nil]! !

!SMSqueakMap methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
accounts
	"Lazily maintain a cache of all known account objects."

	accounts ifNotNil: [^accounts].
	accounts := objects select: [:o | o isAccount].
	^accounts! !

!SMSqueakMap methodsFor: 'queries' stamp: 'gk 12/3/2004 12:03'!
accountsByInitials
	"Return the accounts sorted by the developer initials."

	^self accounts asSortedCollection: [:x :y | x initials caseInsensitiveLessOrEqual: y initials]! !

!SMSqueakMap methodsFor: 'queries' stamp: 'gk 12/3/2004 12:03'!
accountsByName
	"Return the accounts sorted by their name."

	^self accounts asSortedCollection: [:x :y | x name caseInsensitiveLessOrEqual: y name].! !

!SMSqueakMap methodsFor: 'public-master' stamp: 'gk 11/17/2003 10:30'!
addCategory: category inObject: object
	"Add a category in an object."

	^object addCategory: category
! !

!SMSqueakMap methodsFor: 'transactions' stamp: 'gk 10/10/2003 11:30'!
addDirty: anSMObject
	"Add the SMObject to the dirty list making
	sure it gets committed when transaction ends."

"In first SM2 version we do nothing"

"	dirtyList add: anSMObject"! !

!SMSqueakMap methodsFor: 'public-master' stamp: 'gk 2/7/2004 14:56'!
addObject: anSMObject 
	"Add a new object, only if not already added."

	(self object: anSMObject id) ifNil: [
		self transaction: [self newObject: anSMObject]]! !

!SMSqueakMap methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
adminPassword: aString
	"We store the password as a SHA hash so that we can let the slave maps
	have it too."

	adminPassword := SecureHashAlgorithm new hashMessage: aString! !

!SMSqueakMap methodsFor: 'public-packages' stamp: 'gk 11/14/2003 12:06'!
allPackages
	"Answer all packages."

	^self packages! !

!SMSqueakMap methodsFor: 'public-packages' stamp: 'gh 12/1/2002 19:53'!
availablePackages
	"Answer all packages that are old or not installed."

	^self packages select: [:package | package isAvailable]! !

!SMSqueakMap methodsFor: 'accessing' stamp: 'btr 5/28/2003 02:16'!
cache
	^ fileCache! !

!SMSqueakMap methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
categories
	"Lazily maintain a cache of all known category objects."

	categories ifNotNil: [^categories].
	categories := objects select: [:o | o isCategory].
	^categories! !

!SMSqueakMap methodsFor: 'queries' stamp: 'stephaneducasse 2/4/2006 20:38'!
categoryWithId: anIdString 
	"Look up a category. Return nil if missing.
	Raise error if it is not a category."

	| cat |
	cat := self objectWithId: anIdString.
	cat ifNil: [^nil].
	cat isCategory ifTrue:[^cat].
	self error: 'UUID did not map to a category.'! !

!SMSqueakMap methodsFor: 'queries' stamp: 'stephaneducasse 2/4/2006 20:38'!
categoryWithNameBeginning: aString
	"Look up a category beginning with <aString>. Return nil if missing.
	We return the shortest matching one. We also strip out spaces and
	ignore case in both <aString> and the names."

	| candidates shortest answer searchString |
	searchString := (aString asLowercase) copyWithout: Character space.
	candidates := self categories select: [:cat |
		((cat name asLowercase) copyWithout: Character space)
			beginsWith: searchString ].
	shortest := 1000.
	candidates do: [:ca |
		ca name size < shortest ifTrue:[answer := ca. shortest := ca name size]].
	^answer	! !

!SMSqueakMap methodsFor: 'public-master' stamp: 'gh 11/28/2002 20:44'!
changeCategoriesTo: newCategories inObject: object
	"Remove or add categories in an object such that
	it belongs to the categories in <newCategories>.
	Logs the changes."

	newCategories do: [:cat |
		(object hasCategory: cat)
			ifFalse:[self addCategory: cat inObject: object]].
	object categories do: [:cat |
		(newCategories includes: cat)
			ifFalse: [self removeCategory: cat inObject: object]]
! !

!SMSqueakMap methodsFor: 'queries' stamp: 'gk 5/5/2006 02:07'!
check
	"Sanity checks."

	"SMSqueakMap default check"
	
	(((self packages inject: 0 into: [:sum :p | sum := sum + p releases size]) +
	self accounts size +
	self packages size +
	self categories size) = SMSqueakMap default objects size)
		ifFalse: [self error: 'Count inconsistency in map'].
	
	objects do: [:o |
		o map == self
			ifFalse: [self error: 'Object with wrong map']].
	self packages do: [:p |
		(p releases allSatisfy: [:r | r map == self])
			ifFalse: [self error: 'Package with release pointing to wrong map']].
		
	self packageReleases do: [:r |
		r package map == self ifFalse: [self error: 'Release pointing to package in wrong map']]! !

!SMSqueakMap methodsFor: 'private' stamp: 'gh 10/31/2002 11:59'!
checkVersion: string
	"Check the content for a SqueakMap version conflict notification.
	Return true if no conflict is reported, otherwise ask user if we
	should upgrade SqueakMap using the bootstrap method."

	(string beginsWith: 'Server version:')
		ifTrue:[(self confirm: ('The SqueakMap master server is running another version (', (string last: (string size - 15)), ') than the client (', SMSqueakMap version, ').
You need to upgrade the SqueakMap package, would you like to do that now?'))
			ifTrue: [self class bootStrap. ^false]
			ifFalse: [^false]
	].
	^true! !

!SMSqueakMap methodsFor: 'accessing' stamp: 'gk 10/21/2003 23:02'!
checkpointNumber

	^checkpointNumber! !

!SMSqueakMap methodsFor: 'private' stamp: 'stephaneducasse 2/4/2006 20:38'!
clearCaches
	"Clear the caches."

	packages := accounts := users := categories := nil
! !

!SMSqueakMap methodsFor: 'private' stamp: 'stephaneducasse 2/4/2006 20:38'!
clearCachesFor: anObject 
	"Clear the valid caches."

	anObject isPackage ifTrue:[packages := nil].
	anObject isAccount ifTrue:[accounts := users := nil].
	anObject isCategory ifTrue:[categories := nil]
! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'gk 8/3/2004 13:24'!
clearInstalledPackageWithId: aPackageId
	"Clear the fact that any release of this package is installed.
	Can be used even when the map isn't loaded."

	^self registry clearInstalledPackageWithId: aPackageId! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'gk 8/3/2004 13:24'!
clearInstalledPackages
	"Simply clear the dictionary with information on installed packages.
	Might be good if things get corrupted etc. Also see
	SMSqueakMap class>>recreateInstalledPackagesFromChangeLog"

	^self registry clearInstalledPackages! !

!SMSqueakMap methodsFor: 'private' stamp: 'stephaneducasse 2/4/2006 20:38'!
clearUsernames
	"Clear the username cache."

	users := nil! !

!SMSqueakMap methodsFor: 'checkpoints' stamp: 'gk 3/31/2006 20:58'!
compressFile: aFileStream
	"Shamelessly copied and modified from StandardFileStream>>compressFile."
	
	| zipped buffer |
	aFileStream binary.
	zipped := StandardFileStream newFileNamed: (self directory fullNameFor: (aFileStream name, 'gz')).
	zipped binary; setFileTypeToObject.
	"Type and Creator not to be text, so can be enclosed in an email"
	zipped := GZipWriteStream on: zipped.
	buffer := ByteArray new: 50000.
	[[aFileStream atEnd] whileFalse: [
		zipped nextPutAll: (aFileStream nextInto: buffer)]]
		ensure: [zipped close. aFileStream close].
	self directory deleteFileNamed: aFileStream name! !

!SMSqueakMap methodsFor: 'private' stamp: 'gk 5/5/2006 02:07'!
copyFrom: aMap
	"Copy all relevant info from the other map."

	objects := aMap objects.
	objects do: [:o | o map: self].
	accounts := users := packages := categories := nil.
	checkpointNumber := aMap checkpointNumber! !

!SMSqueakMap methodsFor: 'checkpoints' stamp: 'gk 10/10/2003 14:47'!
createCheckpoint
	"Export a new checkpoint of me using an ImageSegment."

	^self createCheckpointNumber: 
		(self nextFileNameForCheckPoint findTokens: '.') second asNumber.
! !

!SMSqueakMap methodsFor: 'checkpoints' stamp: 'stephaneducasse 2/4/2006 20:38'!
createCheckpointNumber: number
	"Export me using an ImageSegment.
	This is used for checkpointing the map on disk
	in a form that can be brought into an independent image.
	We do not overwrite older versions, since using ImageSegments
	is an intermediate hack anyway we don't care about the disk waste!!
	Sidenote: Some refactoring was needed to produce a .gz file directly so
	I didn't bother."

	| is fname stream oldMutex |
	fname := self filename, '.', number asString, '.s'.
	(self directory fileExists: fname) ifTrue: [self error: 'Checkpoint already exists!!'].
	stream := StandardFileStream newFileNamed: (self directory fullNameFor: fname).
	checkpointNumber := number.
	oldMutex := mutex.
	mutex := nil. self clearCaches.
	[is := ImageSegment new.
	is copyFromRoots: (Array with: self) sizeHint: 1000000 areUnique: true.
	is writeForExportOn: stream.
	self compressFile: (StandardFileStream oldFileNamed: (self directory fullNameFor: fname)).
	isDirty := false]
		ensure: [mutex := oldMutex].
	^is! !

!SMSqueakMap methodsFor: 'private' stamp: 'gk 8/8/2003 19:10'!
deleteObject: anObject 
	"Delete an object, remove it from objects.
	This method is called from the #delete method of
	anObject so it will take care of the rest of the
	cleaning up. Clear the valid caches."

	objects removeKey: anObject id.
	self clearCachesFor: anObject
! !

!SMSqueakMap methodsFor: 'accessing' stamp: 'gk 5/22/2004 21:56'!
directory
	"Return the subdirectory that SqueakMap uses."
	
	(FileDirectory default directoryExists: dir)
		ifFalse:[FileDirectory default createDirectory: dir].
	^FileDirectory default directoryNamed: dir! !

!SMSqueakMap methodsFor: 'private' stamp: 'gk 8/4/2003 15:22'!
emailOccupied: aUsername
	"Return true if email already taken."

	^(self accountForEmail: aUsername) notNil! !

!SMSqueakMap methodsFor: 'checkpoints' stamp: 'gk 10/10/2003 12:07'!
extension
	^'sgz'! !

!SMSqueakMap methodsFor: 'checkpoints' stamp: 'gk 10/10/2003 12:07'!
filename
	^'map'! !

!SMSqueakMap methodsFor: 'checkpoints' stamp: 'gk 3/31/2006 09:30'!
getLastCheckpointWithFilename
	"Return a readstream on a fresh checkpoint gzipped imagesegment.
	First we check if we are dirty and must create a new checkpoint.
	The filename is tacked on at the end so that the checkpoint number
	can be used on the client side too."

	| directory fname |
	isDirty ifTrue: [self createCheckpoint].
	directory := self directory.
	fname := self lastCheckpointFilename.
	fname ifNil: [self error: 'No checkpoint available'].
	^((StandardFileStream oldFileNamed: (directory fullNameFor: fname))
		contentsOfEntireFile), ':', fname! !

!SMSqueakMap methodsFor: 'initialize-release' stamp: 'gk 5/5/2006 02:07'!
initializeOn: directoryName
	"Create the local directory for SqueakMap."

	dir := directoryName.
	(FileDirectory default directoryExists: dir)
		ifFalse:[FileDirectory default createDirectory: dir].
	fileCache := SMFileCache newFor: self.
	checkpointNumber := 1! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'stephaneducasse 2/4/2006 20:38'!
installPackage: aPackage
	"Install the package.

	Note: This method should not be used anymore, better
	to specify a specific release."

	| rel |
	rel := aPackage lastPublishedReleaseForCurrentSystemVersion
			ifNil: [self error: 'No published release for this system version found to install.'].
	^self installPackageRelease: rel! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'stephaneducasse 2/4/2006 20:38'!
installPackage: aPackage autoVersion: version
	"Install the release <version> of <aPackage.
	<version> is the automatic version name."

	| r |
	r := aPackage releaseWithAutomaticVersionString: version.
	r ifNil: [self error: 'No package release found with automatic version ', version].
	^self installPackageRelease: r! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'stephaneducasse 2/4/2006 20:38'!
installPackageNamed: aString
	"Install the last published release
	for this Squeak version of the package with a name
	beginning with aString (see method comment
	of #packageWithNameBeginning:).

	Note: This method should not be used anymore.
	Better to specify a specific release."

	| p |
	p := self packageWithNameBeginning: aString.
	p ifNil: [self error: 'No package found with name beginning with ', aString].
	^self installPackage: p! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'stephaneducasse 2/4/2006 20:38'!
installPackageNamed: aString autoVersion: version
	"Install the release <version> of the package with a name
	beginning with aString (see method comment
	of #packageWithNameBeginning:). <version> is the
	automatic version name."

	| p r |
	p := self packageWithNameBeginning: aString.
	p ifNil: [self error: 'No package found with name beginning with ', aString].
	r := p releaseWithAutomaticVersionString: version.
	r ifNil: [self error: 'No package release found with automatic version ', version].
	^self installPackageRelease: r! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'gk 7/14/2004 15:57'!
installPackageRelease: aPackageRelease
	"Install the given package release, no checks made."

	(SMInstaller forPackageRelease: aPackageRelease) install! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'stephaneducasse 2/4/2006 20:38'!
installPackageReleaseWithId: anUUIDString
	"Look up and install the given release."

	| r |
	r := self packageReleaseWithId: anUUIDString.
	r ifNil: [self error: 'No package release available with id: ''', anUUIDString, ''''].
	^self installPackageRelease: r! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'stephaneducasse 2/4/2006 20:38'!
installPackageWithId: anUUIDString
	"Look up and install the latest release of the given package.	

	Note: This method should not be used anymore.
	Better to specify a specific release."

	| package |
	package := self packageWithId: anUUIDString.
	package ifNil: [self error: 'No package available with id: ''', anUUIDString, ''''].
	^self installPackage: package! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'stephaneducasse 2/4/2006 20:38'!
installPackageWithId: anUUIDString autoVersion: version
	"Install the release <version> of the package with id <anUUIDString>.
	<version> is the automatic version name."

	| p |
	p := self packageWithId: anUUIDString.
	p ifNil: [self error: 'No package available with id: ''', anUUIDString, ''''].
	^self installPackage: p autoVersion: version! !

!SMSqueakMap methodsFor: 'public-packages' stamp: 'gh 12/1/2002 19:54'!
installableAndNotInstalledPackages
	"Answer all installable but not installed packages."

	^self packages select: [:package | package isInstallableAndNotInstalled]! !

!SMSqueakMap methodsFor: 'public-packages' stamp: 'gh 12/1/2002 19:54'!
installablePackages
	"Answer all packages that can be (auto)installed -
	we have installers that can install them."

	^self packages select: [:package | package isInstallable]! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'gk 1/18/2004 15:29'!
installedPackageReleases
	"Answer all package releases that we know are installed.
	Lazily initialize. The Dictionary contains the installed packages
	using their UUIDs as keys and the version string as the value."

	^self installedPackages collect: [:p | self installedReleaseOf: p]! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'gk 8/3/2004 13:24'!
installedPackages
	"Answer all packages that we know are installed."

	^self registry installedPackages! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'gk 8/3/2004 13:24'!
installedPackagesDictionary
	"Access the dictionary directly. The UUID of the installed package is the key.
	The value is an OrderedCollection of Arrays.
	The arrays have the smartVersion of the package, the time of the
	installation in seconds and the sequence number (installCounter)."

	^self registry installedPackagesDictionary! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'gk 8/3/2004 13:24'!
installedPackagesDictionary: aDict
	"Set dictionary directly."

	^self registry installedPackagesDictionary: aDict! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'gk 8/3/2004 13:23'!
installedReleaseOf: aPackage
	"If the package is installed, return the release.
	Otherwise return nil. SM2 stores the version as
	an Association to be able to distinguish it."

	^self registry installedReleaseOf: aPackage! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'gk 8/3/2004 13:24'!
installedVersionOf: aPackage
	"If the package is installed, return the version as a String.
	If it is a package installed during SM1 it will return the manual version String,
	for SM2 it returns the automatic version as a String.
	If package is not installed - return nil. If you want it to work without the map loaded you
	should instead use #installedVersionOfPackageWithId:."

	^self registry installedVersionOf: aPackage! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'gk 8/3/2004 13:24'!
installedVersionOfPackageWithId: anId
	"If the package is installed, return the automatic version or version String.
	Otherwise return nil. This can be used without the map loaded."

	^self registry installedVersionOfPackageWithId: anId! !

!SMSqueakMap methodsFor: 'checkpoints' stamp: 'gk 10/10/2003 15:02'!
isCheckpointAvailable
	"Check that there is an 'sm' directory
	and that it contains at least one checkpoint."

	[^self lastCheckpointFilename notNil] on: Error do: [:ex | ^false]! !

!SMSqueakMap methodsFor: 'transactions' stamp: 'gk 11/4/2003 00:19'!
isDirty
	"Is the map modified but not yet checkpointed to disk?"

	^isDirty! !

!SMSqueakMap methodsFor: 'checkpoints' stamp: 'gk 10/10/2003 14:21'!
lastCheckpointFilename
	"Return the filename for the newest checkpoint."

	^self directory lastNameFor: self filename extension: self extension! !

!SMSqueakMap methodsFor: 'checkpoints' stamp: 'gk 10/10/2003 14:52'!
lastCheckpointNumberOnDisk
	"Return the last checkpoint number on disk."

	^(self nextFileNameForCheckPoint findTokens: '.') second asNumber - 1! !

!SMSqueakMap methodsFor: 'public' stamp: 'gh 10/21/2002 14:08'!
loadFull
	"Go through the list of known master servers, ping
	each one using simple http get on a known 'ping'-url
	until one responds and then load the full map from it."
 
	self loadUpdatesFull: true! !

!SMSqueakMap methodsFor: 'private' stamp: 'gk 5/5/2006 02:07'!
loadFullFrom: aServerName
	"Contact the SqueakMap at the url <aSqueakMapUrl>
	and load a full map from scratch."

	| url  zipped |
	url := 'http://', aServerName, '/loadgz?mapversion=', SMSqueakMap version, '&checkpoint=', checkpointNumber asString.
	Transcript show: 'Fetch: ', (Time millisecondsToRun: [ zipped := (HTTPSocket httpGet: url) contents]) asString, ' ms';cr.
	Transcript show: 'Size: ', zipped size asString, ' bytes';cr.
	((self checkVersion: zipped) and: [zipped ~= 'UPTODATE'])
		ifTrue:[
			Transcript show: 'Save checkpoint to disk: ', (Time millisecondsToRun: [
			self saveCheckpoint: zipped]) asString, ' ms';cr.
			Transcript show: 'Full reload from checkpoint: ', (Time millisecondsToRun: [
			self reload]) asString, ' ms';cr.]! !

!SMSqueakMap methodsFor: 'public' stamp: 'gk 11/17/2003 23:16'!
loadUpdates
	"Go through the list of known master servers, ping
	each one using simple http get on a known 'ping'-url
	until one responds and then load updates from it."

	"SM2 starts with using full always"

	self loadFull! !

!SMSqueakMap methodsFor: 'private' stamp: 'stephaneducasse 2/4/2006 20:38'!
loadUpdatesFull: full
	"Find a server and load updates from it."
 
	| server |
	server := self class findServer.
	server ifNotNil: [
		self synchWithDisk.
		full ifTrue: [self loadFullFrom: server]
			ifFalse:[self error: 'Not supported yet!!'."self loadUpdatesFrom: server"]]! !

!SMSqueakMap methodsFor: 'private' stamp: 'gk 11/17/2003 22:57'!
mandatoryCategoriesFor: aClass
	"Return the categories that are mandatory for instances of <aClass>."

	^self categories select: [:c | c mandatoryFor: aClass]! !

!SMSqueakMap methodsFor: 'private-installation' stamp: 'gk 8/3/2004 13:24'!
markInstalled: uuid version: version time: time counter: num
	"Private. Mark the installation. SM2 uses an Association
	to distinguish the automatic version from old versions."

	^self registry markInstalled: uuid version: version time: time counter: num! !

!SMSqueakMap methodsFor: 'public-master' stamp: 'gk 11/17/2003 14:56'!
moveCategory: category toAfter: categoryBefore inParent: parent
	"Move a category to be listed after <categoryBefore> in <parent>."

	parent move: category toAfter: categoryBefore.
	^category
	
! !

!SMSqueakMap methodsFor: 'public-master' stamp: 'gk 5/22/2004 22:22'!
moveCategory: category toParent: parentCategory
	"Move a category into another parent category."

	parentCategory
		ifNil: [category parent: nil]
		ifNotNil: [parentCategory addCategory: category].
	^category
	
! !

!SMSqueakMap methodsFor: 'transactions' stamp: 'stephaneducasse 2/4/2006 20:38'!
mutex
	"Lazily initialize the Semaphore."

	^mutex ifNil: [mutex := Semaphore forMutualExclusion]! !

!SMSqueakMap methodsFor: 'private' stamp: 'gk 8/3/2003 23:23'!
newAccount
	"Create a new account."

	^SMAccount newIn: self! !

!SMSqueakMap methodsFor: 'public-master' stamp: 'stephaneducasse 2/4/2006 20:38'!
newAccount: name username: username email: email
	"Create an account. Checking for previous account should already have been done.
	To add the account to the map, use SMSqueakMap>>addObject:"

	| account |
	account := self newAccount
					name: name;
					initials: username;
					email: email.
	^account
	
! !

!SMSqueakMap methodsFor: 'private' stamp: 'gk 9/26/2003 00:04'!
newObject: anSMObject 
	"Add an SMObject to me. Clear the valid caches."

	self addDirty: anSMObject.
	self clearCachesFor: anSMObject.
	^objects at: anSMObject id put: anSMObject! !

!SMSqueakMap methodsFor: 'checkpoints' stamp: 'gk 10/10/2003 12:12'!
nextFileNameForCheckPoint
	"Return the next available filename for a checkpoint."

	^self directory nextNameFor: self filename extension: self extension! !

!SMSqueakMap methodsFor: 'public-packages' stamp: 'gh 12/1/2002 19:54'!
notInstalledPackages
	"Answer all packages that are not installed."

	^self packages reject: [:package | package isInstalled]! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'gk 11/18/2003 17:49'!
noteInstalled: aPackageRelease
	"The package release was just successfully installed using SM.
	This is the method being called by SM upon a successful installation.

	We record this in our Dictionary of installed package releases
	and log a 'do it' to mark this in the changelog.
	The map used is the default map."

	^self noteInstalledPackageWithId: aPackageRelease package id asString
		autoVersion: aPackageRelease automaticVersion
		name: aPackageRelease package name! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'gk 11/18/2003 17:50'!
noteInstalledPackage: aPackage autoVersion: aVersion
	"Mark that the package release was just successfully installed.
	Can be used to inform SM of an installation not been done using SM."

	
^self noteInstalledPackageWithId: aPackage id asString
		autoVersion: aVersion
		name: aPackage name! !

!SMSqueakMap methodsFor: 'private-installation' stamp: 'gk 8/3/2004 13:24'!
noteInstalledPackage: uuidString version: version
	"Mark a specific version of a package as installed.
	This method is called when replaying a logged installation
	from before SqueakMap 1.07. Such logged installations lacked
	a timestamp and a count. We take the current time and a
	count starting from -10000 and upwards. This should keep
	the sorting order correct."

	^self registry noteInstalledPackage: uuidString version: version! !

!SMSqueakMap methodsFor: 'private-installation' stamp: 'gk 8/3/2004 13:24'!
noteInstalledPackage: uuidString version: version atSeconds: time number: num
	"Mark a package as installed in the Dictionary.
	This method is called when replaying a logged installation.
	<time> is the point in time as totalSeconds of the installation.
	<num> is the installCount of the installation.
	This method is typically called from a doIt in the changelog
	in order to try to keep track of packages installed."

	^self registry noteInstalledPackage: uuidString version: version atSeconds: time number: num! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'stephaneducasse 2/4/2006 20:38'!
noteInstalledPackageNamed: aString autoVersion: aVersion
	"Mark that the package release was just successfully installed.
	<aVersion> is the automatic version as a String.
	Can be used to inform SM of an installation not been done using SM."

	| p |
	p := self packageWithNameBeginning: aString.
	p ifNil: [self error: 'No package found with name beginning with ', aString].
	
^self noteInstalledPackage: p autoVersion: aVersion asVersion! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'gk 11/18/2003 11:59'!
noteInstalledPackageWithId: aPackageId autoVersion: aVersion
	"The package release was just successfully installed.
	Can be used to inform SM of an installation not been
	done using SM, even when the map isn't loaded."

	
^self noteInstalledPackageWithId: aPackageId
		autoVersion: aVersion
		name: '<unknown name>'! !

!SMSqueakMap methodsFor: 'private-installation' stamp: 'gk 8/3/2004 13:24'!
noteInstalledPackageWithId: uuidString autoVersion: version atSeconds: time number: num
	"Mark a package as installed in the Dictionary.
	This method is called when replaying a logged installation.
	<time> is the point in time as totalSeconds of the installation.
	<num> is the installCount of the installation.
	This method is typically called from a doIt in the changelog
	in order to try to keep track of packages installed."

	^self registry noteInstalledPackageWithId: uuidString autoVersion: version atSeconds: time number: num! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'gk 8/3/2004 13:24'!
noteInstalledPackageWithId: aPackageId autoVersion: aVersion name: aName
	"The package release was just successfully installed.
	Can be used to inform SM of an installation not been
	done using SM, even when the map isn't loaded.

	We record the fact in our Dictionary of installed packages
	and log a 'do it' to mark this in the changelog.
	The doit helps keeping track of the packages when
	recovering changes etc - not a perfect solution but should help.
	The map used is the default map.
	The id of the package is the key and the value is an OrderedCollection
	of Arrays with the release auto version, the point in time and the current installCounter."

	^self registry noteInstalledPackageWithId: aPackageId autoVersion: aVersion name: aName! !

!SMSqueakMap methodsFor: 'private-installation' stamp: 'gk 1/22/2004 12:01'!
noteUninstalled: aPackageRelease
	"The package release was just successfully uninstalled using SM.
	This is the method being called by SM upon a successful uninstallation.

	We record this in our Dictionary of installed package releases
	and log a 'do it' to mark this in the changelog.
	The map used is the default map."

	^self noteUninstalledPackageWithId: aPackageRelease package id asString
		autoVersion: aPackageRelease automaticVersion
		name: aPackageRelease package name! !

!SMSqueakMap methodsFor: 'private-installation' stamp: 'gk 8/3/2004 13:25'!
noteUninstalledPackageWithId: aPackageId autoVersion: aVersion name: aName
	"The package release was just successfully uninstalled.
	Can be used to inform SM of an uninstallation not been
	done using SM, even when the map isn't loaded.

	We record the fact in our Dictionary of installed packages
	and log a 'do it' to mark this in the changelog.
	The doit helps keeping track of the packages when
	recovering changes etc - not a perfect solution but should help.
	The map used is the default map.
	The id of the package is the key and the value is an OrderedCollection
	of Arrays with the release auto version, the point in time and the current installCounter."

	^self registry noteUninstalledPackageWithId: aPackageId autoVersion: aVersion name: aName! !

!SMSqueakMap methodsFor: 'queries' stamp: 'gk 11/18/2003 02:06'!
object: aUUID
	"Look up a categorizable object. Return nil if missing."

	^objects at: aUUID ifAbsent: [nil]! !

!SMSqueakMap methodsFor: 'queries' stamp: 'gh 12/1/2002 19:25'!
objectWithId: anIdString
	"Look up a categorizable object. Return nil if missing."

	^objects at: (UUID fromString: anIdString) ifAbsent: [nil]! !

!SMSqueakMap methodsFor: 'accessing' stamp: 'gh 11/28/2002 22:00'!
objects
	^objects! !

!SMSqueakMap methodsFor: 'public-packages' stamp: 'gk 7/13/2004 15:12'!
oldPackages
	"Answer all packages that are installed with a
	newer published version for this Squeak version available."

	^self installedPackages select: [:package | package isSafelyOld]! !

!SMSqueakMap methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
packageCacheDirectory
	"Return a FileDirectory for the package cache of the map.
	Creates it if it is missing."

	| dirName baseDir |
	dirName := self packageCacheDirectoryName.
	baseDir := self directory.
	(baseDir fileOrDirectoryExists: dirName)
		ifFalse:[baseDir createDirectory: dirName].
	^baseDir directoryNamed: dirName! !

!SMSqueakMap methodsFor: 'private' stamp: 'btr 5/28/2003 00:56'!
packageCacheDirectoryName
	"What is the name of the cache directory?"

	^'cache'! !

!SMSqueakMap methodsFor: 'queries' stamp: 'stephaneducasse 2/4/2006 20:38'!
packageReleaseWithId: anIdString 
	"Look up a package release. Return nil if missing.
	Raise error if it is not a package release."

	| r |
	r := self objectWithId: anIdString.
	r ifNil: [^nil].
	r isPackageRelease ifTrue:[^r].
	self error: 'UUID did not map to a package release.'! !

!SMSqueakMap methodsFor: 'accessing' stamp: 'gk 10/19/2005 23:53'!
packageReleases
	"Return subset of objects."

	objects ifNil: [^#()].
	^objects select: [:o | o isPackageRelease]! !

!SMSqueakMap methodsFor: 'queries' stamp: 'stephaneducasse 2/4/2006 20:38'!
packageWithId: anIdString 
	"Look up a package. Return nil if missing.
	Raise error if it is not a package."

	| package |
	package := self objectWithId: anIdString.
	package ifNil: [^nil].
	package isPackage ifTrue:[^package].
	self error: 'UUID did not map to a package.'! !

!SMSqueakMap methodsFor: 'queries' stamp: 'gh 12/1/2002 19:54'!
packageWithName: aName
	"Look up a package by exact match on name. Return nil if missing."

	^self packages detect: [:package | package name = aName ] ifNone: [nil]! !

!SMSqueakMap methodsFor: 'queries' stamp: 'stephaneducasse 2/4/2006 20:38'!
packageWithNameBeginning: aString
	"Look up a package beginning with <aString>. Return nil if missing.
	We return the shortest matching one. We also strip out spaces and
	ignore case in both <aString> and the names."

	| candidates shortest answer searchString |
	searchString := (aString asLowercase) copyWithout: Character space.
	candidates := self packages select: [:package |
		((package name asLowercase) copyWithout: Character space)
			beginsWith: searchString ].
	shortest := 1000.
	candidates do: [:package |
		package name size < shortest ifTrue:[answer := package. shortest := package name size]].
	^answer	! !

!SMSqueakMap methodsFor: 'queries' stamp: 'gk 11/11/2003 18:37'!
packageWithPI: aPIName
	"Look up a package by exact match on PackageInfo name. Return nil if missing."

	aPIName isEmptyOrNil ifTrue: [^nil].
	^self packages detect: [:package | package packageInfoName = aPIName ] ifNone: [nil]! !

!SMSqueakMap methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
packages
	"Lazily maintain a cache of all known package objects."

	packages ifNotNil: [^packages].
	objects ifNil: [^#()].
	packages := objects select: [:o | o isPackage].
	^packages! !

!SMSqueakMap methodsFor: 'queries' stamp: 'gk 12/3/2004 12:01'!
packagesByName
	"Return the packages sorted by their name."

	^self packages asSortedCollection: [:x :y | x name caseInsensitiveLessOrEqual: y name]! !

!SMSqueakMap methodsFor: 'private' stamp: 'gk 9/30/2003 17:03'!
pingServer: aServerName

	^self class pingServer: aServerName! !

!SMSqueakMap methodsFor: 'public' stamp: 'stephaneducasse 2/4/2006 20:38'!
purge
	"Clear out the map from memory. Use this to reclaim space,
	no information is lost because it does not remove information
	about what packages are installed, and the map itself is checkpointed
	to disk. Use #reload to get it back from the latest checkpoint on disk."

	objects := accounts := users := packages := categories := nil.
	checkpointNumber := 0.! !

!SMSqueakMap methodsFor: 'accessing' stamp: 'gk 5/5/2006 02:07'!
registry
	^registry ifNil: [registry := SMInstallationRegistry map: self]! !

!SMSqueakMap methodsFor: 'public' stamp: 'gk 3/31/2006 10:32'!
reload
	"Reload the map from the latest checkpoint on disk.
	The opposite of #purge."

	| fname stream map |
	fname := self lastCheckpointFilename.
	fname ifNil: [self error: 'No checkpoint available!!'].
	"Code below uses good ole StandardFileStream to avoid m17n issues (this is binary data) and
	also uses #unzipped since it works in older Squeaks"
	stream := (StandardFileStream oldFileNamed: (self directory fullNameFor: fname)) asUnZippedStream.
	"stream := (RWBinaryOrTextStream with: contents) reset."
	stream ifNil: [self error: 'Couldn''t open stream on checkpoint file!!'].
	[map := (stream fileInObjectAndCode) install arrayOfRoots first] ensure: [stream close].
	self copyFrom: map! !

!SMSqueakMap methodsFor: 'public-master' stamp: 'gk 11/17/2003 10:30'!
removeCategory: category inObject: object
	"Remove a category from an object."

	^object removeCategory: category
! !

!SMSqueakMap methodsFor: 'public-master' stamp: 'gk 5/5/2006 02:07'!
repair
	"Integrity repairs. This should not be neeed, but
	for some reason the map can obviously get messed up,
	not sure how."

	"SMSqueakMap default repair"
	
	"all objects should point back to me and not at another map"
	objects do: [:o | o map: self].
	
	"all releases should point back at the package they are in"
	self packages do: [:p | p releases do: [:r | r package: p]].
	
	"all releases in this map should point at a package in this map"
	self packageReleases do: [:r | | p |
		p := self object: r package id.
		p ifNil: [self error: 'Unknown package'].
		r package: p]! !

!SMSqueakMap methodsFor: 'checkpoints' stamp: 'gk 3/31/2006 09:34'!
saveCheckpoint: contentWithFilename
	"Save the map checkpoint to disk if it is not there already."

	| file directory sz fname content |
	directory := self directory.
	sz := contentWithFilename size.
	fname := contentWithFilename last: sz - (contentWithFilename lastIndexOf: $:).
	content := contentWithFilename first: sz - fname size - 1.
	(directory fileExists: fname) ifFalse: [
		[file := StandardFileStream newFileNamed: (directory fullNameFor: fname).
		file nextPutAll: content]
			ensure: [file close]]! !

!SMSqueakMap methodsFor: 'transactions' stamp: 'stephaneducasse 2/4/2006 20:38'!
setDirty
	"Set the map modified so that it will get written to disk."

	isDirty := true! !

!SMSqueakMap methodsFor: 'accessing' stamp: 'gk 7/13/2004 02:51'!
silent
	"Can installations ask questions or should they be silent
	and us good defaults?"

	^ silent ifNil: [false] ifNotNil: [true]! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'stephaneducasse 2/4/2006 20:38'!
silentlyDo: aBlock
	"Execute <aBlock> with the Silent flag set.
	This is a crude way of avoiding user interaction
	during batch operations, like loading updates."

	[silent := true.
	aBlock value]
		ensure: [silent := nil]! !

!SMSqueakMap methodsFor: 'private' stamp: 'stephaneducasse 2/4/2006 20:38'!
synchWithDisk
	"Synchronize myself with the checkpoints on disk.
	If there is a newer checkpoint than I know of, load it.
	If there is no checkpoint or if I have a higher checkpoint number,
	create a new checkpoint from me.

	The end result is that I am in synch with the disk and we are both as
	updated as possible."

	| checkpointNumberOnDisk |
	 "If there is no checkpoint, save one from me."
	(self isCheckpointAvailable) ifFalse: [^self createCheckpointNumber: checkpointNumber].
	"If the one on disk is newer, load it"
	checkpointNumberOnDisk := self lastCheckpointNumberOnDisk.
	(checkpointNumber < checkpointNumberOnDisk)
		ifTrue: [^self reload].
	"If I am newer, recreate me on disk"
	(checkpointNumberOnDisk < checkpointNumber)
		ifTrue: [^self createCheckpointNumber: checkpointNumber]! !

!SMSqueakMap methodsFor: 'queries' stamp: 'gh 12/1/2002 20:12'!
topCategories
	^self categories select: [:cat | cat isTopCategory]! !

!SMSqueakMap methodsFor: 'transactions' stamp: 'stephaneducasse 2/4/2006 20:38'!
transaction: aBlock
	"Execute aBlock and then make sure any modified SMObjects
	are committed to disk. We do this inside a mutex in order to
	serialize transactions. Transactions must be initiated from
	service methods in this class and not from inside the domain
	objects - otherwise they could get nested and a deadlock occurs."

"In first version of SM2 we simply set the isDirty flag,
when next client asks for updates, or 30 minutes has passed,
we checkpoint."

"	self mutex critical: ["
		aBlock value.
		isDirty := true
"	]"

"	self mutex critical: [
		dirtyList := OrderedCollection new.
		aBlock value.
		dirtyList do: [:obj | obj commit].
		dirtyList := nil
	]"! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'gk 11/17/2003 13:06'!
upgradeOldPackages
	"Upgrade all upgradeable old packages without confirmation on each."

	^self upgradeOldPackagesConfirmBlock: [:package | true ]! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'stephaneducasse 2/4/2006 20:38'!
upgradeOldPackagesConfirmBlock: aBlock
	"First we find out which of the installed packages are upgradeable and old.
	Then we upgrade them if confirmation block yields true.
	The block will be called with each SMPackage to upgrade.
	We return a Dictionary with the packages we tried to upgrade as keys
	and the value being the result of the upgrade, true or false."

	| result |
	result := Dictionary new.
	self upgradeableAndOldPackages
		do: [:package |
			(aBlock value: package)
				ifTrue:[ result at: package put: package upgrade]].
	^result
! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'gk 7/14/2004 17:33'!
upgradeOrInstallPackage: aPackage
	"Upgrade package (or install) to the latest published release for this Squeak version."

	^aPackage upgradeOrInstall! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'stephaneducasse 2/4/2006 20:38'!
upgradeOrInstallPackageWithId: anUUIDString
	"Upgrade package (or install) to the latest published release for this Squeak version."

	| package |
	package := self packageWithId: anUUIDString.
	package ifNil: [self error: 'No package available with id: ''', anUUIDString, ''''].
	^package upgradeOrInstall! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'gk 5/5/2006 02:07'!
upgradeOrInstallPackageWithId: anUUIDString asOf: aTimeStamp
	"Upgrade package (or install) to the latest published release as it was
	on <aTimeStamp> for this Squeak version. This ensures that the same
	release will be installed (for all Squeak versions) as when it was tested."

	| package |
	package := self packageWithId: anUUIDString.
	package ifNil: [self error: 'No package available with id: ''', anUUIDString, ''''].
	^package upgradeOrInstall! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'gk 7/14/2004 17:16'!
upgradePackage: aPackage
	"Upgrade package to the latest published release for this Squeak version."

	^aPackage upgrade! !

!SMSqueakMap methodsFor: 'public-installation' stamp: 'stephaneducasse 2/4/2006 20:38'!
upgradePackageWithId: anUUIDString
	"Upgrade package to the latest published release for this Squeak version.
	Will raise error if there is no release installed, otherwise use
	#upgradeOrInstallPackageWithId: "

	| package |
	package := self packageWithId: anUUIDString.
	package ifNil: [self error: 'No package available with id: ''', anUUIDString, ''''].
	^package upgrade! !

!SMSqueakMap methodsFor: 'public-packages' stamp: 'gh 10/25/2002 11:33'!
upgradeableAndOldOrInstallableAndNotInstalledPackages
	"This would give you all packages that are available now
	for automatic install or automatic upgrade."

	^self upgradeableAndOldPackages union: self installableAndNotInstalledPackages! !

!SMSqueakMap methodsFor: 'public-packages' stamp: 'gk 7/14/2004 16:17'!
upgradeableAndOldPackages
	"Answer all packages that are installed and which have a
	newer published release for this Squeak version that also
	can be to by an installer."

	^self installedPackages select: [:package | package isSafelyOldAndUpgradeable]! !

!SMSqueakMap methodsFor: 'private' stamp: 'gk 8/3/2003 23:13'!
usernameOccupied: aUsername
	"Return true if name already taken."

	^(self accountForUsername: aUsername) notNil! !

!SMSqueakMap methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:38'!
users
	"Lazily maintain a cache of all known account objects
	keyed by their developer initials."

	users ifNotNil: [^users].
	users := Dictionary new.
	self accounts do: [:a | users at: a initials put: a].
	^users! !

!SMSqueakMap methodsFor: 'private' stamp: 'gk 10/21/2003 23:05'!
verifyAdminPassword: aString
	"Answer true if it is the correct password."

	^adminPassword = (SecureHashAlgorithm new hashMessage: aString)! !

!SMSqueakMap methodsFor: 'views' stamp: 'gh 11/28/2002 22:06'!
viewFor: uiObject
	"This is a double dispatch mechanism for multiple views
	for multiple uis. Used primarily by the web frontend."

	^uiObject squeakMapViewOn: self! !

Object subclass: #SMUtilities
	instanceVariableNames: ''
	classVariableNames: 'MailServer MasterServer'
	poolDictionaries: ''
	category: 'SMBase-UI'!

!SMUtilities commentStamp: 'gk 11/13/2003 23:39' prior: 0!
Various loose functions in SM.!

!SMUtilities classSide methodsFor: 'class initialization' stamp: 'gk 5/5/2006 02:07'!
initialize
	"Initialize server settings."

	"self initialize"

	MasterServer := nil.
	MailServer := nil! !

!SMUtilities classSide methodsFor: 'server' stamp: 'gk 10/13/2005 00:18'!
isServer
	"Is this a running server?"
	
	^MasterServer notNil! !

!SMUtilities classSide methodsFor: 'private' stamp: 'gk 10/12/2005 23:10'!
mail: anAccount subject: sub message: msg
	"Send a mail to the holder of <anAccount>."

	SMTPClient
		deliverMailFrom: 'squeakmap@squeak.org'
		to: {anAccount email}
		text:
('From: SqueakMap <squeakmap@squeak.org>
To: ', anAccount email, '
Subject: ', sub,
'
', msg, (self randomPhrase), ', SqueakMap') squeakToIso usingServer: MailServer! !

!SMUtilities classSide methodsFor: 'server' stamp: 'gk 8/4/2003 16:15'!
mailPassword: randomPass for: anAccount
	"Change the password to a random generated one
	and mail it to the holder of the account."

	self mail: anAccount subject: 'New password at SqueakMap!!' message:
'Hi!!
An extra random password has been added for the account held by "', anAccount name, '":
"', randomPass, '"

You can login to SqueakMap at:

', MasterServer, '/login

The regular password still works, so if it was not you who requested this extra
random password you can safely just delete this email.

This extra password will stop working when you change your regular password.

'! !

!SMUtilities classSide methodsFor: 'server' stamp: 'gk 8/4/2003 16:15'!
mailPassword: aPassword forNew: anAccount
	"Mail the password to the person who just registered the account."

	self mail: anAccount subject: 'Your new account at SqueakMap!!' message:
'Hi!!
You or someone else has registered an account on SqueakMap. You can login to it using this link:

',
MasterServer, '/autologin?u=', anAccount initials, '&p=', aPassword,
'

If it was not you who performed this registration you can safely just delete this email.

'! !

!SMUtilities classSide methodsFor: 'class initialization' stamp: 'stephaneducasse 2/4/2006 20:38'!
mailServer: ipName masterServer: httpUrl
	"Initialize server settings."

	MailServer := ipName.
	MasterServer := httpUrl! !

!SMUtilities classSide methodsFor: 'private' stamp: 'rbb 3/1/2005 11:12'!
mailUserName
	"Answer the mail user's name, but deal with some historical mail senders."

	| mailSender |
	mailSender := (Smalltalk at: #MailSender ifAbsent: [ Smalltalk at: #Celeste ifAbsent: []]).
	^mailSender
		ifNil: [ UIManager default request: 'What is your email address?' ]
		ifNotNil: [ mailSender userName ]! !

!SMUtilities classSide methodsFor: 'server' stamp: 'gk 8/15/2003 12:25'!
masterServer
	"Return the master server url."

	^MasterServer! !

!SMUtilities classSide methodsFor: 'private' stamp: 'gh 11/27/2002 12:37'!
randomPhrase
	"Pick a nice phrase."

	^#('Debug safely' 'Happy Squeaking' 'Just do it' 'Yours truly' 'Stay a Squeaker' 'Squeak rocks') atRandom! !

!SMUtilities classSide methodsFor: 'utilities' stamp: 'gk 3/31/2006 09:50'!
sendMail: aString
	"Send the given mail message, but check for modern mail senders."

	| server |

	Smalltalk at: #MailSender ifPresent: [ :mailSender |
		^mailSender sendMessage: ((Smalltalk at: #MailMessage) from: aString).
	].

	Smalltalk at: #MailComposition ifPresent: [ :mailComposition |
		^mailComposition new
			messageText:  aString;
			open
	].
	
	Smalltalk at: #Celeste ifPresent: [ :celeste |
		celeste isSmtpServerSet ifTrue: [
			Smalltalk at: #CelesteComposition ifPresent: [ :celesteComposition |
				^celesteComposition
					openForCeleste: celeste current 
					initialText: aString
			]
		]
	].

	Smalltalk at: #AdHocComposition ifPresent: [ :adHocComposition |
		server := UIManager default request: 'What is your mail server for outgoing mail?'.
		^adHocComposition 
			openForCeleste: server
			initialText: aString
	].

	^self inform: 'Sorry, no known way to send the message'.
	 	! !

!SMUtilities classSide methodsFor: 'utilities' stamp: 'gk 7/10/2004 03:43'!
sendMailTo: recipient regardingPackageRelease: pr
	"Send mail to the given recipient. Try to use the first of:
	- MailSender (with its registered composition class)
	- Celeste
	- AdHocComposition
	for compatibility with 3.5 and 3.6 images"

	self sendMail: (String streamContents: [:stream |
		stream
			nextPutAll: 'From: '; nextPutAll: self mailUserName; cr;
			nextPutAll: 'To: '; nextPutAll: recipient; cr;
			nextPutAll: 'Subject: Regarding '; nextPutAll: pr printName; cr])! !

!SMUtilities classSide methodsFor: 'utilities' stamp: 'gk 3/31/2006 09:52'!
stripEmailFrom: aString
	"Picks out the email from:
		'Robert Robertson <rob@here.com>' => 'rob@here.com'
	Spamblockers 'no_spam', 'no_canned_ham' and 'spam_block'
	(case insensitive) will be filtered out."

	| lessThan moreThan email pos |
	lessThan := aString indexOf: $<.
	moreThan := aString indexOf: $>.
	(lessThan * moreThan = 0) ifTrue: [^ aString].
	email := (aString copyFrom: lessThan + 1 to: moreThan - 1) asLowercase.
	#('no_spam' 'no_canned_ham' 'spam_block') do: [:block |
		pos := email findString: block.
		pos = 0 ifFalse:[email := (email copyFrom: 1 to: pos - 1), (email copyFrom: pos + block size to: email size)]].
	^email! !

!SMUtilities classSide methodsFor: 'utilities' stamp: 'stephaneducasse 2/4/2006 20:38'!
stripNameFrom: aString
	"Picks out the name from:
		'Robert Robertson <rob@here.com>' => 'Robert Robertson'
	"

	| lessThan |
	lessThan := aString indexOf: $<.
	^(aString copyFrom: 1 to: lessThan - 1) withBlanksTrimmed ! !
SMSqueakMap initialize!
SMUtilities initialize!
